//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// drawMacros.h - Macros to be used in Drawtraces
//
// $Id: drawMacros.h,v 1.7 2000/10/19 15:27:36 goetz Exp $
//
////////////////////////////////////////////////////////////////////////

#ifndef _DRAWTRACES_H_
#define _DRAWTRACES_H_

// draw circle and light circle resp.
#define CIRCLE(pos, rad) "(c " << pos << " " << rad << ")"
#define LCIRCLE(pos, rad) "(c " << pos << " " << rad << "l)"

// draw circle with position and light circle with position resp.
// for pos use a Vector2d !!
#define GCIRCLE(pos, rad) "(c " << pos << " " << rad << "g)"
#define LGCIRCLE(pos, rad) "(c " << pos << " " << rad << "gl)"

// draw position and light position mark
// for pos use a Vector2d !!
#define POSITION(pos) "(p " << pos << ")"
#define LPOSITION(pos) "(p " << pos << " l)"

// draw a line and light line resp
// for start and end use Vector2d !!
#define LINE(start, end) "(l " << start << end << ")"
#define LLINE(start, end) "(l " << start << end << "l)"

// draw a vertical line (eg offside line)
#define VLINE(x) "(l (" << x << ", " << -FIELD_WIDTH/2 << ")(" << x << ", " << FIELD_WIDTH/2 << "))"
#define LVLINE(x) "(l (" << x << ", " << -FIELD_WIDTH/2 << ")(" << x << ", " << FIELD_WIDTH/2 << ") l)"

// draw a corridor and light corridor resp. (without the circle around target)
#define CORRIDOR(source, target, d, alpha) \
LINE( source, \
      source + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) + \
      radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) ) \
LINE( source + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) + \
      radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ), \
      target + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) ) \
LINE( target + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ), \
      target - radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) ) \
LINE( target - radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ), \
      source + radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) + \
      radAng2Vector2d( 0.5 * d, ( target - source ).perp().perp().perp().getAngle() ) ) \
LINE( source + radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) + \
      radAng2Vector2d( 0.5 * d, ( target - source ).perp().perp().perp().getAngle() ), \
      source )
#define LCORRIDOR(source, target, d, alpha) \
LLINE( source, \
       source + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) + \
       radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) ) \
LLINE( source + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) + \
       radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ), \
       target + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) ) \
LLINE( target + radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ), \
       target - radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ) ) \
LLINE( target - radAng2Vector2d( 0.5 * d, ( target - source ).perp().getAngle() ), \
       source + radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) + \
       radAng2Vector2d( 0.5 * d, ( target - source ).perp().perp().perp().getAngle() ) ) \
LLINE( source + radAng2Vector2d( 0.5 * d / tan( alpha ), ( target - source ).getAngle() ) + \
       radAng2Vector2d( 0.5 * d, ( target - source ).perp().perp().perp().getAngle() ), \
       source )
 
// draw a text string to a position pos (Vector2d)
// or draw a text into the corner
#define TEXT(text, pos) "(t " << pos << text << ")"
#define CTEXT(text) "(t (" << (-FIELD_LENGTH/2) << ", " << (-FIELD_WIDTH/2+3) << ") " << text << ")"

#endif
