//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// 	$Id: AdvSkills.C,v 1.33 2000/10/19 15:27:41 goetz Exp $
//      $Manager: aarn (Axel Arnold)$
//
//////////////////////////////////////////////////////////////////////

// for axels wonderful emacs-extension
// cwz-build-command: (cd ../player_axel; make all)

#include "constants.h"
#include "worldmodel.h"
#include "AdvSkills.h"
#include "mytrace.h"
#include "minmax.h"
#include "drawMacros.h"

#ifdef FIGS
#define KICKFIGS
#undef TURNHEADFIGS
#undef DRIBBLEFIGS
#define INTERCEPTDRAW(time, output)
//#define INTERCEPTDRAW(time, output) cout << time << "% " << output << endl
#undef TURNFIGS

#include <fstream.h>
#include <stdlib.h>

// a class writing out .fig-pictures. Good for debugging/optimizing
// Based on Miha's Positions.C code from RoboCup 98

class FigDraw
{
  ofstream    *out;
public:
  FigDraw(const char *str, const Snapshot *snap);
  ~FigDraw() { delete out; }

  // elementary functions
  void Circle(const Vector2d &p, double rad, int color = -1, int muster = -1);
  void Line(const Vector2d &p, const Vector2d &v, int color = -1);
  void Text(const Vector2d &p, const char *txt);
  void Text(const char *txt);
  void Value(const char *, double);
  void VecValue(const char *vecname, double x, double y);
  
private:

  static const int FigDrawFactor = 2400;
  static const int OffsetX        = 3*FigDrawFactor;
  static const int OffsetY        = 4*FigDrawFactor;
  
  double line;
  static int    serial_no;
};

int FigDraw::serial_no = 0;

void FigDraw::Circle(const Vector2d &p, double r, int color, int ms)
{
  int mx = OffsetX + int(FigDrawFactor*p.getX()); 
  int my = OffsetY - int(FigDrawFactor*p.getY());
  int mr = int(FigDrawFactor*r);
  
  *out << "1 3 0 1 " << color << " 0 0 0 "
       << ms << " 0.000 1 0.0000 "
       << mx << " " << my << " " << mr << " " << mr << " " << mx << " " << my
       << " " << mx+150 << " " << my+150 << endl;
}

void FigDraw::Line(const Vector2d &p, const Vector2d &v, int color)
{
  int mx = OffsetX + int(FigDrawFactor*p.getX()); 
  int my = OffsetY - int(FigDrawFactor*p.getY());
  int mx2 = OffsetX + int(FigDrawFactor*(p.getX()+v.getX())); 
  int my2 = OffsetY - int(FigDrawFactor*(p.getY()+v.getY()));
  
  *out << "2 1 0 1 "  << color << " 0 0 -1 0.000 0 0 -1 1 0 2" << endl
       << "0 0 1.00 60.00 120.00" << endl
       << mx << " " << my << " " << mx2 << " " << my2 << " " << endl;
}

void FigDraw::Text(const Vector2d &p, const char *txt)
{
  int mx = OffsetX + int(FigDrawFactor*p.getX()); 
  int my = OffsetY - int(FigDrawFactor*p.getY());

  *out << "4 0 -1 0 0 0 18 0.0000 4 180 1425 " << mx << " "
       << my << " " << txt << "\\001\n"
       << endl;
}

void FigDraw::Text(const char *txt)
{
  Text(Vector2d(1, -(line += 0.15)), txt);
}

FigDraw::FigDraw(const char *fn, const Snapshot *sn)
{
  char buf[256];
  char *id = getenv("USER");
  Vector2d mp = sn->my.pos;
  Vector2d c(0,0);
  sprintf(buf, "/tmp/figs_%s/p%05dt%03d-%s.fig", id, serial_no++,
	  sn->time, fn);
  cout << "-----" << buf << endl;
  out = new ofstream(buf);
  *out << "#FIG 3.1" << endl
       << "Landscape" << endl
       << "Center" << endl
       << "Inches" << endl
       << "1200 2" << endl;

  line = 0.0;

  Circle(c, PLAYER_SIZE, 0, 0);
  Line(c, sn->my.v, 1);
  Line(Vector2d(0, 0), radAng2Vector2d(0.5, sn->my.bodyDir), 3);
  Line(Vector2d(0, 0), radAng2Vector2d(0.5, sn->my.faceDir), 2);

  Circle(sn->ball.pos - mp, BALL_SIZE, 0, 0);
  Line(sn->ball.pos - mp, sn->ball.v, 1);

  sprintf(buf, "%s@%d", fn, sn->time);
  Text(buf);
  sprintf(buf, "ba:%d", sn->ball.age);
  Text(buf);
}

inline void FigDraw::VecValue(const char *text,
			      double a, double b)
{
  char buf[256];
  sprintf(buf, "%s (%f,%f)", text, a, b);
  Text(buf);
}

inline void FigDraw::Value(const char *text, double a)
{
  char buf[256];
  sprintf(buf, "%s = %f", text, a);
  Text(buf);
}

#else // no figures output, fake FigDraw

class FigDraw {
public:
  inline void makefig(const char *, const Snapshot *) {}
  inline void Circle(const Vector2d &, double, int, int = -1) {}
  inline void Line(const Vector2d &, const Vector2d &, int = -1) {}
  inline void Text(const Vector2d &, const char *) {}
  inline void Text(const char *) {}
  inline void Value(const char *, double) {}
  inline void VecValue(const char *, double, double) {}
};

#define INTERCEPTDRAW(time, output)
#endif




/////////////////////////////////////////////////// ENTIRE SKILLS STUFF

// const double KICKRANGE(BALL_SIZE+PLAYER_SIZE+KICKABLE_MARGIN);
const double MINDIST(BALL_SIZE+PLAYER_SIZE);

/*********************************** The constructor ************************/

AdvSkills::AdvSkills(Comm *c):
  // pseudo constants
  precision(1e-5),
  kickRange(PLAYER_SIZE + BALL_SIZE + 0.8*KICKABLE_MARGIN),
  minSpeed(0.1),
  prepareOuterLimit(3*kickRange),
  prepareInnerLimit(2*kickRange),
  minInterceptTime(30),
  goToThreshStretch(0.008),
  turnMinAngle(DEG_TO_RAD(15)),
  turnLatency(2),
  directThreshold(0.1),
  laybackErr(0.25),
  innerSafeCircle(PLAYER_SIZE+BALL_SIZE+0.3*KICKABLE_MARGIN),
  safeCircle(PLAYER_SIZE+BALL_SIZE+0.55*KICKABLE_MARGIN),
  stopableV(0.5*KICK_POWER_RATE*MAXPOWER),
  dribbleBodyErr(DEG_TO_RAD(10)),
  dribbleHoldErr(DEG_TO_RAD(30)),
  dribbleMaxV(PLAYER_SPEED_MAX),
  controlMaxSteps(20),
  // real vars
  comm(c),
  turnH_time(-20)
{}

/********************************* the AdvSkills ****************************/


/************************************ kick *********************************/
#include "AdvSkillsKick.C"

/********************************** intercept ******************************/
#include "AdvSkillsIntercept.C"

/*********************************** dribble *******************************/
#include "AdvSkillsDribble.C"

void AdvSkills::changeView(const Snapshot *sn, view_width width)
{
  if (width != sn->senseBody.viewWidth) {
    comm->sendChangeView(width, sn->senseBody.viewQuality);
  }
}

void AdvSkills::changeView(const Snapshot *sn, view_quality quality)
{
  if (quality != sn->senseBody.viewQuality) {
    comm->sendChangeView(sn->senseBody.viewWidth, quality);
  }
}

void AdvSkills::changeView(const Snapshot *sn, view_width width,
			   view_quality quality)
{
  if ((width != sn->senseBody.viewWidth) ||
      (quality != sn->senseBody.viewQuality)) {
    comm->sendChangeView(width, quality);
  }
}

void AdvSkills::basicDash (double d) {
  comm->sendDash(d);
}

void AdvSkills::basicKick (double a,double b) {
  comm->sendKick(a,b);
}

void AdvSkills::basicTurn (double d) {
  //TRACE("TURN");
  comm->sendTurn(d);
}

void AdvSkills::basicCatch (double d) {
  comm->sendCatch(d);
}

void AdvSkills::noAction(void) {
  comm->noAction();
}

void AdvSkills::basicMove(double x, double y) {
  comm->sendMove(x,y);
}

void AdvSkills::sendTurnNeck(double dir)
{
  comm->sendTurnNeck(dir);
}

void AdvSkills::sendChangeView(view_width width, view_quality quality)
{
  comm->sendChangeView(width, quality);
}

void AdvSkills::sendSenseBody()
{
  comm->sendSenseBody();
}

/*********************************** catch *******************************/

bool AdvSkills::catchSafe(const Snapshot* sn) 
{
  
  return sn->my.ballCatchable (sn->ball, 
			       (sn->ball.pos - sn->my.pos).getAngle()
			       - sn->my.bodyDir );
}

AdvSkills::CatchCommand AdvSkills::goalieCatch(const Snapshot *sn)
{
  CatchCommand com;
  com.catchDir = (sn->ball.pos - sn->my.pos).getAngle();
  // check if player is goalie and if ball is in catchable area
  if(!comm->amIGoalie()) {
    com.state = CNOTGOALIE;
    return com;
  }
  if ((sn->time - sn->lastCatchTime) < CATCH_BAN_CYCLE) {
    com.state = CBAN;   
    return com;
  }
  if (!catchSafe(sn)) {
    com.state = CRANGE;
    return com;
  }
  com.state = COK;
  return com;
}

void AdvSkills::exec(const Snapshot *sn, const AdvSkills::CatchCommand &com) {
  comm->sendCatch(com.catchDir - sn->my.bodyDir);
}

/*********************************** turn *******************************/
AdvSkills::TurnCommand AdvSkills::turn(const Snapshot *sn, Angle a)
{
  TurnCommand com;

#ifdef TURNFIGS
  FigDraw draw("TURN", sn);
  draw.Line(Vector2d(0,0), radAng2Vector2d(1, a), 1);
#endif

  // NOT an angle since we need to know if we turn to hard (inertiaMoment)
  double delta = double(a - sn->my.bodyDir) * sn->my.inertiaMoment();
  com.state = (fabs(delta) < turnMinAngle) ? RFINISHED : ROK;

  if (delta >  M_PI * MAXMOMENT / 180.0) {
    delta = M_PI * MAXMOMENT / 180.0;
    com.state = RANGLE;
  }

  if (delta < M_PI * MINMOMENT / 180.0) {
    delta = M_PI * MINMOMENT / 180.0;
    com.state = RANGLE;
  }

  com.turnMoment = delta;

  com.bodyDir = sn->my.bodyDir + Angle(com.turnMoment / sn->my.inertiaMoment());
  
#ifdef TURNFIGS
  draw.Line(Vector2d(0,0), radAng2Vector2d(1, com.bodyDir), 4);
#endif
  return com;
}

void AdvSkills::exec(const Snapshot *, const AdvSkills::TurnCommand &com) { 
  basicTurn(com.turnMoment);
  return;
}

/***************************** turnHead *******************************/

AdvSkills::TurnHeadCommand AdvSkills::turnHead(const Snapshot *sn, Angle a)
{
  TurnHeadCommand com;
  com.state = THOK;

#ifdef TURNHEADFIGS
  FigDraw draw("TURNHEAD", sn);
  draw.Line(Vector2d(0,0), radAng2Vector2d(1, a), 4);
#endif

#ifdef TURNHEADFIGS
  draw.Value("TH", turnH_time);
  draw.Value("TL", turnLatency);
  draw.Value("ST", sn->time);
#endif
    
  if (turnH_time + turnLatency > sn->time) {
#ifdef TURNHEADFIGS
    draw.Text("Latency interval");
    draw.Line(Vector2d(0,0), radAng2Vector2d(1, turnH_oldHead), 6);
    draw.Line(Vector2d(0,0), radAng2Vector2d(1, turnH_reqHead), 7);
#endif
    // in latency, check if last turn_neck seems to be done
    if (fabs(turnH_oldHead - sn->my.faceDir) <
	fabs(turnH_reqHead - sn->my.faceDir)) {
#ifdef TURNHEADFIGS
      draw.Text("Delayed for latency");
#endif
      com.state = THBAN;
      return com;
    }
  }

  a -= sn->my.bodyDir;

  if (a > M_PI * MAXNECKANG / 180.0) {
    a = M_PI * MAXNECKANG / 180.0;
    com.state = THANGLE;
  }
  else if (a < M_PI * MINNECKANG / 180.0) {
    a = M_PI * MINNECKANG / 180.0;
    com.state = THANGLE;
  }

#ifdef TURNHEADFIGS
  draw.Line(Vector2d(0,0), radAng2Vector2d(1, a + sn->my.bodyDir), 5);
#endif

  // NOT an angle, since we need to know if we turn too hard
#ifdef TURNHEADFIGS
  draw.Value("a", a);
#endif
  double delta = double(a) - double(sn->senseBody.headAngle);

  if ((com.state == THOK) && (fabs(delta) < turnMinAngle))
    com.state = THFINISHED;

#ifdef TURNHEADFIGS
  draw.VecValue("delta, ha", delta, sn->senseBody.headAngle);
#endif

  if (delta > M_PI * MAXNECKMOMENT / 180.0) {
    delta = M_PI * MAXNECKMOMENT / 180.0;
    com.state = THANGLE;
  }
  else if (delta < M_PI * MINNECKMOMENT / 180.0) {
    delta = M_PI * MINNECKMOMENT / 180.0;
    com.state = THANGLE;
  }

  com.faceDir = sn->my.faceDir + Angle(delta);
  com.turnHeadMoment = delta;
  return com;
}

void AdvSkills::exec(const Snapshot *sn, const AdvSkills::TurnHeadCommand &com) {
  // ignore turnHead while banned
  if (com.state != THBAN) {
    sendTurnNeck(com.turnHeadMoment);
    turnH_oldHead = sn->my.faceDir;
    turnH_reqHead = com.faceDir;
    turnH_time = sn->time;
  }
  return;
}
