//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: comm.C,v 1.29 2000/10/19 15:27:44 goetz Exp $
//  $Manager: eri$
// 
// 
//  comm.C - Communication with soccer server
// 
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <iostream.h>
#include <sys/file.h>

#include <fcntl.h>
#include <errno.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "comm.h"

//
// Remember: slow_down_factor (since sserver-6.06) has not been implemented yet
//

///////////////////////////////////////////////////////////////////////

#undef CERR
#define CERR(output) cerr << output << endl

//////////////////////////////////////////////////////////////////////


// prompt initialization process
#undef INITTRACE
#define INITTRACE(cmd)

// prompt all communication to and from server
#undef SERVERTRACE
#define SERVERTRACE(cmd)

// general prompting of work
#undef TRACE
#define TRACE(cmd)

// prompt incoming signals (careful - just for debugging purposes !! )
// ( Enabling this will almost surely cause segfaults during a normal game.
//   It will also mess up the output stream. )
#undef SIGNALTRACE
#define SIGNALTRACE(cmd)

// prompt encoding/decoding of say messages
#undef SAYTRACE
#define SAYTRACE(cmd)

#ifndef CERR
#define CERR(output)
#warning Comm error output macro CERR disabled!
#endif

////////////////////////////////////////////////////////////////////


// define this symbol to use pause instead of sigsuspend 
// (pause is more reliable at the moment, 
//  sigsuspend is the more elegant solution)
#undef USE_PAUSE
#define USE_PAUSE


/////////////////////////////////////////////////////////////////////


// initialize the static Comm pointer once at the beginning, to define it.
Comm *(Comm::theComm) = NULL;

// static member methods for signal handling
void Comm::alarmHandler(int signo) {

  if (signo) {} // prevent warning when compiling
  if (Comm::theComm != NULL) {
    SIGNALTRACE( cout << "SIGALRM" << endl);
    Comm::theComm->alarmAction(); 
  } else {
    CERR("Comm: cannot handle SIGALRM as Comm is missing !" );
  }
  
}

void Comm::ioHandler(int signo) {

  if (signo) {} // prevent warning when compiling
  if (Comm::theComm != NULL) {
    SIGNALTRACE( cout << "SIGIO" << endl);
    Comm::theComm->ioAction(); 
  } else {
    CERR("Comm: cannot handle SIGIO as Comm is missing !" );
  }

}
  

///////////////////////////////////////////////////////////////////////////


Comm::Comm(const char* _team) {
  
  TRACE( cout << "Comm: constructor entered" << endl);

  if (Comm::theComm == NULL) { // only if no other Comm exists yet
    Comm::theComm = this; // store static reference

    sd = -1;
    serverOk = false;

    sigemptyset(&iomask);
    sigaddset(&iomask, SIGIO);
    sigemptyset(&alrmask);
    sigaddset(&alrmask, SIGALRM);
    alarmFlag = 0;
    ioFlag    = 0;
    
    // [HACK]
    info = new AsyncCommInfo(_team, this);
    parser   = new Parser();
    store    = new SensorDataStore();
    audiomgr = new AudioMgr();

    hadMessage = false;
    noMessageCount = 0;
    printCommands  = false;
    
  } else {
    CERR("Comm: cannot construct more than one Comm "
	 << "in any single process !");
  }

  TRACE( cout << "Comm: constructor completed" << endl);

}


Comm::~Comm() {

  TRACE( cout << "Comm: destructor entered" << endl);

  // log out
  if (serverOk) sendToServer ("(bye)");

  removeSigHandlers();
  
  // remove static reference
  Comm::theComm = NULL;
  
  delete info;
  if (sd >= 0) close(sd);
  if (parser) delete parser;
  if (store) delete store;
  if (audiomgr) delete audiomgr;

  TRACE( cout << "Comm: destructor completed" << endl);

}

int
Comm::getTimeExpression() {
  const int MAX_SEC = 100;
  struct timeval now;
  gettimeofday (&now, NULL);
  long time_ms = (now.tv_sec%MAX_SEC)*1000 + now.tv_usec/1000;
  return time_ms % 10000;
}

comm_error 
Comm::sendToServer(char *msg) {

  // Prints the string and sends it to the server via the UDP socket.
      
  SERVERTRACE( cout << getServerTime() << " systime: " << getTimeExpression() 
	       << " snd: "<< msg << endl);
  
  if (printCommands) {
    if (strncmp(msg, "(say", 4) != 0) { // avoid logging say commands
      cout << info->getServerTime() << "> " << msg << endl;
    }
  }
  
  int len = strlen(msg);
  int actual;
  do { // repeat sending if interrupted
    actual = sendto(sd, msg, len, 0, (struct sockaddr *)
		    &server_addr, sizeof(server_addr));
  } while (actual < 0 && errno == EINTR);
  
  if (actual != len) return COMM_ERROR;
  else return COMM_OK;
}


comm_error 
Comm::receiveFromServer(char *msg, long timeout_msec) {
  // select() the socket to find out if data is there. Then keep receiving
  // until the whole string was read from the socket, and return it.
  
  fd_set                watchMe;
  timeval               timeout;
  int	                n;

#ifdef Linux
  unsigned int          fromlen;
#else
  int                   fromlen;
#endif

  struct sockaddr_in    from_addr;
  bool                  repeatSelect;
  
  // create the file descriptor set to be watched by select()
  FD_ZERO (&watchMe);
  FD_SET (sd, &watchMe);
  timeout.tv_sec  = (1000*timeout_msec) / 1000000;
  timeout.tv_usec = (1000*timeout_msec) % 1000000;
  
  do {
    repeatSelect = false;
    switch (select (sd+1, &watchMe, NULL, NULL, &timeout)) {
      
    case 1: // the socket has data
      fromlen = sizeof(from_addr) ;
      
      do { // repeat receiving if interrupted
	n = recvfrom(sd, msg, MAX_SERVER_MSG_SIZE, 0,
		     (struct sockaddr *)&from_addr, &fromlen);
      } 
      while ((n == -1) && (errno == EINTR));
      
      if (n<0) return COMM_ERROR; // not an EINTR error !
      msg[n] = 0;

      SERVERTRACE(cout << getServerTime() << " systime: " << getTimeExpression() 
		  << " rcv: " << msg << endl);

      // sender of this message will be next adressee (when receiving init)
      server_addr.sin_port = from_addr.sin_port;

      return COMM_OK; 
      
    case 0: // socket had no data and timed out
      return COMM_ERROR_NO_DATA;
      
    case -1: // select() returned an error
      if (errno != EINTR) return COMM_ERROR; // select returned a serious error
      else repeatSelect = true; // repeat on interrupt
    } //switch
  } //do
  while (repeatSelect);
    
  // if control would reach this point, something were seriously wrong
  return COMM_ERROR;
  
}

void 
Comm::alarmAction(){

  alarmFlag = 1;
  
}
  
void
Comm::finishAlarm() {
  
  TRACE(cout << "Comm: handling SIGALRM" << endl);
  
  if (hadMessage) {
    noMessageCount = 0;
    hadMessage = false;
  } else {
    TRACE(cout << "Comm: no message from server since last alarm (for " 
	  << noMessageCount+1 << " steps now)" << endl);
    noMessageCount++;
  }
  
  if (noMessageCount > COMM_MAX_DEAD_SERVER_CYCLES) {
    TRACE(cout << "Comm: no messages for more than " 
	  << COMM_MAX_DEAD_SERVER_CYCLES 
	  << " supposed timesteps, server is dead." << endl);
    serverOk = false;
  }
  
}

void 
Comm::ioAction(){

  ioFlag = 1;

}

void 
Comm::finishIo() {

  TRACE(cout << "Comm: handling SIGIO" << endl);
  char msg [MAX_SERVER_MSG_SIZE];
  
  while (receiveFromServer(msg, 0) == COMM_OK) {
    
    hadMessage = true;
    
    SensorData data;
    if (!parser->parseServerString(msg, data)) {
      CERR(getServerTime() 
	   << " Comm (" << info->getTeam() << " No." << info->getNumber()
	   << "): could not parse server string " << msg);
    } else {
      
      // use data to adjust time, play mode etc.
      bool firstSenseBody = info->adjust(data); 

      switch (data.type) {

      case SENSOR_TYPE_OK:
	TRACE(cout << "Comm: skipping ok message" << endl);
	break;
      case SENSOR_TYPE_ERROR: 
	CERR(getServerTime()
	     << " Comm (" << info->getTeam() << " No." << info->getNumber()
	   << "): received error message from server: "
	     << msg );
	break;
	
      case SENSOR_TYPE_SENSE_BODY:
	if (info->isTimeRunning() && firstSenseBody) {
	  // first sense_body of a timestep received while clock is running:
	  // insert it before all data from same timestep 
	  // (needed by worldmodel)
	  TRACE(cout << "Comm: inserting first sense_body data into store" 
		<< endl);
	  store->insert(data);
	} else {
	  // additional sense_body, or sense_body received while clock 
	  // is stopped: just store it according to receive time
	  TRACE(cout << "Comm: storing sense_body data" << endl);
	  store->put(data);
	}
	break;
      case SENSOR_TYPE_EXACT:
	TRACE(cout << "Comm: storing exact data" << endl);
	store->put(data);
	break;
      case SENSOR_TYPE_SEE:
	TRACE(cout << "Comm: storing see data" << endl);
	store->put(data); // worldmodel can handle twisted neck
	break;
      case SENSOR_TYPE_HEAR:
	
	switch (data.sender) {
	case SENDER_SELF: 
	  TRACE(cout << "Comm: skipping own message" << endl);
	  break;
	case SENDER_REFEREE:
	  TRACE(cout << "Comm: skipping referee play mode" << endl);
	  // information contained was already used in info->adjust() above, 
	  // so also forget about it
	  break;
	case SENDER_DIR:    // message from player heard by player
	  TRACE(cout << "Comm: storing heard message" << endl);
	  SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		   << data.message << endl);
	  audiomgr->storeMsg(AudioMsg(data.message));
	  break;
	case SENDER_PLAYER: // message from player heard by any coach
	  TRACE(cout << "Comm: storing player message" << endl);
	  SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		   << data.message << endl);
	  audiomgr->storeMsg(AudioMsg(data.message));
	  break;
	case SENDER_COACH:  // message from train coach heard by player/olcoach
	  if (info->isTrainCoach()) {
	    TRACE(cout << "Comm: skipping own message" << endl);
	  } else {
	    TRACE(cout << "Comm: storing coach message relayed "
		  << "through the referee" << endl);
	    SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		     << data.message << endl);
	    audiomgr->storeMsg(AudioMsg(data.message));
	  }
	  break;
	case SENDER_ONLINECOACH:// msg from onlinecoach heard by player/coach
	  if (info->isOnlineCoach()) {
	    TRACE(cout << "Comm: skipping own message" << endl);
	  } else {
	    TRACE(cout << "Comm: storing onlinecoach message" << endl);
	    SAYTRACE(cout << "Comm: decoding via AudioMgr " 
		     << data.message << endl);
	    audiomgr->storeMsg(AudioMsg(data.message));
	  }
	  break;
	} // switch on sender

	break;
	
      } //switch on sensor_type
      
    } //parse ok

  } // while data

}


comm_error 
Comm::createSocket (const char *hostname, const int port) {
  // lookup the server's address, create a socket and bind it
  
  INITTRACE( cout << "Comm:     creating socket at " 
	     << hostname << ":" << port << endl);

  // Init remote and local addresses
  memset(&server_addr, 0, sizeof(sockaddr_in));
  server_addr.sin_family = AF_INET;

  struct sockaddr_in local_addr;
  memset(&local_addr, 0, sizeof(sockaddr_in));
  local_addr.sin_family = AF_INET ;

  INITTRACE( cout << "Comm:         looking network address for " 
	     << hostname << endl);
  // Look up the server's network address
  struct hostent *hp;
  if ((hp = gethostbyname(hostname)) != NULL)
    memcpy(&server_addr.sin_addr, hp->h_addr, sizeof(server_addr.sin_addr));
  else 
    if ((server_addr.sin_addr.s_addr = inet_addr(hostname)) == 
	(unsigned long)-1) return COMM_ERROR_SERVER_NOT_FOUND;
  
  server_addr.sin_port = htons(port);


  INITTRACE( cout << "Comm:         creating socket on remote port " 
	     << port << endl);
  // Create UDP socket
  if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) 
    return COMM_ERROR_SOCKET_INIT;

  // get SIGIO delivered to current process
  if (fcntl (sd, F_SETOWN, getpid())) return COMM_ERROR_SOCKET_INIT;
  
  // make socket non-blocking
  int val = fcntl(sd, F_GETFL, 0) ;
  //val |= O_ASYNC; // BSD-style option, does not work properly !
  val |= FASYNC; 
  // from CMU98: undocumented option, but the only one that works ;-)
  if (fcntl(sd, F_SETFL, val) ) return COMM_ERROR_SOCKET_INIT;
  
  INITTRACE( cout << "Comm:         binding local address to socket" << endl);
  // Bind local address
  local_addr.sin_addr.s_addr = htonl(INADDR_ANY) ;
  local_addr.sin_port = htons(0) ;
  if (bind(sd, (struct sockaddr *)&local_addr, sizeof(local_addr)) < 0) {
    return COMM_ERROR_SOCKET_INIT;
  }
  INITTRACE( cout << "Comm:     socket sucessfully opened" << endl);
  
  return COMM_OK;

}

void
Comm::generateCheckinMessage(char* outmsg, 
			     client_type type, int reconnectNumber) {
  
  if (reconnectNumber <= 0) { // player first init or any coach
    switch (type) {
    case CLIENT_GOALIE:
      sprintf(outmsg, "(init %s (version 6.00) (goalie))", 
	      info->getTeam()); break;
    case CLIENT_FIELDPLAYER:
      sprintf(outmsg, "(init %s (version 6.00))", info->getTeam()); break;
    case CLIENT_ONLINECOACH:
      sprintf(outmsg, "(init %s (version 6.00))", info->getTeam()); break;
    case CLIENT_TRAINCOACH:
      sprintf(outmsg, "(eye on)"); break;
    default: 
      break;
    }
  } else {
    switch (type) {
    case CLIENT_GOALIE:
    case CLIENT_FIELDPLAYER:
      sprintf(outmsg, "(reconnect %s %d)", 
	      info->getTeam(), reconnectNumber); break;
    case CLIENT_ONLINECOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
    case CLIENT_TRAINCOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
    default:
      break;
    }
  } 

}

comm_error
Comm::parseCheckinResponse(char* inmsg, client_type type, 
			   int reconnectNumber, playing_side coachSide) {

  char   c;
  char   playmode [MAX_SERVER_MSG_SIZE];

  // needed to parse play mode returned by server
  Parser* dummy = new Parser("", 0, SIDE_UNKNOWN, false);

  int          time;
  bool         secondHalf;
  int          number = -1;
  play_mode    mode   = PLAY_MODE_UNKNOWN;
  playing_side side   = SIDE_UNKNOWN;
  
  if (reconnectNumber <= 0) { // parse the init response
    time = 0;
    secondHalf = false;
    switch (type) {
    case CLIENT_GOALIE:
    case CLIENT_FIELDPLAYER:
      if (sscanf(inmsg, "(init %c %d %[^)])", &c, &number, playmode) != 3)
	return COMM_ERROR_PARSE;
      if (c == 'l') side = SIDE_LEFT;
      else if (c == 'r') side = SIDE_RIGHT;
      else return COMM_ERROR_PARSE;
      mode = dummy->parsePlayMode(playmode);
      break;
    case CLIENT_ONLINECOACH:
      if (sscanf(inmsg, "(init %c ok)" , &c) != 1) 
	return COMM_ERROR_PARSE;
      else {
	number = 0;
	if (c == 'l') side = SIDE_LEFT;
	else if (c == 'r') side = SIDE_RIGHT;
	else return COMM_ERROR_PARSE;
	mode = PLAY_MODE_BEFORE_KICK_OFF;
	sendEye(true);
      }
      break;
    case CLIENT_TRAINCOACH:
      if (strcmp(inmsg,"(ok eye on)")) return COMM_ERROR_PARSE;
      else {
	number = 0;
	side = coachSide;
	if (coachSide == SIDE_LEFT)
	  mode = PLAY_MODE_KICK_OFF_US;
	else 
	  mode = PLAY_MODE_KICK_OFF_THEM;
	sendEar(true);
      }
      break;
    default:
      break;
    }
  } else {  // parse the reconnect response
    time = 3000;
    secondHalf = true;
    switch (type) {
    case CLIENT_GOALIE:
    case CLIENT_FIELDPLAYER:
      if (sscanf(inmsg, "(reconnect %c %[^)])", &c, playmode) != 2)
	return COMM_ERROR_PARSE;
      number = reconnectNumber;
      if (c == 'l') side = SIDE_LEFT;
      else if (c == 'r') side = SIDE_RIGHT;
      else return COMM_ERROR_PARSE;
      mode = dummy->parsePlayMode(playmode);
      break;

    case CLIENT_ONLINECOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
      break;
      
    case CLIENT_TRAINCOACH:
      ensure(reconnectNumber == 0); // coach cannot reconnect
      break;
      
    default:
      break;
    }
  }
  
  info->setStates(type, number, time, mode, side, secondHalf);
  
  delete (dummy);
  return COMM_OK;

}

comm_error 
Comm::checkinAtServer (client_type type,
		       int reconnectNumber, playing_side coachSide) {
  
  INITTRACE( cout << "Comm:     ";
	     switch(type) {
	     case CLIENT_GOALIE: cout << "goalie "; break;
	     case CLIENT_FIELDPLAYER: cout << "fieldplayer "; break;
	     case CLIENT_ONLINECOACH: cout << "onlinecoach "; break;
	     case CLIENT_TRAINCOACH: cout << "trainingcoach "; break;
	     default: break;
	     }
	     if (reconnectNumber > 0) cout << "reconnecting" << endl;
	     else cout << "checking into server" << endl;);
  
  
  char outmsg[MAX_SERVER_MSG_SIZE];
  char inmsg[MAX_SERVER_MSG_SIZE];

  generateCheckinMessage (outmsg, type, reconnectNumber);
  
  // send message to server
  INITTRACE (cout << "Comm:     sending checkin message " 
	     << outmsg << " to server" << endl);
  if (sendToServer(outmsg) != COMM_OK) return COMM_ERROR_SERVER_INIT;
  
  INITTRACE( cout << "Comm:     waiting for server response" << endl);
  
  if (receiveFromServer(inmsg, COMM_MAX_DEAD_SERVER_CYCLES*SIMULATOR_STEP) 
      != COMM_OK)
    return COMM_ERROR_NO_DATA;
  INITTRACE (cout << "Comm:     received server response " 
	     << inmsg << endl);
  
  comm_error ce;
  ce = parseCheckinResponse (inmsg, type, reconnectNumber, coachSide);
  if (ce != COMM_OK) {
    CERR("Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): cannot handle server checkin response " << inmsg );
    return ce;
  }

  INITTRACE( cout << "Comm:     completed checkin at sserver" << endl);

  return COMM_OK;
  
}

comm_error 
Comm::initSigHandlers() {

  INITTRACE( cout << "Comm:     initializing signal handlers" << endl);

  struct sigaction saction;
  
  saction.sa_mask = alrmask; // SIGIO-exec blocks SIGALRM
  saction.sa_flags = SA_RESTART;
  saction.sa_handler = Comm::ioHandler; //(void (*) (...)) Comm::ioHandler;
  if (sigaction (SIGIO, &saction, NULL)) return COMM_ERROR_SIGNAL_INIT;
  
  saction.sa_mask = iomask; // SIGALRM-exec blocks SIGIO
  saction.sa_flags = SA_RESTART;
  saction.sa_handler = Comm::alarmHandler; //(void (*) (...)) Comm::alarmHandler;
  if (sigaction (SIGALRM, &saction, NULL)) return COMM_ERROR_SIGNAL_INIT;
  
  return COMM_OK;

}

comm_error 
Comm::setAlarmInterval(int msec) {
  
  INITTRACE(cout << "Comm:     setting alarm interval to " <<msec<< " msec" <<endl);

  long usec = msec*1000;
  
  itimerval itv;
  
  // time left in current interval
  itv.it_value.tv_sec     = usec / 1000000;  
  itv.it_value.tv_usec    = usec % 1000000;
  
  // stadard interval time
  itv.it_interval.tv_sec  = usec / 1000000;
  itv.it_interval.tv_usec = usec % 1000000;
  
  if (setitimer (ITIMER_REAL, &itv, NULL )) return COMM_ERROR_SIGNAL_INIT;
  else return COMM_OK;
  
}

void 
Comm::removeSigHandlers() {

  INITTRACE(cout << "Comm:     removing all signal handlers" << endl);

  struct sigaction saction;

  sigset_t emptymask;
  sigemptyset(&emptymask);

  saction.sa_mask = emptymask;
  saction.sa_flags = SA_RESTART;
  saction.sa_handler = SIG_DFL;

  sigaction(SIGIO, &saction, NULL);
  sigaction(SIGALRM, &saction, NULL);
  
}

comm_error
Comm::blockSignals() {
  
  INITTRACE(cout << "Comm:     blocking all signals" << endl);

  if ( sigprocmask(SIG_BLOCK, &alrmask, NULL) ||
       sigprocmask(SIG_BLOCK, &iomask, NULL)  )
    return COMM_ERROR_SIGNAL_INIT;
  else return COMM_OK;
}

comm_error
Comm::unblockSignals() {

  INITTRACE(cout << "Comm:   unblocking all signals" << endl);

  if ( sigprocmask(SIG_UNBLOCK, &alrmask, NULL) ||
       sigprocmask(SIG_UNBLOCK, &iomask, NULL)  )
    return COMM_ERROR_SIGNAL_INIT;
  else return COMM_OK;
}


comm_error 
Comm::init(client_type type,
	   const char* hostname,
	   int port,
	   int reconnectNumber,
	   bool _printCommands,
	   playing_side coachSide) {
  
  INITTRACE(cout << "Comm: initializing and checking in" << endl);

  printCommands = _printCommands;
  
  comm_error e;
  
  e = blockSignals(); if(e) return e;
  
  e = createSocket (hostname, port); if (e) return e;
  
  e = checkinAtServer (type, reconnectNumber, coachSide); 
  if (e) return e;

  // set parameters known after receiving first info

  audiomgr->setStates(info->getNumber(), info->getSide()==SIDE_LEFT );
  
  parser->setStates(info->getTeam(), info->getNumber(), 
		    info->getSide(), info->isCoach());
  
  e = initSigHandlers(); if (e) return e;
  
  e = setAlarmInterval(SIMULATOR_STEP); if (e) return e;
  
  e = unblockSignals(); if(e) return e;

  serverOk = true;
  
  INITTRACE(cout << "Comm: init sucessfully completed" << endl);

  return COMM_OK;
}


bool 
Comm::getNextData(SensorData &data) {
  return store->getNext(data);
}

void
Comm::getLastAction(CycleAction &action) {
  info->getLastAction(action);
}

void 
Comm::handleSignals() {

#ifdef USE_PAUSE

  TRACE(cout << "Comm: pausing" << endl);
  pause();

#else

  sigset_t oldmask;
  sigset_t blockmask;
  
  sigemptyset(&blockmask);
  sigaddset(&blockmask, SIGIO);
  sigaddset(&blockmask, SIGALRM);
  
  TRACE(cout << "Comm: blocking signals" << endl);
  sigprocmask(SIG_BLOCK, &blockmask, &oldmask); // block all
  
  TRACE(cout << "Comm: suspending execution" << endl);
  sigsuspend(&oldmask);

#endif

  if (alarmFlag) { 
    alarmFlag = 0;
    finishAlarm();
  }
  
  if (ioFlag) {
    ioFlag = 0;
    finishIo();
  }

#ifdef USE_PAUSE
#else
  TRACE(cout << "Comm: unblocking signals" << endl);
  sigprocmask(SIG_UNBLOCK, &blockmask, NULL);
#endif

}

bool 
Comm::amIGoalie() {
  return info->isGoalie();
}

//////////////////////////////////////////////////////////////////////////


bool 
Comm::isServerAlive() {
  return serverOk;
}

bool
Comm::isGameOver() {
  return (info->getPlayMode() == PLAY_MODE_TIME_OVER);
}


int
Comm::getServerTime() {
  return info->getServerTime();
}

bool 
Comm::isTimeRunning() {
  return info->isTimeRunning();	   
}

bool 
Comm::isActionNeeded() {
  return info->actNow();
}

const char*  
Comm::getMyTeam() {
  return info->getTeam();
}

int 
Comm::getMyNumber() {
  return info->getNumber();
}

play_mode
Comm::getPlayMode() {
  return info->getPlayMode();
}

playing_side 
Comm::getMySide() {
  return info->getSide();
}

int
Comm::getOurScore() {
  return info->getOurScore();
}

int
Comm::getOtherScore() {
  return info->getOtherScore();
}

bool
Comm::inSecondHalf() {
  return info->inSecondHalf();
}

void
Comm::say(AudioMsg &msg) {

  const char* charmsg = audiomgr->SpeakTo(msg);
  SAYTRACE(cout << "Comm: encoding via AudioMgr " << msg << endl);
  sendSay(charmsg);
  
}

bool
Comm::getNextMsg(AudioMsg &msg) {

  bool hadMsg = audiomgr->getNextMsg(msg);
  SAYTRACE(if (hadMsg) 
	   cout << "Comm: retrieved from AudioMgr " << msg << endl);
  return hadMsg;
  
}

void 
Comm::noAction() {

  TRACE(cout << "Comm: agent doing no action" << endl);
  info->setSentAction(NO_COMMAND, info->getServerTime(), 0, 0);

}

void 
Comm::sendKick    (double power, double dir_rad) { 

  double dir_deg = RAD_TO_DEG( dir_rad );

  if (power < MINPOWER - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MINPOWER );
    power = MINPOWER;
  } else if (power > MAXPOWER + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MAXPOWER );
    power = MAXPOWER;
  } else if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (power >= MINPOWER - EPS) && 
		 (power <= MAXPOWER + EPS) && 
		 (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN somewhere ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (kick " << power << " " 
	 << dir_deg << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(kick %f %f)", power, dir_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->setSentAction(KICK_COMMAND, info->getServerTime(), 
			power, DEG_TO_RAD( dir_deg ));
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendDash    (double power) { 

  if (power < MINPOWER - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (dash " << power 
	 << "), normalizing to " << MINPOWER );
    power = MINPOWER;
  } else if (power > MAXPOWER + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (dash " << power 
	 << "), normalizing to " << MAXPOWER );
    power = MAXPOWER;
  } else if (! ( (power >= MINPOWER - EPS) && 
		 (power <= MAXPOWER + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (dash " << power 
	 << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(dash %f)", power);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->setSentAction(DASH_COMMAND, info->getServerTime(), power, 0);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendTurn    (double moment_rad) { 

  double moment_deg = RAD_TO_DEG( moment_rad );

  if (moment_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn " << moment_deg 
	 << "), normalizing to " << MINMOMENT );
    moment_deg = MINMOMENT;
  } else if (moment_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn " << moment_deg 
	 << "), normalizing to " << MINMOMENT );
    moment_deg = MAXMOMENT;
  } else if ( ! ( (moment_deg >= MINMOMENT - EPS) && 
		  (moment_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn " << moment_deg 
	 << "), dropping command completely" );
    return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(turn %f)", moment_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->setSentAction(TURN_COMMAND, info->getServerTime(), 
			moment_deg * M_PI / 180.0, 0);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
  
}

void 
Comm::sendCatch   (double dir_rad) { 

  double dir_deg = RAD_TO_DEG( dir_rad );

  if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime()
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (catch " << dir_deg 
	 << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime()
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (catch " << dir_deg 
	 << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime()
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (catch " << dir_deg 
	 << "), dropping command completely" );
      return;
  }
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(catch %f)", dir_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->setSentAction(CATCH_COMMAND, info->getServerTime(), 
			DEG_TO_RAD( dir_deg ), 0);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
  
}

void 
Comm::sendMove    (double x, double y) { 

  if ( (!info->isGoalie() && info->isTimeRunning() ) ||
       (info->isGoalie() && info->isTimeRunning() && 
	(info->getPlayMode() != PLAY_MODE_FREE_KICK_US) ) ) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << ") game is running, agent not allowed to (move)" );
    return;
  } else {

    // no range checking when moving: there are no limits !
    
    char msg [MAX_SERVER_MSG_SIZE];
    sprintf(msg, "(move %f %f)", x, y);
    if (sendToServer(msg) == COMM_OK) {
      TRACE(cout << "Comm: agent sent " << msg << endl);
      info->setSentAction(MOVE_COMMAND, info->getServerTime(), x, y);
    } else {
      CERR(getServerTime() 
	   << " Comm (" << info->getTeam() << " No." << info->getNumber()
	   << "): agent cannot send " << msg << " to server" );
    }
  }
  
}

void 
Comm::sendTurnNeck   (double angle_rad) { 

  double angle_deg = RAD_TO_DEG( angle_rad );

  if (angle_deg < MINNECKMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn_neck " << angle_deg 
	 << "), normalizing to " << MINNECKMOMENT );
    angle_deg = MINNECKMOMENT;
  } else if (angle_deg > MAXNECKMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn_neck " << angle_deg 
	 << "), normalizing to " << MAXNECKMOMENT );
    angle_deg = MAXNECKMOMENT;
  } else if (! ( (angle_deg >= MINNECKMOMENT - EPS) && 
		 (angle_deg <= MAXNECKMOMENT + EPS) ) ) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (turn_neck " << angle_deg 
	 << "), dropping command completely" );
    return;
  }
  
  // to make sure there are no (see) infos left on the socket, just read 
  // and parse all outstanding IO into the SensorData buffer:
  TRACE(cout << "Comm: reading all IO just before turning neck" << endl);
  finishIo();
  
  char msg [MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(turn_neck %f)", angle_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    
    info->incTurnNeckCount();

    /* no need to check for twisted neck, worldmodel can handle it anyway.
       
       info->twistNeck();
       
    */
    
    /* server handles turn_neck synchronously, no need to send sense_body !
       
       // if (see) info is received with a twisted neck an warning message 
       // will be printed. Only after a sense_body arrived with a confirmed 
       // head_angle will these warnings be removed. 
       // ( see AsyncInfo::adjust() and AsyncInfo::isNeckTwisted() )
       
       sendSenseBody();
    
    */
    
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
  
}

void 
Comm::sendChangeView (view_width width, view_quality quality) { 
  
  char *w = "";
  switch (width) {
  case VIEW_WIDTH_NARROW: w = "narrow"; break;
  case VIEW_WIDTH_NORMAL: w = "normal"; break;
  case VIEW_WIDTH_WIDE: w = "wide"; break;
  }

  char *q = "";
  switch (quality) {
  case VIEW_QUALITY_LOW: q = "low"; break;
  case VIEW_QUALITY_HIGH: q = "high"; break;
  }

  char msg[MAX_SERVER_MSG_SIZE];
  sprintf(msg, "(change_view %s %s)", w, q);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime()
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
  
}

void 
Comm::sendSenseBody  () {

  if (sendToServer("(sense_body)") == COMM_OK) {
    TRACE(cout << "Comm: agent sent (sense_body)" << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send (sense_body) to server" );
  }
}

void 
Comm::sendSay(const char *message) { 

  unsigned int maxSize = MAX_SAY_MSG_SIZE;
  if (info->isCoach()) maxSize = SAY_COACH_MSG_SIZE;
  // [!] wie lang darf duerfen die messages hoechstens sein (+-10 Zeichen?)
  if (strlen(message)+10 > maxSize) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): client tries to send too long say message, dropping "
	 << "it completely" );
    return;
  }
  
  char msg[maxSize];
  sprintf(msg, "(say %s)", message);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
    info->incSayCount();
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}


void 
Comm::sendChangeMode(play_mode playMode) { 
  char* newMode;
  switch (playMode) {
  case PLAY_MODE_BEFORE_KICK_OFF: newMode = "before_kick_off"; break;
  case PLAY_MODE_CORNER_KICK_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "corner_kick_l";
    else newMode = "corner_kick_r";
    break;
  case PLAY_MODE_CORNER_KICK_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "corner_kick_r";
    else newMode = "corner_kick_l";
    break;
  case PLAY_MODE_FREE_KICK_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "free_kick_l";
    else newMode = "free_kick_r";
    break;
  case PLAY_MODE_FREE_KICK_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "free_kick_r";
    else newMode = "free_kick_l";
    break;
  case PLAY_MODE_GOAL_KICK_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "goal_kick_l";
    else newMode = "goal_kick_r";
    break;
  case PLAY_MODE_GOAL_KICK_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "goal_kick_r";
    else newMode = "goal_kick_l";
    break;
  case PLAY_MODE_KICK_IN_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "kick_in_l";
    else newMode = "kick_in_r";
    break;
  case PLAY_MODE_KICK_IN_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "kick_in_r";
    else newMode = "kick_in_l";
    break;
  case PLAY_MODE_KICK_OFF_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "kick_off_l";
    else newMode = "kick_off_r";
    break;
  case PLAY_MODE_KICK_OFF_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "kick_off_r";
    else newMode = "kick_off_l";
    break;
  case PLAY_MODE_GOAL_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "goal_l";
    else newMode = "goal_r";
    break;
  case PLAY_MODE_GOAL_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "goal_r";
    else newMode = "goal_l";
    break;
  case PLAY_MODE_OFFSIDE_US: 
    if (info->getSide()==SIDE_LEFT) newMode = "offside_l";
    else newMode = "offside_r";
    break;
  case PLAY_MODE_OFFSIDE_THEM: 
    if (info->getSide()==SIDE_LEFT) newMode = "offside_r";
    else newMode = "offside_l";
    break;
  case PLAY_MODE_PLAY_ON : newMode = "play_on"; break;
  case PLAY_MODE_TIME_OVER : newMode = "time_over"; break;
  default: 
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): trying to set unknown play mode: No." 
	 << (int)playMode << ", droping command" ); 
    return;
  }
  
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(change_mode %s)", newMode);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendMoveObject (const char *object, double x, double y) { 
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(move %s %f %f)", object, x, y);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendMoveObject (const char *object, double x, double y, 
		      double dir_rad) { 

  double dir_deg = RAD_TO_DEG( dir_rad );

  if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), dropping command completely" );
    return;
  }
  
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(move %s %f %f %f)", object, x, y, dir_deg);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendMoveObject (const char *object, double x, double y, 
		      double dir_rad, double xvel, double yvel) { 
  
  double dir_deg = RAD_TO_DEG( dir_rad );

  if (dir_deg < MINMOMENT - EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MINMOMENT );
    dir_deg = MINMOMENT;
  } else if (dir_deg > MAXMOMENT + EPS) {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), normalizing to " << MAXMOMENT );
    dir_deg = MAXMOMENT;
  } else if (! ( (dir_deg >= MINMOMENT - EPS) && 
		 (dir_deg <= MAXMOMENT + EPS) ) ) {
    // NaN ?
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent tries to send (move " << object << " " 
	 << x << " " << y << " " 
	 << dir_deg << "), dropping command completely" );
    return;
  }
  
  char msg[MAX_SERVER_MSG_SIZE];
  sprintf (msg, "(move %s %f %f %f %f %f)", object, x, y, dir_deg, xvel, yvel);
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendEye        (bool eyeOn) {
  
  char* msg;
  if (eyeOn) msg = "(eye on)";
  else msg = "(eye off)";
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendEar        (bool earOn) { 
  char* msg;
  if (earOn) msg = "(ear on)";
  else msg = "(ear off)";
  if (sendToServer(msg) == COMM_OK) {
    TRACE(cout << "Comm: agent sent " << msg << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send " << msg << " to server" );
  }
}

void 
Comm::sendLook       () { 
  if (sendToServer("(look)") == COMM_OK) {
    TRACE(cout << "Comm: agent sent (look)" << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send (look) to server" );
  }
}  

void 
Comm::sendCheckBall  () { 
  if (sendToServer("(check_ball)") == COMM_OK) {
    TRACE(cout << "Comm: agent sent (check_ball)" << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send (check_ball) to server" );
  }
}

void 
Comm::sendRecover  () { 
  if (sendToServer("(recover)") == COMM_OK) {
    TRACE(cout << "Comm: agent sent (recover)" << endl);
  } else {
    CERR(getServerTime() 
	 << " Comm (" << info->getTeam() << " No." << info->getNumber()
	 << "): agent cannot send (recover) to server" );
  }
}
