//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
//
// $Id: worldmap.h,v 1.7 2000/10/19 15:27:48 goetz Exp $	
// $Manager: goetz$
// 
// 
//  worldmap.h - World model display
// 
//  ( Christian Bauer <cebix> )
//   Goetz Schwandtner
//
//
//////////////////////////////////////////////////////////////////////

#ifndef _WORLDMAP_H_
#define _WORLDMAP_H_

#include "worldmodel.h"

#include <X11/Xlib.h>
#include <X11/keysym.h>

#include "coordinates.h"

// Display of actual worldmodel data
class WorldMap{
public:
  WorldMap(const char *windowtitle = NULL, double mag = 5.0);
  ~WorldMap();
  // Open map window. True if oK
  bool Open();
  void Close();

  // Display snapshot with onbjects with age <= max_age only
  void Map(Snapshot *snap, int max_age);

  // Some drawing functions:
  // draw a circle with radius rad, center (x,y)
  void drawCircle(const double x, const double y, const double rad= 1, 
		  const bool lines= true, const bool bright= false);
  void drawLine(const double x1, const double y1, 
		const double x2, const double y2,
		const bool bright= false);
  void drawText(const double x, const double y, 
		const char *text, const bool bright= false);

protected:
  // Constants and inline functions as elements to override in subclasses

  // Map magnification factor
  double MAGNIFY;

  // Velocities are displayed in meters per simulation step
  // Velocity magnification factor 
  double VELOCITY_MAG;

  // PLayer magnification factor
  double PLAYER_MAG;

  // Player size on screen
  int PSIZE;

  // Length factor of head angle lines 
  int VIEW_CONE_LENGTH;

  // max age of elements to be displayed as "now"
  int MAX_AGE_AS_NOW;

  // Functions for converting field coordinates to pixmap coordinates
  // X and Y coordinates 
  int mx(double x) {return int((x + FIELD_LENGTH/2) * MAGNIFY);}
  int my(double y) {return int((y + FIELD_WIDTH/2) * MAGNIFY);}	
  // Distances
  int md(double d) {return int(d * MAGNIFY);}			
  // Velocities
  int mv(double d) {return int(d * MAGNIFY * VELOCITY_MAG);}	

  // their inverses
  double inv_mx(int x) { return (double) x / MAGNIFY - FIELD_LENGTH/2; }
  double inv_my(int y) { return (double) y / MAGNIFY - FIELD_WIDTH/2; }
  double inv_md(int d) { return (double) d / MAGNIFY; }
  double inv_mv(int v) { return (double) v / (MAGNIFY * VELOCITY_MAG); }


  // Internal Constants
  int WIN_WIDTH; 
  int WIN_HEIGHT;

  // Internal elements
  char *title;
  Display *the_disp;
  int the_screen;
  
  Window the_root;
  Window the_win;
  Colormap the_cmap;
  GC win_gc;
  GC map_gc;
  Pixmap the_map;
  // Available colors
  // color names beginning with l "light" are used for now seen objects
  XColor red_color;
  XColor lred_color;
  XColor green_color;
  XColor lgreen_color;
  XColor blue_color; 
  XColor yellow_color;
  XColor lyellow_color;
  XColor lightgrey_color; 
  XColor darkgrey_color;
  XColor pink_color;
  XColor dpink_color;
  unsigned long black;
  unsigned long white;
  unsigned long red;
  unsigned long lred;
  unsigned long green;
  unsigned long lgreen;
  unsigned long blue;
  unsigned long yellow;
  unsigned long lyellow;
  unsigned long lightgrey;
  unsigned long darkgrey;
  unsigned long pink;
  unsigned long dpink;

  // private drawing functions
  void drawviewcone(Vector2d pos, Angle orient, Angle viewangle, 
		    view_quality viewqu);
  void drawlines();
  void drawownplayer(WorldPlayer ownplayer);
  void drawball(WorldBall ball);
  void drawplayer(WorldPlayer player, int max_age);
  void drawmyteam(Snapshot *snap, int max_player_age);
  void drawotherteam(Snapshot *snap, int max_player_age);
};

#endif
