//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
//
// $Id: worldmap_r.h,v 1.4 2000/10/19 15:27:49 goetz Exp $	
// $Manager: goetz$
// 
// 
//  worldmap_r.h - World model display with additional relative map
// 
//  ( Christian Bauer <cebix> )
//   Goetz Schwandtner
//
//
//////////////////////////////////////////////////////////////////////

#ifndef _WORLDMAP_R_H_
#define _WORLDMAP_R_H_

#include "worldmodel.h"
#include "worldmap.h"
#include "coordinates.h"

#include <X11/Xlib.h>
#include <X11/keysym.h>


// Display of actual worldmodel data
class WorldMap_r : public WorldMap {
public:
  // create new worldmaps
  // absolute worldmap has magnification factor "mag"
  // relative worldmap has magnification factor "relmag"
  // If "relmag" is set to negative values, "relmag" gets same value as
  // "mag". If "relmag" is set to 0.0, it means that there will be no
  // relative map at all
  WorldMap_r(const char *windowtitle = NULL, 
	     double mag = 5.0,
	     double relmag = -1.0);
  ~WorldMap_r();
  // Open map window. True if oK
  bool Open();
  void Close();

  // Display snapshot with onbjects with age <= max_age only
  void Map(Snapshot *snap, int max_age);
protected:
  // Is a relative map wished ?
  bool has_rel_map;
  
  // Map magnification factor of relative map
  double REL_MAGNIFY;

  // Player size on screen
  int REL_PSIZE;

  // Length factor of head angle lines 
  int REL_VIEW_CONE_LENGTH;

  // Functions for converting field coordinates to pixmap coordinates
  // X and Y coordinates 
  int mx(double x) {return int((x + FIELD_LENGTH/2) * REL_MAGNIFY);}
  int my(double y) {return int((y + FIELD_WIDTH/2) * REL_MAGNIFY);}	
  // Distances
  int md(double d) {return int(d * REL_MAGNIFY);}			
  // Velocities
  int mv(double d) {return int(d * REL_MAGNIFY * VELOCITY_MAG);}	

  // their inverses
  double inv_mx(int x) { return (double) x / REL_MAGNIFY - FIELD_LENGTH/2; }
  double inv_my(int y) { return (double) y / REL_MAGNIFY - FIELD_WIDTH/2; }
  double inv_md(int d) { return (double) d / REL_MAGNIFY; }
  double inv_mv(int v) { return (double) v / (REL_MAGNIFY * VELOCITY_MAG); }

  // Internal Constants
  int REL_WIN_WIDTH; 
  int REL_WIN_HEIGHT;

  // Position and velocity of own player for relative map window
  // this is set by setownpos(Worldplayer)
  double ownx;
  double owny;
  double ownvx;
  double ownvy;
  void setownposandvel(WorldPlayer ownplayer);

  // Internal elements
  char *rel_title;
  Window rel_win;
  GC rel_gc;
  GC rel_map_gc;
  Pixmap rel_map;
  // private drawing functions
  void drawviewcone(Angle orient, Angle viewangle, view_quality viewqu);
  void drawownplayer(WorldPlayer ownplayer);
  void drawball(WorldBall ball);
  void drawplayer(WorldPlayer player, int max_age);
  void drawmyteam(Snapshot *snap, int max_player_age);
  void drawotherteam(Snapshot *snap, int max_player_age);
};

#endif
