//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module goalkickmodule
//
// \$Id: goalkickmodule.h,v 1.21 2000/10/19 15:27:53 goetz Exp $
// \$manager: schappel$
//
//////////////////////////////////////////////////////////////////////


#ifndef __GOALKICKMODULE_H__
#define __GOALKICKMODULE_H__

#include "grade.h"
#include "goalkickeval.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"


// the GOALTRACE
////////////////////////

#undef GOALTRACE

// use this line if you want prompting of players decision process
//#define GOALTRACE(out) cout << worldSnap->time << " GOAL: " << out << endl 

#ifndef GOALTRACE
#define GOALTRACE(out)
#else
#warning GOALTRACE enabled
#endif

#undef GOALTRACESHORT

// use this line if you want prompting of players decision process
//#define GOALTRACESHORT(out) cout << "GOAL: " << out << endl 

#ifndef GOALTRACESHORT
#define GOALTRACESHORT(out)
#else
#warning GOALTRACESHORT enabled
#endif



class GoalKickModuleInfo;


// the class GoalKickModule
///////////////////////////////

class GoalKickModule
{
public:
  // constructor of module GoalKick 
  GoalKickModule(bool isGoalie, GoalKickEval *_goalkickeval);

  ~GoalKickModule();
  
  // feed information about new world and tactics into module
  bool feedbackWorld(const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap);

  // Feed audio information into module. This method is called for
  // each audio message. audioMsg.getSource() is player No or 0 for coach
  // Audiomsg audioInfo is acting like a stream.
  // usage: audioMsg >> auInfo (where GoalKickModuleInfo auInfo holds the information
  // for module GoalkickModule in other players)"
  bool feedbackAudio(AudioMsg &audioMsg);

  // Do the evaluation for the next action
  // For definition of Grade: see grade.h
  Grade evaluate(AdvSkills* skills,
		 const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap);

  // Perform the action for this module  
  bool act(AdvSkills* skills,
	   const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap);

  // description of stream AudioMsg: see above. Here you should implement:
  // audioMsg << audioMsg;
  bool getSayMsg(AudioMsg &audioMsg);

private:
  // some constants this class needs
  //////////////////////////////////////

  
  // is the player the goalie ?
  bool goalie;       

  // elder things will not be considered
  int maxAge;   

  // some variables this class needs
  //////////////////////////////////////

  // the goalkickeval-instance
  GoalKickEval *goalkickeval;

  // to store the messages that shall be sent
  GoalKickModuleInfo *sendInfo;

  // did the player prepare a superkick and must do a rekick?
  bool mustRekick;   

  // did the player a goalkick, that has been finished?
  bool finishedKick;

  // time of the last snapshot the module received;
  // to check if feedback, evaluate and act have the same judgement-basis
  int lastSnapshotTime;

  // the player himself
  WorldPlayer me;

  // the time the player has to prepare his kick-command
  int availableKickTime;
  
  // which point in the goal should be used to do the goal kick
  Vector2d goalTarget;

  // enough time to use a superkick ?
  bool useSuperKick;

  // enough time to do a fixed kick ?
  bool useFixedKick;

  // turn to optimal direction before kicking?
  bool turnBeforeKick;

  // for remembering a calculated kick-command
  AdvSkills::ExtKickCommand kickParams;
  

  // some private methods
  /////////////////////////////////////////

  // updates the above datastorage of the module by means of a snapshot
  bool updateDataStorage(const Snapshot *worldSnap);
  
  // calculates if a given target is reachable with a kick that only has
  // availableTime to do the kick
  bool targetReachable(const Snapshot *worldSnap, AdvSkills *skills, 
		       Vector2d target, int availableTime);


  // determines the best corridor for the player and evaluates it; 
  // returns the grade for the best corridor and the belonging target
  Grade searchAndEvalBestCorridor(const Snapshot *worldSnap, AdvSkills *skills, 
				  Vector2d &target, int availableTime);
  

}; // class GoalKickModule


// Class to keep information for modules of other agents.
//////////////////////////////////////////////////////////////////////

class GoalKickModuleInfo
{
public:

  // constructor of Info class for module GoalKickModule
  GoalKickModuleInfo(bool _valid);
  ~GoalKickModuleInfo();

  // flag to determine if instance of GoalKickModuleInfo is valid
  bool valid;
  
}; // class GoalKickModuleInfo

// put information to AudioMsg-Stream 
class AudioMsg &operator<< (AudioMsg &aumsg, const GoalKickModuleInfo &info);

// fill with information from AudioMsg-Stream 
class AudioMsg &operator>> (AudioMsg &aumsg, GoalKickModuleInfo &info);

#endif

