//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module PassModule
//
// \$Id: passmodule.h,v 1.34 2000/10/19 15:27:54 goetz Exp $
// \$manager: holgi$
//
//////////////////////////////////////////////////////////////////////

#ifndef __PASSMODULE_H__   
#define __PASSMODULE_H__

#include "passeval.h"
#include "goalkickeval.h"

#include "grade.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"
#include "drawMacros.h"

// the TRACEs
/////////////////////

#undef PASSTRACE
// comment out the following line to disable trace
// #define PASSTRACE(cmd) cmd
#ifndef PASSTRACE
#define PASSTRACE(cmd)
#else
#warning PASSTRACE enabled
#endif

#undef DRAWTRACE
// comment out the following line to disable trace
// #define DRAWTRACE(output) cout << worldSnap->time << "% " << output << endl
#ifndef DRAWTRACE
#define DRAWTRACE(cmd)
#else
#warning DRAWTRACE enabled
#endif

#undef TESTPASSTRACE
// comment out the following line to disable trace
// #define TESTPASSTRACE(cmd) cmd
#ifndef TESTPASSTRACE
#define TESTPASSTRACE(cmd)
#else
#warning TESTPASSTRACE enabled
#endif

// the class PassModule
//////////////////////////

class PassModule
{
public:
  // constructor of module PassModule
  PassModule(PassEval* _passeval, const GoalKickEval* _goalkickeval);
  ~PassModule();
  
  // feed information about new world and tactics into module
  bool feedbackWorld(const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap);

  // feed audio information into module
  bool feedbackAudio(AudioMsg &audioMsg);

  // Do the evaluation for the next action
  // For definition of Grade: see grade.h
  Grade evaluate(AdvSkills *skills, const Snapshot *worldSnap,
                 TacticalSnapshot *tacticalSnap);

  // Perform the action for this module
  bool act(AdvSkills *skills, const Snapshot *worldSnap, TacticalSnapshot *tacticalSnap);

  // write passmodule audio information into audioMsg stream
  bool getSayMsg(AudioMsg &audioMsg);

private:
  // recall of this module needed, if true
  bool callMeAgain;

  // we want to use our own passeval-method
  PassEval* passeval;

  // and perhaps the goalkickeval-method
  const GoalKickEval* goalkickeval;

  // time of latest snapshot the module received;
  // to check if feedback, evaluate and act work with the same snapshot
  int lastSnapshotTime;

  // time available for a kick
  int time_for_kick;

  // property-array for playerpasses
  PropOf pprops[12];

  // property array for rangepasses:
  PropOf rprops[12];

  // information on the passtarget
  /////////////////////////////////////////
  bool     target_selected;
  bool     target_breed;     // true: pass to player; false: rangepass
  int      target_number;
  Vector2d target_pos;

  double   target_vel;       // desired velocity when ball leaves player
  int      target_corridor;  // width of free corridor to target

  Vector2d target_source;

  // emergency kick away
  /////////////////////////
  Vector2d em_vec;    // direction of emergency-kick-away
  bool     em_case;   // true if we have an emergency kick situation

  // information concerning the looking after ball
  //////////////////////////////////////
  bool look_act;              // true, if act shall only look for ball
  bool look_for_ball;         // true, if player shall look for ball
  int  look_for_ball_time;    // time, looking for ball started
  Vector2d look_for_ball_pos; // position, where player is looking for the ball

  // information about the kick-commands
  //////////////////////////////////////////
  AdvSkills::ExtKickCommand kickparams;

  // needed "constants"
  //////////////////////////
  // value between 0 and 1, 1 means MAXPOWER, 0 means nothing (look in get_externGrade())
  double OPP_POWER_FACTOR;

  // some private methods
  /////////////////////////////////////////

  // updates the above datastorage of the module by a snapshot
  bool updateDataStorage(const Snapshot *worldSnap);

  // this method helps to gather some information
  // it fills the pprop- and the rprop-arrays;
  bool getPasses(const Snapshot *wS);

  // calculate emergency kick away vector (=em_vec)
  Vector2d get_ePass(const Snapshot *wS, TacticalSnapshot *tS );

  // lock the pass target
  Grade lockPass(PropOf prop);

  // get an extern grade
  // depending on the pass with internGrade from player to partner with kickmode
  // calculate an extern grade;
  // consider if opponents can reach the passline (shortest way) before ball passes
  // the meeting point; refines the grade (just downgrade, if necessary)
  Grade get_externGrade(const Snapshot *wS, Vector2d player, Vector2d partner,
			Grade internGrade, int playerTimeToKick) const;
}; // class PassModule


//////////////////////////////////////////////////////////////////////
//
// Class to keep information for modules of other agents.
//
//////////////////////////////////////////////////////////////////////


class PassModuleInfo
{
};  // class PassModuleInfo

// put information to AudioMsg-Stream

AudioMsg &operator<< (AudioMsg &aumsg, const PassModuleInfo &info);

// fill with information from AudioMsg-Stream 

AudioMsg &operator>> (AudioMsg &aumsg, PassModuleInfo &info);

#endif

