//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
//
// MRB player module StandardSituationModule
//
// \$Id:
// \$manager: gauer$
//
//////////////////////////////////////////////////////////////////////

//////////////////////////////
// changes since saturday
/////////////////////////////////////

//
// - removed unneccesary DRAWTRACEs from going[Backwards]To()
//   and replaced a DECISIONTRACE
// - walkAround players is now more inteligent and uses stamina 
//   so header changed to
// - fixed a bug in calculating the second pos when it is offside
//   in calcPlayers.*(KickInUs|FreeKickUs)()
// - calcWayTo() calcs a way out of the circle now if from is inside
// - avoidBall() is more exact when running a way form the ball now
//

#include "standardsituationmodule.h"
#include "Config.h"

// ------------------------------------------------------------------

bool AfterStandardSituationPlan::executePlan(AdvSkills *skills, const Snapshot *worldSnap)
{
  switch (thePlan)
    {
    case PLAN_SUPERKICK:
      {
	if (time > 0)
	  {
	    time --;
	    if (skills->kickSafe(worldSnap))
	      {
		kickCommand = skills->rekick(worldSnap, kickCommand);
		skills->exec(worldSnap, kickCommand);
		PLANTRACE( " Execute Superkick, time = " << time );
		return true; 
	      }	    
	    PLANTRACE( " planned Superkick seems to be ready (ball not kickable), time =  "
		       << time );
	  }
	else
	  PLANTRACE ( " planned Superkick interrupted because of time-out " );
	setNothing();
	return false; 
      }     
      
    case PLAN_RUN_TO_POSITION:
      {
	if (time > 0)
	  {
	    time --;
	    double usedStam = max(stamina, STAMINA_INC_MAX);
	    AdvSkills::TurtleCommand turtle;
	    turtle = skills->goToPosition(worldSnap, pos, 2.0, usedStam);
	    if (turtle.state == AdvSkills::TFINISHED)
	      {
		PLANTRACE( " Do not execute planned RunToPosition, because destination is reached ");
		setNothing();
		return false;
	      }
	    PLANTRACE( " Execute planned RunToPosition to " << pos << " with " 
		       << usedStam << " stamina, " 
		       << " my.pos = " << worldSnap->my.pos
		       << ", time = " << time);
	    skills->exec(worldSnap, turtle);
	    // we want to look at the ball
	    AdvSkills::TurnHeadCommand thcmd = skills->turnHeadToBall(worldSnap, turtle);
	    if ((thcmd.state == AdvSkills::THOK) || (thcmd.state == AdvSkills::THANGLE))
	      {      
		skills->exec(worldSnap, thcmd);
		DECISIONTRACE( "turning to ball" );
	      }
	    return true;
	  }
	PLANTRACE( "RunToPosition stopped because of time-out" );
	setNothing();
	return false;
    }      
    
    case PLAN_NO_ACTION:
      if (time > 0)
	{
	  skills->noAction();
	  time --;
	  PLANTRACE( " Executing planned noAction, time =  " << time );
	  if (time == 0)
	    setNothing();
	  return true;
	}
      PLANTRACE( " End of doing nothing " );
      return false;
 
    case PLAN_TURNING:
      if (time > 0)
	{
	  skills->exec(worldSnap, skills->turn(worldSnap, dir));
	  time --;
	  PLANTRACE( " Executing planned turn to " << dir << ", time = " << time );
	  if (time == 0)
	    setNothing();
	  return true;
	}
      PLANTRACE( " End of turning " );
      return false;

    case PLAN_NOTHING:
    default:
      PLANTRACE( " UNKNOWN PLAN !?! GRRRRRRRRR " );
      CMDTRACE( cerr << " UNKNOWN PLAN !?! GRRRRRRRRR " << endl );
      return false;
    }
}


// ------------------------------------------------------------------

StandardSituationModule::StandardSituationModule(bool isGoalie, 
					         const PassEval *_passEval, 
						 const GoalKickEval *_goalKickEval ) :
  goalKickEval(_goalKickEval),
  passEval(_passEval)

{
  // the agents modules
  if (goalKickEval == NULL)
    {
      CMDTRACE( cerr << "STANDARDSITUATIONMODULE: goalKickEval == NULL" << endl );
    }
  if (passEval == NULL)
    {
      CMDTRACE( cerr << "STANDARDSITUATIONMODULE: passEval == NULL" << endl );
    }

  readAgent_conf();


  goalieMovesInCurrentSituation = 0;
  goalieDoesFreeKick = false;
  goalieMayMove = false;
  goalieShouldMoveAgain = false;
  nextCycleDoNothing = false;
  weAreKickingOff = false;

  lastTime_PLAY_ON = -1;

  if (isGoalie)    
    CMDTRACE( cout << "StandardSituationModule (STDSIT) starting ... as goalie" 
	      << endl );
  else
    CMDTRACE( cout << "StandardSituationModule (STDSIT) starting ... as fieldplayer" 
	      << endl );
  iAmGoalie = isGoalie;
  isSpecialSituation = false;
  maxAge = 20;
  firstPlayer = secondPlayer = UNKNOWN;
  needSecond = true;
  waitingTime = -1;
  nextTurnAction = TURN_TO_LEFT_SIDE;

  sendSomething = false;
  sendInfo = new StandardSituationModuleInfo();

  // calculating the positions where the goalie should move when doing
  // a free kick after moving
  goalieMovePositions = GOALIE_MOVE_POSITIONS;
  if (isGoalie)
    {
      CMDTRACE(cout << "STDSIT: calculating Goalie Move positions)" << endl);
      double yStep = (PENALTY_AREA_WIDTH - 5.0) / (goalieMovePositions - 1.0);
      for (int i=0; i<goalieMovePositions; i++)
	{
	  goalieMovePos[i].setXY( -FIELD_LENGTH/2.0 + PENALTY_AREA_LENGTH - 1.0, 
				  -PENALTY_AREA_WIDTH / 2.0 + 2.5 + i * yStep  );
	  CMDTRACE(cout << "STDSIT: " << i << ": " << goalieMovePos[i] << endl);
	}
    }
}

StandardSituationModule::~StandardSituationModule()
{
  if (sendInfo)
    delete (sendInfo);
  CMDTRACE( cout << "StandardSituationModule leaving ..." << endl );  
}  


bool StandardSituationModule::feedbackWorld(const Snapshot* worldSnap, 
					    TacticalSnapshot* tacticalSnap)
{
  if (worldSnap == NULL)
    ERRORTRACE( "feedbackWorld: worldSnap == NULL" );
  if (tacticalSnap == NULL)
    ERRORTRACE( "feedbackWorld: tacticalSnap == NULL" );
  return false;
}

bool StandardSituationModule::feedbackAudio(AudioMsg &audioMsg)
{
  bool state;
  audioMsg >> state;
  if (state)
    {
      // audioMsg >> sendInfo;
      cerr <<  " Message received but not send" << endl;
    }
  return false;
}

Grade StandardSituationModule::evaluate(AdvSkills* skills,
					const Snapshot* worldSnap, 
					TacticalSnapshot* tacticalSnap)
{
  isSpecialSituation = false;
  Grade value;
  
  if (skills == NULL)
    {
      ERRORTRACE( "eval: skills == NULL" );
      value.set_IMPOSSIBLE();
      return value;
    }
  if (worldSnap == NULL)
    {
      ERRORTRACE( "eval: worldSnap == NULL" );
      value.set_IMPOSSIBLE();
      return value;
    }
  if (tacticalSnap == NULL)
    {
      ERRORTRACE( "eval: tacticalSnap == NULL" );
      value.set_IMPOSSIBLE();
      return value;
    }

  // for drawing free-kick-circles
  switch (worldSnap->playMode)
    {
    case PLAY_MODE_FREE_KICK_THEM:
    case PLAY_MODE_FREE_KICK_US:
    case PLAY_MODE_CORNER_KICK_US:
    case PLAY_MODE_CORNER_KICK_THEM:
    case PLAY_MODE_KICK_IN_US:
    case PLAY_MODE_KICK_IN_THEM:
      DRAWTRACE( CIRCLE(worldSnap->ball.pos, FREE_KICK_FORBIDDEN_AREA) );
    default: ;
    }
   
  // the evaluation
  switch (worldSnap->playMode)
    {
      // case PLAY_MODE_UNKNOWN:
    case PLAY_MODE_TIME_OVER:
      value.set_IMPOSSIBLE();
      break;
    case PLAY_MODE_TIME_EXTENDED:
      value.set_ABSOLUTE();
      weAreKickingOff = (worldSnap->ourSide == SIDE_LEFT);
      if (weAreKickingOff)
	DECISIONTRACE( "Extended time, our kick off" );
      else
	DECISIONTRACE( "Extended time, other kick off" );
      break;
    case PLAY_MODE_KICK_IN_US :
      {
	calcPlayersAndPos4KickInUs(worldSnap, tacticalSnap, firstPlayer, secondPlayer, 
				   needSecond, targetPos1, targetPos2);
	if (worldSnap->my.number == firstPlayer ||
	    (worldSnap->my.number == secondPlayer && needSecond) )
	  {
	    value.set_ABSOLUTE();
	    // max time to wait for second player to arrive its position
	    if (waitingTime < 0)
	      waitingTime = standardWaitingTime;
	  }
	else
	  {
	    if (mustAvoidBall(worldSnap))
	      value.set_ABSOLUTE();
	    else
	      value.set_BAD();
	  }
	break;
      }
    case PLAY_MODE_KICK_IN_THEM :
      calcPlayersAndPos4KickInThem(worldSnap, tacticalSnap, firstPlayer, secondPlayer,
				   needSecond, targetPos1, targetPos2);
      if (worldSnap->my.number == firstPlayer ||
	  (worldSnap->my.number == secondPlayer && needSecond) )
	value.set_ABSOLUTE();
      else
	value.set_BAD();
      break;   
    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_FREE_KICK_US :
      value = evalFreeKickUs(skills, worldSnap, tacticalSnap);
      break;
    case PLAY_MODE_OFFSIDE_US :
    case PLAY_MODE_FREE_KICK_THEM :
      {
	firstPlayer = UNKNOWN;
	secondPlayer = UNKNOWN;
	// Do we need more than one player to build a wall
	needSecond = true;
	if (!iAmGoalie)
	  calcPlayersAndPos4FreeKickThem(worldSnap, tacticalSnap, firstPlayer, secondPlayer, 
					 needSecond, targetPos1, targetPos2);
	if ((iAmGoalie &&
	     sqrnorm(worldSnap->ourGoal.pos - worldSnap->ball.pos) <= sqr(maxWallDist)) ||
	    worldSnap->my.number == firstPlayer || 
	    ((worldSnap->my.number == secondPlayer) && needSecond))
	  {
	    // if the opponent goalie will perform a free kick in his penalty area
	    // do not build a wall (because he will move!)
	    if (worldSnap->ball.inOtherPenaltyArea() &&
		tacticalSnap->firstDangerousOpponent == worldSnap->otherGoalieNo)
	      value.set_BAD();
	    else
	      value.set_ABSOLUTE();
	  }
	else
	  value.set_BAD();
      }
    break;
    case PLAY_MODE_GOAL_US :
      weAreKickingOff = false;
      value.set_ABSOLUTE();
      DECISIONTRACE( "Oh we scored a goal (through the other goalies trouserholders)" );
      break;
    case PLAY_MODE_GOAL_THEM :
      weAreKickingOff = true;
      value.set_ABSOLUTE();
      DECISIONTRACE( "Oh my god, the other team scored a goal" );
      break;      
    case PLAY_MODE_KICK_OFF_US :
      value.set_ABSOLUTE();
      weAreKickingOff = true;
      break;
    case PLAY_MODE_KICK_OFF_THEM :
      weAreKickingOff = false;
      value.set_ABSOLUTE();
      break;
    case PLAY_MODE_BEFORE_KICK_OFF :
      if  ( ( (worldSnap->ourSide == SIDE_LEFT) && 
	      (worldSnap->time == 0) ) ||
	    ( (worldSnap->ourSide == SIDE_RIGHT) &&  
	      (worldSnap->time == HALF_TIME_IN_CYCLES) )  ||
	    ( (worldSnap->ourSide == SIDE_LEFT) && 
	      (worldSnap->time == 2*HALF_TIME_IN_CYCLES) ) 
	    )
	{
	  // start of one of the half-times or the extended time
	  weAreKickingOff = true;
	}
      if  ( (worldSnap->time == 0)  ||
	    (worldSnap->time == HALF_TIME_IN_CYCLES ) ||
	    (worldSnap->time == 2*HALF_TIME_IN_CYCLES) 
	    )
	lastTime_PLAY_ON = worldSnap->time-1;  // the goalie should not catch if a new 
                                               // half-time just started
      
      value.set_ABSOLUTE();
      if (weAreKickingOff)
	DECISIONTRACE( "Check your trikots, we are before our kick-off" );
      else
	DECISIONTRACE( "Check your trikots, we are before their kick-off" );
      break;
    case PLAY_MODE_CORNER_KICK_US :
      {
	firstPlayer = UNKNOWN;
	secondPlayer = UNKNOWN;
	// Do we need more than one player for the freekick
	needSecond = false;
	calcPlayersAndPos4CornerKickUs(worldSnap, tacticalSnap, firstPlayer, secondPlayer, 
				       needSecond, targetPos1, targetPos2);
	if (worldSnap->my.number == firstPlayer ||
	    (worldSnap->my.number == secondPlayer && needSecond))
	  {
	    value.set_ABSOLUTE();
	    // max time to wait for second player to arrive its position
	    if (waitingTime < 0)
	      waitingTime = standardWaitingTime / 2;
	  }
	else
	  {
	    if (mustAvoidBall(worldSnap))
	      value.set_ABSOLUTE();
	    else
	      value.set_BAD();
	  }
	break;
      }
    case PLAY_MODE_CORNER_KICK_THEM :
      calcPlayersAndPos4CornerKickThem(worldSnap, tacticalSnap, firstPlayer, secondPlayer,
				       needSecond, targetPos1, targetPos2);
      if (worldSnap->my.number == firstPlayer ||
	  (worldSnap->my.number == secondPlayer && needSecond) )
	value.set_ABSOLUTE();
      else
	value.set_BAD();
      break;
    case PLAY_MODE_GOAL_KICK_THEM :
      value.set_IMPOSSIBLE();
      break;
    case PLAY_MODE_GOAL_KICK_US :
      if (iAmGoalie)
	{
	  if (nextCycleDoNothing)
	    {
	      value.set_BAD();
	      break;
	    }
	  if (waitingTime < 0)
	    waitingTime = standardWaitingTime;
	  value.set_ABSOLUTE();
	  goalieMayMove = false;
	  goalieShouldMoveAgain = false;
	}
      else
	{
	  if (mustAvoidBall(worldSnap))
	    value.set_ABSOLUTE();
	  else
	    value.set_BAD();
	}
      break;
    case PLAY_MODE_HALF_TIME :
      value.set_ABSOLUTE();
      weAreKickingOff = (worldSnap->ourSide == SIDE_RIGHT);
      if (weAreKickingOff)
	DECISIONTRACE( "Half-Time, our kick-off");
      else
	DECISIONTRACE( "Half-Time, other kick-off");
      break;
    case PLAY_MODE_PLAY_ON:
      waitingTime = -1;
      weAreKickingOff = false;
      goalieMovesInCurrentSituation = 0;
      // looking for standardsituations that have been in the last cycles
      lastTime_PLAY_ON = worldSnap->time;
      if (myPlan.planExists())
	value.set_ABSOLUTE();
      else
	value.set_IMPOSSIBLE();
      break;
    case PLAY_MODE_DROP_BALL :
      goalieMovesInCurrentSituation = 0;
      value.set_IMPOSSIBLE();
      break;
    default:
      value.set_IMPOSSIBLE();
      ERRORTRACE( "Playmode Unkown !?" );
      break;
    } 
  DECISIONTRACE( "evaluate returning " << value << ", playmode = " << worldSnap->playMode );

  return value;
}


bool StandardSituationModule::act(AdvSkills* skills,
				  const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap)
{
  if (skills == NULL)
    {
      ERRORTRACE( "act: skills == NULL" );
      return false;
    }
  if (worldSnap == NULL)
    {
      ERRORTRACE( "act: wordlSnap == NULL" );
      skills->noAction();
      return false;
    }
  if (tacticalSnap == NULL)
    {
      ERRORTRACE( "act: tacticalSnap == NULL" );
      skills->noAction();
      return false;
    }

  // first ensure that we look in the standard mode
  skills->changeView(worldSnap, VIEW_WIDTH_NORMAL, VIEW_QUALITY_HIGH);

  // do we have to speak because we see the ball in a new standardsituation
  // where the ball might have been moved 

  if ( (worldSnap->ball.age < worldSnap->time - lastTime_PLAY_ON) &&
       (worldSnap->time - lastTime_PLAY_ON < 9)
     )

    isSpecialSituation = true;
  else
    isSpecialSituation = false;

  switch (worldSnap->playMode)
    {
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_GOAL_THEM :
      return doingKickOff(skills, worldSnap, tacticalSnap);
    case PLAY_MODE_KICK_IN_US :
      {
	if (worldSnap->my.number == firstPlayer ||
	    (worldSnap->my.number == secondPlayer && needSecond))
	  return doingKickInUs(skills, worldSnap,
			       firstPlayer, secondPlayer, needSecond,
			       targetPos1, targetPos2);
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing avoidBall" );
	    avoidBall(skills, worldSnap, tacticalSnap);
	    return false;
	  }
	break;
      }
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_KICK_IN_THEM :
	if (worldSnap->my.number == firstPlayer ||
	    (worldSnap->my.number == secondPlayer && needSecond))
	  return doingKickInCornerKickThem(skills, worldSnap,
					   firstPlayer, secondPlayer, 
					   needSecond, targetPos1, targetPos2);
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing avoidBall" );
	    avoidBall(skills, worldSnap, tacticalSnap);
	    return false;
	  }
	break;
    case PLAY_MODE_OFFSIDE_THEM :
    case PLAY_MODE_FREE_KICK_US :
      {
	// if the goalie does it all ...
	if ( goalieDoesFreeKick )
	  return doingGoalieFreeKickUs(skills, worldSnap, tacticalSnap);
	    
	if (worldSnap->my.number == firstPlayer ||
	    (worldSnap->my.number == secondPlayer && needSecond))
	  // if we are close to the edge of the field
	  if (fabs(fabs(worldSnap->ball.pos.getY())-FIELD_WIDTH/2) < 2)
	    // do a kick in
	    return doingKickInUs(skills, worldSnap,
				 firstPlayer, secondPlayer, needSecond,
				 targetPos1, targetPos2);
	  else
	    // do a free kick
	    return doingFreeKickUs(skills, worldSnap,
				   firstPlayer, secondPlayer, needSecond,
				   targetPos1, targetPos2);
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing avoidBall" );
	    avoidBall(skills, worldSnap, tacticalSnap);
	    return false;
	  }
      break;
      }
    case PLAY_MODE_OFFSIDE_US :    
    case PLAY_MODE_FREE_KICK_THEM :
      {
	if (iAmGoalie && 
	    sqrnorm(worldSnap->ourGoal.pos-
		    worldSnap->ball.pos) <= sqr(maxWallDist))
	  return doingGoalieFreeKickThem(skills, worldSnap);
	
	if (((worldSnap->my.number == firstPlayer || 
	      ((worldSnap->my.number == secondPlayer) && needSecond))))
	  {
	    DECISIONTRACE( " Doing fieldplayer freekick them" );
	    return doingFieldPlayerFreeKickThem(skills, worldSnap, 
						firstPlayer, secondPlayer, needSecond,
						targetPos1, targetPos2);
	  }
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing avoidBall" );
	    avoidBall(skills, worldSnap, tacticalSnap);
	    return false;
	  }
	break;
      }
    case PLAY_MODE_CORNER_KICK_US :
      if (worldSnap->my.number == firstPlayer ||
	  (worldSnap->my.number == secondPlayer && needSecond))
	return doingCornerKickUs(skills, worldSnap, 
				 firstPlayer, secondPlayer, needSecond,
				 targetPos1, targetPos2);
      else
	{
	  DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			 " situation, doing avoidBall" );
	  avoidBall(skills, worldSnap, tacticalSnap);
	  return false;
	}
      break;
    case PLAY_MODE_GOAL_KICK_US :
      return doingGoalieFreeKickUs(skills, worldSnap, tacticalSnap);
//     case PLAY_MODE_GOAL_KICK_THEM :

    case PLAY_MODE_HALF_TIME :
      {
	skills->noAction();
	DECISIONTRACE( " half-time -> act does not do a thing" );
	return true;
      }
    case PLAY_MODE_PLAY_ON:
      {
	if (myPlan.planExists())
	  {
	    if (!myPlan.executePlan(skills, worldSnap))
	      {
		DECISIONTRACE ( " act: plan failed => doing nothing ");
		skills->noAction();
		return false;
	      }
	    return true;
	  }
	else
	  {
	    DECISIONTRACE( "I am not responsible in " << worldSnap->playMode << 
			   " situation, doing nothing" );
	    skills->noAction();
	    return false;
	  }	
      }
    break;
    default:
      skills->noAction();
      DECISIONTRACE( "act does not do a thing" );
      return false;
    }
  return false;
}



bool StandardSituationModule::getSayMsg(AudioMsg &audioMsg)
{
  if (sendSomething) 
    {
      audioMsg.markWrite();      
      audioMsg << true;      
      audioMsg << sendInfo;
      
      cerr << " Sending a message " << endl;

      // check if error occurred
      if( !audioMsg.getState() )
	{
	  cerr << "STDSIT: Stream overflow: " << audioMsg << endl;
	  audioMsg.rollbackWrite().clearState();
	  
	  audioMsg << false;	  
	}
    } 
  else 
    {
      audioMsg << false;
    }    
  return true;
}



void StandardSituationModule::pushPlayer(WorldPlayer &player, 
					 Vector2d pos)
{
  CMDTRACE(cout << ".");
  if ( (player.age > maxAge)  // too old players are ignored 
       || ( player.distTo(pos) > FREE_KICK_FORBIDDEN_AREA) )
    return;
  
  // player has to be pushed outside the circle and his speed has to be set to 0.
  player.v.setXY(0.0, 0.0);
  Vector2d distToPos = player.pos;
  distToPos -= pos;
  double length = distToPos.getLength();
  if (length < 0.1) 
    {
      // player is very close, so place him in
      // front of me
      player.pos.setXY(pos.getX(), 
			pos.getY()-FREE_KICK_FORBIDDEN_AREA);
      CMDTRACE(cout << "s");
    }
  else
    {
      // player is not so close, so pushing him the server-like way
      distToPos.stretch(FREE_KICK_FORBIDDEN_AREA / length);
      player.pos = pos;
      player.pos += distToPos;
      CMDTRACE(cout << "p"); 
    }
}


/** put information to AudioMsg-Stream */
class AudioMsg &operator<< (AudioMsg &aumsg, const StandardSituationModuleInfo &)
{
  // write all data from Info class here, and make sure there are 
  // AudioMsg-operators defined for the specific type of data you write !
  
  // aumsg << info.value1;
  // aumsg << info.value2;

  return aumsg;
}

/** fill with information from AudioMsg-Stream */
class AudioMsg &operator>> (AudioMsg &aumsg, StandardSituationModuleInfo &)
{

  // read data for Info class in same order as writing it

  // aumsg >> info.value1;
  // aumsg >> info.value2;

  return aumsg;

}



Vector2d StandardSituationModule::calcKickOffTarget(int playerNumber,
						    const Snapshot *worldSnap,
						    TacticalSnapshot *tacticalSnap)
{
  if (worldSnap);     // to avoid warnings

  int tacticalNumber 
    = tacticalSnap->ourTeamTacticalNo(playerNumber);
  if (tacticalNumber == UNKNOWN)
    tacticalNumber = playerNumber;

  int ourSystem = getOurTacticalSystem(tacticalSnap);

  switch(ourSystem)
    {
    case 442:
      if (weAreKickingOff)
	return kickOffPosWeKickOff_442[tacticalNumber-1];
      else
	return kickOffPosTheyKickOff_442[tacticalNumber-1];
    case 343:
      if (weAreKickingOff)
	return kickOffPosWeKickOff_343[tacticalNumber-1];
      else
	return kickOffPosTheyKickOff_343[tacticalNumber-1];
    case 352:
      if (weAreKickingOff)
	return kickOffPosWeKickOff_352[tacticalNumber-1];
      else
	return kickOffPosTheyKickOff_352[tacticalNumber-1];
    default: // 343
      ERRORTRACE( "unknown tactical system: " << ourSystem );
      if (weAreKickingOff)
	return kickOffPosWeKickOff_343[tacticalNumber-1];
      else
	return kickOffPosTheyKickOff_343[tacticalNumber-1];
      
    }
}




void StandardSituationModule::calcGoalieMoveAndKick
(
 const Snapshot *worldSnap, 
 TacticalSnapshot *tacticalSnap,
 Vector2d &moveTarget,
 Vector2d &kickTarget,
 bool goalieMayMoveOnceMore)
{
  if (moveTarget == kickTarget); // to avoid warnings
  if (tacticalSnap);             // to avoid warnings

  kickTarget.setXY(0.0, 0.0); // in the case that something goes wrong ...
  // the grades for all movePositions and the best found grade

  int bestPassPartner = UNKNOWN;

  // advanced ourEval:
  PropOf props;
  props.perhaps = false;
  props.grade.set_BAD();
  PropOf actProps;


  if (!goalieMayMoveOnceMore)
    {
      DECISIONTRACE( "calcGoalieMoveAndKick: goalie must not move" );
      // goalie must not move
      for (int passPartner=1; passPartner<=11; passPartner++)
	{
	  if (passPartner == worldSnap->ourGoalieNo)
	    continue;   // goalie does not want the ball
	  if (worldSnap->ourTeam(passPartner).age > maxAge)
	    continue;   // old players are no good !!
	  actProps = passEval->ourEvalProp(worldSnap, worldSnap->ourGoalieNo, passPartner);
	  if (passEval->betterPass(worldSnap, actProps, props))
	    {
	      props           = actProps;
	      bestPassPartner = props.number;
	      kickTarget      = props.pos;
	    }
	} // passPartner      
      DECISIONTRACE( "bestPassPartner = " << bestPassPartner << ", grade = " 
		     << props.grade);
      return;
    }
  
  // goalie can do one more move-command

  Grade positionGrades[goalieMovePositions];
  int bestMovePos     = 0;

  Snapshot simSnap = *worldSnap;

  WorldPlayer ourPlayer;
  WorldPlayer otherPlayer;  

  for (int posNumber=0; posNumber<goalieMovePositions; posNumber++)
    {
      // copy positions of all players
      PUSHTRACE( cout << "the pushings while goalieMove to " 
		 << goalieMovePos[posNumber] << endl );
      for (int playerNumber=1; playerNumber<=11; playerNumber++)
	{
	  // pushing our player
	  ourPlayer   = worldSnap->ourTeam(playerNumber);
	  PUSHTRACE(cout << worldSnap->time << " STDSIT[" << worldSnap->my.number << "]: " 
		   << "ourTeam[" << playerNumber << "].pos = " << ourPlayer.pos
		   << " -> " );
	  pushPlayer(ourPlayer, goalieMovePos[posNumber]);
	  simSnap.setOurTeam(playerNumber, ourPlayer); 
	  PUSHTRACE(cout << simSnap.ourTeam(playerNumber).pos << endl);
	  
	  // pushing other player
	  otherPlayer = worldSnap->otherTeam(playerNumber); 
	  PUSHTRACE(cout << worldSnap->time << " STDSIT[" << worldSnap->my.number << "]: " 
		   << "otherTeam[" << playerNumber << "].pos = " << otherPlayer.pos
		   << " -> " );
	  pushPlayer(ourPlayer, goalieMovePos[posNumber]);
	  simSnap.setOtherTeam(playerNumber, otherPlayer);
	  PUSHTRACE(cout << simSnap.otherTeam(playerNumber).pos << endl);
	} // playerNumber

      // moving our goalie
      ourPlayer = simSnap.ourTeam(simSnap.ourGoalieNo);
      ourPlayer.pos = goalieMovePos[posNumber];
      simSnap.setOurTeam(simSnap.ourGoalieNo, ourPlayer);
      simSnap.my = simSnap.ourTeam(simSnap.ourGoalieNo);
      // for debugging pushing-algorithm
      PUSHTRACE(
		if (posNumber == 1)
		{
		  double sizeFactor = 2;
		  for (int playerNumber=1; playerNumber <= 11; playerNumber++)
		    {
		      if (simSnap.ourTeam(playerNumber).age < maxAge)
			DRAWTRACE( CIRCLE(simSnap.ourTeam(playerNumber).pos, 
					  sizeFactor*PLAYER_SIZE ) );
		      if (simSnap.otherTeam(playerNumber).age < maxAge)
			DRAWTRACE( LCIRCLE(simSnap.otherTeam(playerNumber).pos, 
					   sizeFactor*PLAYER_SIZE ) );
		    }
		  DRAWTRACE( GCIRCLE(simSnap.ball.pos, sizeFactor*BALL_SIZE) );
		  DRAWTRACE( LGCIRCLE(simSnap.my.pos, sizeFactor*PLAYER_SIZE) );
		}		
		);
      
      // look for the best pass
      positionGrades[posNumber].set_IMPOSSIBLE();
      for (int passPartner=1; passPartner<=11; passPartner++)
	{
	  if (passPartner == simSnap.ourGoalieNo)
	    continue;   // goalie does not want the ball
	  if (simSnap.ourTeam(passPartner).age > maxAge)
	    continue;   // old players are no good !!
	  
	  actProps = passEval->ourEvalProp(&simSnap, worldSnap->ourGoalieNo, passPartner);
	  if (passEval->betterPass(&simSnap, actProps, props))
	    {
	      props           = actProps;
	      bestPassPartner = props.number;
	      kickTarget      = props.pos;
	      bestMovePos     = posNumber;
	    }
	  if (props.grade.is_better_than(positionGrades[posNumber]))
	    positionGrades[posNumber] = props.grade;
	} // passPartner      
      DECISIONTRACE( "moving to " << goalieMovePos[posNumber] 
		     << " => grade=" << positionGrades[posNumber]
		     << ", bestGrade=" << props.grade
		     );
    } // posNumber
  
  
  // now we move to goalieMovePos[bestMovePos] and think of passing to passPartner
  moveTarget = goalieMovePos[bestMovePos];
  DECISIONTRACE( "goalie has done kick-off decision: move to " << moveTarget << " ..." );
  DECISIONTRACE( " ... perhaps trying to kick to player " 
		 << bestPassPartner << " at " << kickTarget
		 << " grade= " << props.grade );
  
  if (props.grade.is_BAD())
    {
      DECISIONTRACE( "best pass is bad, so think about moving ..." );
      moveTarget = calcEmergencyMovePosition(worldSnap);
      DECISIONTRACE( "move to " << moveTarget );
    }
}


void StandardSituationModule::calcPos4CornerKickUs(const Snapshot *worldSnap, 
						   Vector2d &targetPos1,
						   Vector2d &targetPos2) 
{
  // Distance between first and second Player
  double DIST = 11.;
  // the first position is directly behind the ball on the line
  // form ball to (0,0)
  targetPos1 =  -worldSnap->ball.pos;
  targetPos1 *= -1.1*(BALL_SIZE+PLAYER_SIZE)/norm(targetPos1);
  // 1.1 as the circle I will not walk into has a radius of
  // BALL_SIZE+PLAYER_SIZE
  targetPos1 += worldSnap->ball.pos;
  // the second position is on the same line
  targetPos2 =  -worldSnap->ball.pos;
  targetPos2 *= DIST/norm(targetPos2);
  targetPos2 += worldSnap->ball.pos;
}

// read data from the agent's config file (agent.conf)
void StandardSituationModule::readAgent_conf( void )
{
  goalieStaminaFactor 
    = Config::getValueAsDouble("std_goalieStaminaFactor");
  wallStaminaFactor 
    = Config::getValueAsDouble("std_wallStaminaFactor");
  passiveStaminaFactor 
    = Config::getValueAsDouble("std_passiveStaminaFactor");
  activeStaminaFactor 
    = Config::getValueAsDouble("std_activeStaminaFactor");
  standardWaitingTime 
    = Config::getValueAsInt("std_maxWaitingTime");
  waitingTimeAfterGoalieMove
    = Config::getValueAsInt("std_waitingTimeAfterGoalieMove");
  lookAroundAngle
    = Config::getValueAsDouble("std_lookAroundAngle");
  maxSecondDist
    = Config::getValueAsDouble("std_maxSecondDist");
  minWallDist
    = Config::getValueAsDouble("std_minWallDist");
  maxWallDist
    = Config::getValueAsDouble("std_maxWallDist");
  ballAvoidDist
    = Config::getValueAsDouble("std_ballAvoidDist");

  char valueString[100];

  // reading start positions for the different tactical systems
  for (int i=0; i<11; i++)
    {
      // 4-4-2
      sprintf(valueString, "std_442_KickOffPosWeKickOff_%d", i+1);
      kickOffPosWeKickOff_442[i] 
	= Config::getValueAsVector(valueString);
      sprintf(valueString, "std_442_KickOffPosTheyKickOff_%d", i+1);
      kickOffPosTheyKickOff_442[i] 
	= Config::getValueAsVector(valueString);

      // 3-4-3
      sprintf(valueString, "std_343_KickOffPosWeKickOff_%d", i+1);
      kickOffPosWeKickOff_343[i] 
	= Config::getValueAsVector(valueString);
      sprintf(valueString, "std_343_KickOffPosTheyKickOff_%d", i+1);
      kickOffPosTheyKickOff_343[i] 
	= Config::getValueAsVector(valueString);

      // 3-5-2
      sprintf(valueString, "std_352_KickOffPosWeKickOff_%d", i+1);
      kickOffPosWeKickOff_352[i] 
	= Config::getValueAsVector(valueString);
      sprintf(valueString, "std_352_KickOffPosTheyKickOff_%d", i+1);
      kickOffPosTheyKickOff_352[i] 
	= Config::getValueAsVector(valueString);
    }

  // reading player types and count for stdsits
  for (int t = 0; t < 2; t++)
    {
      const char *str1;
      switch (t)
	{
	case 0:
	  str1 = "our";
	  break;
	case 1:
	  str1 = "other";
	  break;
	default:
	  str1 = "looser";
	}
      for (int fp = 0; fp < 3; fp++)
	{
	  const char *str2;
	  switch (fp)
	    {
	    case 0:
	      str2 = "leftField";  
	      break;
	    case 1:
	      str2 = "middleField";
	      break;
	    case 2:
	      str2 = "rightField";
	      break;
	    default:
	      str2 = "looser";
	    }
	  for (int ts = 0; ts < 3; ts++)
	    {
	      const char *str3;
	      switch (ts)
		{
		case 0:
		  str3 = "442";  
		  break;
		case 1:
		  str3 = "352";
		  break;
		case 2:
		  str3 = "343";
		  break;
		default:
		  str3 = "looser";
		}

	      sprintf(valueString, "std_%s_%s_%s_defenders", str3, str1, str2);
	      players[t][fp][ts].defenders = Config::getValueAsInt(valueString);

	      sprintf(valueString, "std_%s_%s_%s_midfield", str3, str1, str2);
	      players[t][fp][ts].midfield = Config::getValueAsInt(valueString);

	      sprintf(valueString, "std_%s_%s_%s_offenders", str3, str1, str2);
	      players[t][fp][ts].offenders = Config::getValueAsInt(valueString);
	    }
	}
    }
}

void StandardSituationModule::calcClosestDesiredPlayersTo(const Vector2d &target, 
							  const Snapshot *worldSnap,
							  const TacticalSnapshot *ts,
							  int &firstPlayer,
							  int &secondPlayer)
{
  int off = 0;
  int mid = 0;
  int def = 0;
  

  int defender[10], d = 0;
  int midfield[10], m = 0;
  int offender[10], o = 0;
  int team[11] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11};
  for (int i = 0; i < 10; i++)
    {
      defender[i] = UNKNOWN;
      midfield[i] = UNKNOWN;
      offender[i] = UNKNOWN;
    }

  // calc our defenders, midfielders and offenders
  for (int player = 1; player < 12; player++)
    {
      int tacNo = ts->ourTeamTacticalNo(player);
      if (tacNo != UNKNOWN)
	{
	  if (ts->ourTeamDefender(tacNo))
	    defender[d++] = player;
	  if (ts->ourTeamMidfield(tacNo))
	    midfield[m++] = player;
	  if (ts->ourTeamOffender(tacNo))
	    offender[o++] = player;
	}
    }  

  OTHERTRACE( " known defenders: " << d );
  CMDTRACE( cout << worldSnap->time << " "; 
	    for (int i = 0; i < d; i++) { cout << defender[i] << " "; } cout << endl;);
  OTHERTRACE( " known midfield: " << m );
  CMDTRACE( cout << worldSnap->time << " "; 
	    for (int i = 0; i < m; i++) { cout << midfield[i] << " "; } cout << endl;);
  OTHERTRACE( " known offenders: " << o );
  CMDTRACE( cout << worldSnap->time << " "; 
	    for (int i = 0; i < o; i++) { cout << offender[i] << " "; } cout << endl;);

  calcDesiredPlayerTypes(worldSnap, ts, target, def, mid, off);

  if (def+mid+off > 2)
    {
      ERRORTRACE( " too many players desired for stdsit, taking two middfielders " );
      def = 0;
      mid = 2;
      off = 0;
    }

  if (def+mid+off  <= 1)
    {
      ERRORTRACE( " not enough players desired for stdsit, taking two middfielders " );
      def = 0;
      mid = 2;
      off = 0;
    }
  
  // if we want one more defender than we know
  if (def-d == 1)
    {
      DECISIONTRACE( " not enough defenders known, taking a midfielder " );
      def--;
      mid++;
    }
  
  // remember: 0 <= def, mid, off <= 2
  // if we want two defenders but we know none 
  if (def-d == 2)
    {
      DECISIONTRACE( " no defenders known, taking two midfielders " );
      def = 0;
      mid = 2;
      off = 0;
    }

  // if we want one more offender than we know
  if (off-o == 1)
    {
      DECISIONTRACE( " not enough offenders known, taking a midfielder " );
      off--;
      mid++;
    }
  
  // remember: 0 <= def, mid, off <= 2
  // if we want two offenders but we know none 
  if (off-o == 2)
    {
      DECISIONTRACE( " no offenders known, taking two midfielders " );
      def = 0;
      mid = 2;
      off = 0;
    }

  // if we want one more midfielder than we know
  if (mid-m == 1)
    {
      DECISIONTRACE( " not enough middfielders known " );
      mid--;
      if (d > def && target.getX() <= 0)
	{
	  DECISIONTRACE( " taking a defender " );
	  def++;
	}
      else if (o > off && target.getX() > 0)
	{
	  DECISIONTRACE( " taking an offender " );
	  off++;
	}
    }

  // if we want two midfielders but we know none 
  if (mid-m == 2)
    {
      DECISIONTRACE( " no middfielders known " );
      mid = 0;
      if (d > def && target.getX() <= 0)
	{
	  DECISIONTRACE( " taking a defender " );
	  def++;
	}
      else if (o > off && target.getX() > 0)
	{
	  DECISIONTRACE( " taking an offender " );
	  off++;
	}
    }

  
  // calc the first and second player
  firstPlayer = secondPlayer = UNKNOWN;
  if (def == 1)
    {
      calcClosestPlayerTo(target, defender, d, worldSnap, firstPlayer);
      // make sure this player is not calculated as closest player
      // if we do not know enough players
      team[firstPlayer-1] = UNKNOWN;
      OTHERTRACE( " closest defender: " << firstPlayer );
    }

  int dummyPlayer = UNKNOWN;
  if (mid == 1)
    {
      calcClosestPlayerTo(target, midfield, m, worldSnap, dummyPlayer);
      team[dummyPlayer-1] = UNKNOWN;
      OTHERTRACE( " closest midfielder: " << dummyPlayer );
    }
  if (firstPlayer == UNKNOWN)
    firstPlayer = dummyPlayer;
  else
    secondPlayer = dummyPlayer;

  if (off == 1)
    {
      calcClosestPlayerTo(target, offender, o, worldSnap, dummyPlayer);
      team[dummyPlayer-1] = UNKNOWN;
      OTHERTRACE( " closest offender: " << dummyPlayer );
    }
  if (firstPlayer == UNKNOWN)
    firstPlayer = dummyPlayer;
  else
    secondPlayer = dummyPlayer;

  if (def == 2)
    {
      calcClosestTwoPlayersTo(target, defender, d, worldSnap, firstPlayer, secondPlayer);
      OTHERTRACE( " closest defenders: " << firstPlayer << " " << secondPlayer );
    }
  if (mid == 2)
    {
      calcClosestTwoPlayersTo(target, midfield, m, worldSnap, firstPlayer, secondPlayer);
      OTHERTRACE( " closest midfielders: " << firstPlayer << " " << secondPlayer );
    }
  if (off == 2)
    {
      calcClosestTwoPlayersTo(target, offender, o, worldSnap, firstPlayer, secondPlayer);
      OTHERTRACE( " closest offenders: " << firstPlayer << " " << secondPlayer );
    }

  // if we know only one player calc the closest wich is not the first one
  if (def+mid+off == 1)
    {
      calcClosestPlayerTo(target, team, 11, worldSnap, secondPlayer);
      OTHERTRACE( " closest player: " << secondPlayer );
    }

  // if we don't know a player calc the closest two
  if (def+mid+off == 0)
    {
      calcClosestTwoPlayersTo(target, team, 11, worldSnap, firstPlayer, secondPlayer);
      OTHERTRACE( " closest players: " << firstPlayer << " secondPlayer: " << secondPlayer);
    }

  DECISIONTRACE( " firstPlayer: " << firstPlayer << " secondPlayer: " << secondPlayer );
}

void StandardSituationModule::calcDesiredPlayerTypes(const Snapshot *worldSnap, 
						     const TacticalSnapshot *ts,
						     const Vector2d &target, int &noDef, 
						     int &noMid, int &noOff)
{
  // calc the indices to access the array with the number of player types
  // for the numbers see readAgent_conf()
  int team = 0;
  int fieldPart = 0;
  int tacticalSystem = 0;

  switch (worldSnap->playMode)
    {
      // we have the ball
    case PLAY_MODE_FREE_KICK_US:
    case PLAY_MODE_KICK_IN_US:
    case PLAY_MODE_CORNER_KICK_US:
    case PLAY_MODE_OFFSIDE_THEM:
      team = 0;
      break;
    // they have the ball
    case PLAY_MODE_FREE_KICK_THEM:
    case PLAY_MODE_KICK_IN_THEM:
    case PLAY_MODE_CORNER_KICK_THEM:
    case PLAY_MODE_OFFSIDE_US:
      team = 1;
      break;
    default: ;
    }

  switch (getOurTacticalSystem(ts))
    {
    case 442:
      tacticalSystem = 0;
      break;
    case 352:
      tacticalSystem = 1;
      break;
    case 343:
      tacticalSystem = 2;
      break;
    default: ;
    }
  
  // if the target is in our third of the field
  if (target.getX() <= -FIELD_LENGTH/6.+2.)
    fieldPart = 0;

  // if the target is in the middle third of the field
  if (target.getX() <= FIELD_LENGTH/6. &&
      target.getX() > -FIELD_LENGTH/6.+2.)
    fieldPart = 1;

  // if the target is in their third of the field
  if (target.getX() >= FIELD_LENGTH/6.)
    fieldPart = 2;

  noDef = players[team][fieldPart][tacticalSystem].defenders;
  noMid = players[team][fieldPart][tacticalSystem].midfield;
  noOff = players[team][fieldPart][tacticalSystem].offenders;

  DECISIONTRACE( " Desired player types: defenders: " << noDef );
  DECISIONTRACE( "                       midefield: " << noMid );
  DECISIONTRACE( "                       offenders: " << noOff );
  
}

void StandardSituationModule::calcClosestPlayerTo(const Vector2d &target, const int team[], 
						  const int count, const Snapshot *ws, 
						  int &player)
{
  // calc distances
  double dists[11];
  
  for (int i = 0; i < 11; i++)
    dists[i] = 50000;

  for (int i = 0; i < count; i++)
    {
      if (team[i] == UNKNOWN)
	continue;
      dists[team[i]-1] = sqrnorm(ws->ourTeam(team[i]).pos-target);
    }
  // make sure the goalie is not the closest player
  if (ws->ourGoalieNo > 0)
    dists[ws->ourGoalieNo-1] = 60000;
  
  // find closest player
  player = UNKNOWN;
  double dist = 60000;
  for (int i = 0; i < 11; i++)
    {
      if (dists[i] < dist)
	{
	  dist = dists[i];
	  player = i+1;
	}
    }

  // find players close to target too
  dist = sqr(sqrt(dist)+1.5);
  int closePlayers[11], closeCount = 0;
  
  for (int i = 0; i < 11; i++)
    {
      if (dists[i] < dist)
	closePlayers[closeCount++] = i+1;
    }

  // the closest player is the close player with 
  // smallest player number or the closest
  if (closeCount > 0)
    player = closePlayers[0];
}


void StandardSituationModule::calcClosestTwoPlayersTo(const Vector2d &target, 
						      const int team[], 
						      const int count, const Snapshot *ws, 
						      int &player1, int &player2)
{
  calcClosestPlayerTo(target, team, count, ws, player1);
  int *myTeam = new int[count];
  for (int i = 0; i < count; i++)
    {
      if (team[i] == player1)
	myTeam[i] = UNKNOWN;
      else
	myTeam[i] = team[i];
    }
  calcClosestPlayerTo(target, myTeam, count, ws, player2);
  delete myTeam;
}


void StandardSituationModule::calcPlayersAndPos4CornerKickThem(const Snapshot *worldSnap, 
							       const TacticalSnapshot *ts,
							       int &firstPlayer,
							       int &secondPlayer,
							       bool &needSecond,
							       Vector2d &targetPos1,
							       Vector2d &targetPos2) 
{

  firstPlayer = secondPlayer = UNKNOWN;
  needSecond = true;
  double vz = worldSnap->ball.pos.getY()/
    fabs(worldSnap->ball.pos.getY());
  if (vz > 0)
    {
      targetPos1.setRadAng(FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE,
			   DEG_TO_RAD(-45.));
      targetPos2.setRadAng(FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE,
			   DEG_TO_RAD(-22.5));
    }
  else
    {
      targetPos1.setRadAng(FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE,
			   DEG_TO_RAD(45.));
      targetPos2.setRadAng(FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE,
			   DEG_TO_RAD(22.5));
    }
  targetPos1 += worldSnap->ball.pos;
  targetPos2 += worldSnap->ball.pos;
  DRAWTRACE( POSITION(targetPos1) );
  DRAWTRACE( POSITION(targetPos2) );
  calcClosestDesiredPlayersTo(worldSnap->ball.pos, worldSnap,
			      ts, firstPlayer, secondPlayer);

  if (secondPlayer != UNKNOWN)
    {
      calcShortWays(worldSnap, firstPlayer, secondPlayer,
		    targetPos1, targetPos2);
    }
  else
    {
      needSecond = false;
    }
  

  DECISIONTRACE( " firstPlayer: " << firstPlayer << " secondPlayer: " << secondPlayer );
  DECISIONTRACE( " needSecond: " << needSecond );
  OTHERTRACE( " targetPos1: " << targetPos1 << " targetPos2: " << targetPos2 );
}

void StandardSituationModule::calcPlayersAndPos4KickInThem(const Snapshot *worldSnap, 
							   const TacticalSnapshot *ts,
							   int &firstPlayer,
							   int &secondPlayer,
							   bool &needSecond,
							   Vector2d &targetPos1,
							   Vector2d &targetPos2) 
{
  // if they are near to our goal treat a kick in as a corner kick
  if (worldSnap->ball.pos.getX() < -FIELD_LENGTH/2+GOAL_AREA_LENGTH)
    {
      calcPlayersAndPos4CornerKickThem(worldSnap, ts, firstPlayer, secondPlayer,
				       needSecond, targetPos1, targetPos2);
      return;
    }

  firstPlayer = secondPlayer = UNKNOWN;
  needSecond = true;
  // the first pos is between ball and our goal
  targetPos1 = worldSnap->ourGoal.pos-worldSnap->ball.pos;
  targetPos1 *= (FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE)/norm(targetPos1);
  targetPos1 += worldSnap->ball.pos;

  double vz = worldSnap->ball.pos.getY()/
    fabs(worldSnap->ball.pos.getY());
  if (worldSnap->otherOffsideLine(maxAge) > targetPos1.getX())
    {
      // try to block their way back
      if (vz < 0)
	targetPos1.setRadAng(FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE,
			     DEG_TO_RAD(45.));
      else
	targetPos1.setRadAng(FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE,
			     DEG_TO_RAD(-45.));
      targetPos1 += worldSnap->ball.pos;
    }
  // the second pos is directly above (below) the ball
  targetPos2.setRadAng(FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE,
		       DEG_TO_RAD(90));
  targetPos2 *= -vz;
  targetPos2 += worldSnap->ball.pos;
  DRAWTRACE( POSITION(targetPos1) );
  DRAWTRACE( POSITION(targetPos2) );
  calcClosestDesiredPlayersTo(worldSnap->ball.pos, worldSnap, ts, 
			      firstPlayer, secondPlayer);
  
  if (sqrnorm(targetPos1-targetPos2) <= sqr(4.))
    secondPlayer = UNKNOWN;

  if (secondPlayer != UNKNOWN)
    {
      calcShortWays(worldSnap, firstPlayer, secondPlayer,
		    targetPos1, targetPos2);
    }
  else
    {
      needSecond = false;
    }

  DECISIONTRACE( " firstPlayer: " << firstPlayer << " secondPlayer: " << secondPlayer );
  DECISIONTRACE( " needSecond: " << needSecond );
  OTHERTRACE( " targetPos1: " << targetPos1 << " targetPos2: " << targetPos2 );
}

void StandardSituationModule::calcPlayersAndPos4CornerKickUs(const Snapshot *worldSnap, 
							     const TacticalSnapshot *ts,
							     int &firstPlayer,
							     int &secondPlayer,
							     bool &needSecond,
							     Vector2d &targetPos1,
							     Vector2d &targetPos2) 
{
  needSecond = true;
  calcClosestDesiredPlayersTo(worldSnap->ball.pos, worldSnap, ts, firstPlayer, secondPlayer);
  calcPos4CornerKickUs(worldSnap, targetPos1, targetPos2);
  if (secondPlayer != UNKNOWN)
    {
      calcShortWays(worldSnap, firstPlayer, secondPlayer,
		    targetPos1, targetPos2);
      // if second player is too far away
      if (sqrnorm(targetPos2-worldSnap->ourTeam(secondPlayer).pos) 
	  > sqr(maxSecondDist))
	{
	  DECISIONTRACE( " second Player to far away " );
	  secondPlayer = UNKNOWN;
	  needSecond = false;
	}
    }
  else
    {
      needSecond = false;
    }

  DECISIONTRACE( " firstPlayer: " << firstPlayer << " secondPlayer: " << secondPlayer );
  DECISIONTRACE( " needSecond: " << needSecond );
  OTHERTRACE( " targetPos1: " << targetPos1 << " targetPos2: " << targetPos2 );
  DRAWTRACE( POSITION(targetPos1) );
  DRAWTRACE( POSITION(targetPos2) );
}

void StandardSituationModule::calcPlayersAndPos4KickInUs(const Snapshot *worldSnap, 
							 const TacticalSnapshot *ts,
							 int &firstPlayer,
							 int &secondPlayer,
							 bool &needSecond,
							 Vector2d &targetPos1,
							 Vector2d &targetPos2) 
{
  // calc the targets for the players

  double ourOffsideLine = worldSnap->ourOffsideLine(maxAge);
  // Distance of second player to first player
  double DIST = 6.;

  // if the ball is the offside line
  if (fabs(ourOffsideLine-worldSnap->ball.pos.getX()) < 1e-4)
    {
      DECISIONTRACE( " Ball is offsideline" );
      // we can only kick towards the other goal from
      // behind the ball in order to be not offside
      needSecond = true;
      // the first position is directly behind the ball on the line
      // from ball to other Goal
      targetPos1 = worldSnap->otherGoal.pos - worldSnap->ball.pos;
      targetPos1 *= -1.1*(BALL_SIZE+PLAYER_SIZE)/norm(targetPos1);
      targetPos1 += worldSnap->ball.pos;
      // the second position is directly in front of the ball on the line
      // from ball to our Goal
      targetPos2 = worldSnap->ourGoal.pos - worldSnap->ball.pos;
      targetPos2 *= DIST/norm(targetPos2);
      targetPos2 += worldSnap->ball.pos;
    }
  else
    {
      needSecond = true;
	
      // if we are not close to the other goal
      if (worldSnap->ball.pos.getX() < (FIELD_LENGTH/2-PENALTY_AREA_LENGTH))
	{
	  // the first position is directly behind the ball on the line
	  // from ball to other Goal
	  targetPos1 = worldSnap->otherGoal.pos - worldSnap->ball.pos;
	  targetPos1 *= -1.1*(BALL_SIZE+PLAYER_SIZE)/norm(targetPos1);
	  targetPos1 += worldSnap->ball.pos;
	  // the second position is on the same line
	  targetPos2 = worldSnap->otherGoal.pos - worldSnap->ball.pos;
	  targetPos2 *= DIST/norm(targetPos2);
	  targetPos2 += worldSnap->ball.pos;
	  // if this lies in offside, correct the misplacing
	  if (targetPos2.getX() > ourOffsideLine-1)
	    {  // Pythagoras
	      DECISIONTRACE( " targetPos2 is offside: " << targetPos2 );
	      double vz = worldSnap->ball.pos.getY();
	      vz /= fabs(vz);
	      double dx = fabs(ourOffsideLine-1-worldSnap->ball.pos.getX());
	      targetPos2.setX(ourOffsideLine-1);
	      targetPos2.setY(worldSnap->ball.pos.getY()-
			      vz*sqrt(sqr(DIST)-sqr(dx-1)));
	    }
	}
      else // if we are close to the other goal
	{
	  DECISIONTRACE( " We are close to the other goal doing corner kick" );
	  calcPos4CornerKickUs(worldSnap, targetPos1, targetPos2);
	}
    }
    
  // Find closest players to the ball
  calcClosestDesiredPlayersTo(worldSnap->ball.pos, worldSnap, ts, firstPlayer, secondPlayer);
  
  if (secondPlayer != UNKNOWN)
    {
      calcShortWays(worldSnap, firstPlayer, secondPlayer,
		    targetPos1, targetPos2);
      // if second player is too far away
      if (sqrnorm(targetPos2-worldSnap->ourTeam(secondPlayer).pos) 
	  > sqr(maxSecondDist))
	{
	  DECISIONTRACE( " second Player to far away " );
	  secondPlayer = UNKNOWN;
	  needSecond = false;
	}
    }
  else
    {
      needSecond = false;
    }

  DECISIONTRACE( " firstPlayer: " << firstPlayer << " secondPlayer: " << secondPlayer );
  DECISIONTRACE( " needSecond: " << needSecond );
  OTHERTRACE( " targetPos1: " << targetPos1 << " targetPos2: " << targetPos2 );
  DRAWTRACE( POSITION(targetPos1) );
  DRAWTRACE( POSITION(targetPos2) );
}

void StandardSituationModule::calcPlayersAndPos4FreeKickUs(const Snapshot *worldSnap, 
							   const TacticalSnapshot *ts,
							   int &firstPlayer,
							   int &secondPlayer,
							   bool &needSecond,
							   Vector2d &targetPos1,
							   Vector2d &targetPos2) 
{
  // calc the targets for the players

  double ourOffsideLine = worldSnap->ourOffsideLine(maxAge);
  // Distance of second player to first player
  double DIST = 6.;

  // if the ball is the offside line
  if (fabs(ourOffsideLine-worldSnap->ball.pos.getX()) < 1e-4)
    {
      DECISIONTRACE( " Ball is offsideline" );
      // we can only kick towards the other goal (see calc...4KickInUs)
      needSecond = true;
      // the first position is directly behind the ball on the line
      // from ball to other Goal
      targetPos1 = worldSnap->otherGoal.pos - worldSnap->ball.pos;
      targetPos1 *= -1.1*(BALL_SIZE+PLAYER_SIZE)/norm(targetPos1);
      targetPos1 += worldSnap->ball.pos;
      // the second position is directly in front of the ball on the line
      // from ball to our Goal
      targetPos2 = worldSnap->ourGoal.pos - worldSnap->ball.pos;
      targetPos2 *= DIST/norm(targetPos2);
      targetPos2 += worldSnap->ball.pos;
    }
  else
    {
      needSecond = true;
      // if we are close to the other goal
      if (worldSnap->ball.pos.getX() > FIELD_LENGTH/2-PENALTY_AREA_LENGTH-5)
	{
	  // if we are in their penalty area
	  if (fabs(worldSnap->ball.pos.getY()) <
	      PENALTY_AREA_WIDTH/2+3)
	    {
	      // the first position is directly behind the ball on the line
	      // from ball to other Goal
	      targetPos1 = worldSnap->otherGoal.pos - worldSnap->ball.pos;
	      targetPos1 *= -1.1*(BALL_SIZE+PLAYER_SIZE)/norm(targetPos1);
	      targetPos1 += worldSnap->ball.pos;
	      // the second pos is slightly in front of the first
	      targetPos2.setX(worldSnap->ball.pos.getX()+3);
	      // and to the hopefully open part of the goal
	      if (worldSnap->ball.pos.getY() < 0)
		targetPos2.setY(worldSnap->ball.pos.getY() + 7);
	      else
		targetPos2.setY(worldSnap->ball.pos.getY() - 7);
	    }
	  else
	    {
	      // the first position is directly behind the ball on the line
	      // form ball to (FIELD_LENGTH/4,0)
	      targetPos1 = Vector2d(FIELD_LENGTH/4,0)-worldSnap->ball.pos;
	      targetPos1 *= -1.1*(BALL_SIZE+PLAYER_SIZE)/norm(targetPos1);
	      targetPos1 += worldSnap->ball.pos;
	      // the second player is on the same line
	      targetPos2 = Vector2d(FIELD_LENGTH/4,0)-worldSnap->ball.pos;
	      targetPos2 *= DIST/norm(targetPos2);
	      targetPos2 += worldSnap->ball.pos;
	    }
	}
      else
	{
	  // the first position is directly behind the ball on the line
	  // from ball to other Goal
	  targetPos1 = worldSnap->otherGoal.pos - worldSnap->ball.pos;
	  targetPos1 *= -1.1*(BALL_SIZE+PLAYER_SIZE)/norm(targetPos1);
	  targetPos1 += worldSnap->ball.pos;
	  // if we are on their side
	  if (worldSnap->ball.pos.getX()-DIST > 0)
	    {
	      targetPos2 = -worldSnap->ball.pos;
	      targetPos2 *= -DIST/norm(targetPos2);
	    }
	  else
	    {
	      targetPos2 = worldSnap->otherGoal.pos - worldSnap->ball.pos;
	      targetPos2 *= DIST/norm(targetPos2);
	    }
	  targetPos2 += worldSnap->ball.pos;
	}
      // if this lies in offside, correct the misplacing
      if (targetPos2.getX() > ourOffsideLine-1)
	{  // Pythagoras
	  DECISIONTRACE( " targetPos2 is offside: " << targetPos2 );
	  double vz = worldSnap->ball.pos.getY();
	  vz /= fabs(vz);
	  double dx = fabs(ourOffsideLine-1-worldSnap->ball.pos.getX());
	  targetPos2.setX(ourOffsideLine-1);
	  targetPos2.setY(worldSnap->ball.pos.getY()-
			  vz*sqrt(sqr(DIST)-sqr(dx-1)));
	}
    }

  // find two players with minimal distance to ball
  calcClosestDesiredPlayersTo(worldSnap->ball.pos, worldSnap, ts, firstPlayer, secondPlayer);
  // if there is no second player take only one player
  if (secondPlayer == UNKNOWN)
    needSecond = false;

  if (needSecond)
    {
      calcShortWays(worldSnap, firstPlayer, secondPlayer,
		    targetPos1, targetPos2);
      // if second player is too far away
      if (sqrnorm(targetPos2-worldSnap->ourTeam(secondPlayer).pos) 
	  > sqr(maxSecondDist))
	{
	  DECISIONTRACE( " second Player to far away " );
	  secondPlayer = UNKNOWN;
	  needSecond = false;
	}
    }

  DECISIONTRACE( " firstPlayer: " << firstPlayer << " secondPlayer: " << secondPlayer );
  DECISIONTRACE( " needSecond: " << needSecond );
  OTHERTRACE( " targetPos1: " << targetPos1 << " targetPos2: " << targetPos2 );
  DRAWTRACE( POSITION(targetPos1) );
  DRAWTRACE( POSITION(targetPos2) );
}
  
void StandardSituationModule::calcPlayersAndPos4FreeKickThem(const Snapshot *worldSnap, 
							     const TacticalSnapshot *ts,
							     int &firstPlayer,
							     int &secondPlayer,
							     bool &needSecond,
							     Vector2d &targetPos1,
							     Vector2d &targetPos2) 
{
  // if the ball is close to our goal
  if (sqr(12.) < sqrnorm(worldSnap->ball.pos-worldSnap->ourGoal.pos) &&
      sqrnorm(worldSnap->ball.pos-worldSnap->ourGoal.pos) < sqr(30.))
    {
      // find two players with minimal distance to ball and goal
      double dists[11];
      firstPlayer = secondPlayer = UNKNOWN;
      Vector2d intersect = calcPossibleGoalPoint(worldSnap);
      // calc distances of our players
      for (int i = 0; i < 11; i++)
	{
	  WorldPlayer player = worldSnap->ourTeam(i+1);
	  dists[i] = 50000;
	  if (player.age < maxAge)
	    {
	      dists[i] = sqrnorm(player.pos-intersect)+
		sqrnorm(player.pos-worldSnap->ball.pos);
	      // make sure the goalie is not the first player
	      if ((i+1) == worldSnap->ourGoalieNo) dists[i] = 60000;
	      CMDTRACE( cout << "Player: " << worldSnap->ourTeam(i+1).number 
			<< " dist: " << dists[i] << endl );
	    }
	}
      // find the first
      double dist = 50000;
      for (int i = 0; i < 11; i++)
	{
	  if (dists[i] < dist) 
	    {
	      dist = dists[i];
	      firstPlayer = i+1;
	    }
	}
      dists[firstPlayer-1] = 50000;
      dist = 50000;
      // find the second
      for (int i = 0; i < 11; i++)
	{
	  if (dists[i] < dist) 
	    {
	      dist = dists[i];
	      secondPlayer = i+1;
	    }
	}
      
      // calc the postitions the players are going to
      
      // The left and the right edges of our goal
      Vector2d leftPost(worldSnap->ourGoal.pos.getX(),
			worldSnap->ourGoal.pos.getY()+GOAL_WIDTH/2),
	rightPost(worldSnap->ourGoal.pos.getX(),
		  worldSnap->ourGoal.pos.getY()-GOAL_WIDTH/2);
      
      DECISIONTRACE( " closest Player is " << firstPlayer );
      DECISIONTRACE( " second closest Player is " << secondPlayer );
      
      // calc the positions they will go to
      // on the lines from the ball to the left (right) edge of the goal
      targetPos1 = (leftPost - worldSnap->ball.pos); 
      targetPos2 = (rightPost - worldSnap->ball.pos);
      double angleBetweenPosts = angleBetween(targetPos1, targetPos2);
      needSecond = fabs(angleBetweenPosts) > DEG_TO_RAD(14.);
      double range = 1./3.;  // How far from the circle are we
      // a+range(b-a) == ...
      targetPos1.stretch(FREE_KICK_FORBIDDEN_AREA/norm(targetPos1));
      targetPos1 += worldSnap->ball.pos;
      targetPos1.stretch(1-range);
      targetPos1 += range*leftPost;
      targetPos2.stretch(FREE_KICK_FORBIDDEN_AREA/norm(targetPos2));
      targetPos2 += worldSnap->ball.pos;
      targetPos2.stretch(1-range);
      targetPos2 += range*rightPost;
      
  
      // if there is enough space between the two positions
      // move players a little together
      if (angleBetweenPosts > DEG_TO_RAD(18.))
	{
	  targetPos1 -= worldSnap->ball.pos;
	  targetPos1.turn(DEG_TO_RAD(-4.));
	  targetPos1 += worldSnap->ball.pos;
	  targetPos2 -= worldSnap->ball.pos;
	  targetPos2.turn(DEG_TO_RAD(4.));
	  targetPos2 += worldSnap->ball.pos;
	}
      
      if (secondPlayer == UNKNOWN)
	needSecond = false;
      
      if (needSecond)
	calcShortWays(worldSnap, 
		      firstPlayer, secondPlayer,  
		      targetPos1, targetPos2);
      else
	{
	  DECISIONTRACE( " dist to left target: " << norm(targetPos1-leftPost) 
			 << " dist to right target: " << norm(targetPos2-rightPost) );
	  if (sqrnorm(targetPos1 - leftPost) <= sqrnorm(targetPos2 - rightPost))
	    targetPos1 = targetPos2;
	}
      
      OTHERTRACE( " Wall positions are " << targetPos1 << " "  << targetPos2 );
      DRAWTRACE( POSITION(targetPos1) );
      DRAWTRACE( POSITION(targetPos2) );
      
      OTHERTRACE( " My pos: " << worldSnap->my.pos << " dist to ball: " 
		  << norm(worldSnap->ball.pos-worldSnap->my.pos) );
      OTHERTRACE( " Firstplayer pos: " << worldSnap->ourTeam(firstPlayer).pos 
		  << " dist to ball: " 
		  << norm(worldSnap->ball.pos-worldSnap->ourTeam(firstPlayer).pos) );
      if (needSecond)
	OTHERTRACE( " Secondplayer pos: " << worldSnap->ourTeam(secondPlayer).pos 
		    << " dist to ball: " 
		    << norm(worldSnap->ball.pos-worldSnap->ourTeam(secondPlayer).pos) );
      OTHERTRACE( " Angle between posts: " << fabs(RAD_TO_DEG(angleBetweenPosts)) );
      OTHERTRACE( " Ballpos: " << worldSnap->ball.pos << " Ballage: " 
		  << worldSnap->ball.age );
    }
  else 
    {
      if (sqrnorm(worldSnap->ball.pos-worldSnap->ourGoal.pos) > sqr(minWallDist))
	{
	  // the ball is not close to our goal
	  DECISIONTRACE( " ball is not close to our Goal" );
	  // 
	  //                     ---
	  //     targetPos1  x /     \ 
	  //     targetPos2  x \    /
	  //                     ---
	  //
	  //      : Ball
	  //         angleBetween positions == 45
	  //         bisector of this angle points to our goal
	  targetPos1.setRadAng(FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE,
			       DEG_TO_RAD(-157.5)); 
	  targetPos1.turn(-angleBetween(worldSnap->ourGoal.pos-worldSnap->ball.pos,
					Vector2d(-10,0)));
	  targetPos1 += worldSnap->ball.pos;
	  targetPos2.setRadAng(FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE,
			       DEG_TO_RAD(157.5)); 
	  targetPos2.turn(-angleBetween(worldSnap->ourGoal.pos-worldSnap->ball.pos,
					Vector2d(-10,0)));
	  targetPos2 += worldSnap->ball.pos;
	  
	  // if one of those pos is outside the field
	  // which means we are close to our goal but above or below
	  needSecond = true;
	  if (targetPos1.getX() < -FIELD_LENGTH/2)
	    {
	      targetPos1 = targetPos2;
	      needSecond = false;
	    }
	  if (targetPos2.getX() < -FIELD_LENGTH/2)
	    {
	      needSecond = false;
	    }
	  int dummyPlayer = firstPlayer = secondPlayer = UNKNOWN;
	  calcClosestDesiredPlayersTo(targetPos1, worldSnap, ts, firstPlayer, dummyPlayer);
	  calcClosestDesiredPlayersTo(targetPos2, worldSnap, ts, secondPlayer, dummyPlayer);
	  if (firstPlayer == secondPlayer)
	    if (dummyPlayer != UNKNOWN && 
		dummyPlayer != secondPlayer)
	      secondPlayer = dummyPlayer;
	    else
	      {
		secondPlayer = UNKNOWN;
		needSecond = false;
	      }
	  // if second player is too far away
	  if (secondPlayer != UNKNOWN && 
	  sqrnorm(targetPos2-worldSnap->ourTeam(secondPlayer).pos) 
	      > sqr(maxSecondDist))
	    {
	      DECISIONTRACE( " second Player to far away " );
	      secondPlayer = UNKNOWN;
	      needSecond = false;
	    }
	  
	  if (needSecond)
	    {
	      DECISIONTRACE( " closest Player is " << firstPlayer );
	      DECISIONTRACE( " second closest Player is " << secondPlayer );
	      DRAWTRACE( POSITION(targetPos1) );
	      DRAWTRACE( POSITION(targetPos2) );
	    }
	  else
	    {
	      // if there is only one player 
	      // the pos is between ball and our goal
	      targetPos1 = worldSnap->ourGoal.pos-worldSnap->ball.pos;
	      targetPos1 *= (FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE)/norm(targetPos1);
	      targetPos1 += worldSnap->ball.pos;
	      calcClosestDesiredPlayersTo(targetPos1, worldSnap, ts, firstPlayer, secondPlayer);
	      DECISIONTRACE( " closest Player is " << firstPlayer );
	      DRAWTRACE( POSITION(targetPos1) );
	    }
	}
      else
	{
	  DECISIONTRACE( " Ball is closer than " << minWallDist << " to our goal! " );
	  firstPlayer = secondPlayer = UNKNOWN;
	}
    }
}

Vector2d StandardSituationModule::calcPossibleGoalPoint(const Snapshot *worldSnap)
{
  // The left and the right edges of our goal
  Vector2d leftPost(worldSnap->ourGoal.pos.getX(),
		    worldSnap->ourGoal.pos.getY()+GOAL_WIDTH/2),
    rightPost(worldSnap->ourGoal.pos.getX(),
	      worldSnap->ourGoal.pos.getY()-GOAL_WIDTH/2);

  // calc the intersection of the bisector of the angle between the lines 
  // from the ball to the edges of the goal and
  // the line between the edges of the goal
  Vector2d intersect(rightPost-worldSnap->ball.pos);

  double a = sqrnorm(leftPost-worldSnap->ball.pos), 
    b = sqrnorm(rightPost-worldSnap->ball.pos),
    c = sqrnorm(leftPost-rightPost),
    alpha = acos((a+b-c)/(2.*sqrt(a*b)));
  
  intersect.turn(-alpha/2);
  
  double detA = intersect.getX()*(rightPost.getY()-leftPost.getY())+
    intersect.getY()*(rightPost.getX()-leftPost.getX());
  if (detA)
    {
      double lambda = 1/detA*((leftPost.getY()-rightPost.getY())*
			      (worldSnap->ball.pos.getX()-rightPost.getX())+
			      (rightPost.getX()-leftPost.getX())*
			      (worldSnap->ball.pos.getY()-rightPost.getY()));
      intersect.stretch(lambda);
    }
  
  intersect = intersect + worldSnap->ball.pos;  // point between the goal posts
  
  
  // if detA == 0 then the ball is on the line constructed by the goal posts
  // and a good place for the goalie is the goal post closest to the ball
  return intersect;
}

bool StandardSituationModule::goingTo(Vector2d &target, 
				      const double stamina,
				      AdvSkills *skills,
				      const Snapshot *worldSnap, 
				      const Vector2d *m = NULL,
				      const double r = 0)

{
  if (m && r>0)
    calcWayTo(worldSnap->my.pos, target, stamina, *m, r, worldSnap);
  walkAroundPlayers(worldSnap, target, stamina, m, r);

  // and run to it
  AdvSkills::TurtleCommand cmd = 
    skills->goToPosition(worldSnap, target, PLAYER_SIZE, stamina);

  if (cmd.state != AdvSkills::TFINISHED)
    {
      // we are not at the target yet, so run
      skills->exec(worldSnap, cmd);
      skills->exec(worldSnap, skills->turnHeadToBall(worldSnap, cmd));
      DECISIONTRACE( " Going towards " << target << " angle: " 
		     << RAD_TO_DEG((target-worldSnap->my.pos).getAngle()) );
      DRAWTRACE( POSITION(target) );
      return true;
    }
  else
    {
      // we arrived at the target, so turn to ball
      AdvSkills::TurnCommand tcmd = 
	skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->ball)+
		     worldSnap->my.bodyDir);
      if (tcmd.state == AdvSkills::RANGLE ||
	  tcmd.state == AdvSkills::ROK)
	{
	  DECISIONTRACE( " Turning to ball " );
	  skills->exec(worldSnap, tcmd);
	  skills->exec(worldSnap, 
		       skills->turnHead(worldSnap, 
					(worldSnap->ball.pos-
					 worldSnap->my.pos).getAngle()+
					worldSnap->my.bodyDir-tcmd.bodyDir));
	  return true;
	}
    }
  // we reached the target and are looking to ball, so we are !goingTo()
  return false;
}

bool StandardSituationModule::goingBackwardsTo(Vector2d &target, 
					       const double stamina,
					       AdvSkills *skills,
					       const Snapshot *worldSnap, 
					       const Vector2d *m = NULL,
					       const double r = 0)

{
  if (m && r>0)
    calcWayTo(worldSnap->my.pos, target, stamina, *m, r, worldSnap);
  walkAroundPlayers(worldSnap, target, stamina, m, r);

  AdvSkills::TurtleCommand cmd = 
    skills->goToPosition(worldSnap, target, PLAYER_SIZE, stamina, 
			 true, (worldSnap->ball.pos - worldSnap->my.pos)
			 .getAngle());

  if (cmd.state != AdvSkills::TFINISHED)
    {
      // we are not at the target yet, so run
      skills->exec(worldSnap, cmd);
      skills->exec(worldSnap, skills->turnHeadToBall(worldSnap, cmd));
      DECISIONTRACE( " Going towards " << target << " angle: " 
		     << RAD_TO_DEG((target-worldSnap->my.pos).getAngle()) );
      DRAWTRACE( POSITION(target) );
      return true;
    }
  else
    {
      AdvSkills::TurnCommand tcmd = 
	skills->turn(worldSnap, worldSnap->my.bodyAngleTo(worldSnap->ball)+
		     worldSnap->my.bodyDir);
      if (tcmd.state == AdvSkills::RANGLE ||
	  tcmd.state == AdvSkills::ROK)
	{
	  DECISIONTRACE( " Turning to ball " );
	  skills->exec(worldSnap, tcmd);
	  skills->exec(worldSnap, 
		       skills->turnHead(worldSnap, 
					(worldSnap->ball.pos-
					 worldSnap->my.pos).getAngle()+
					worldSnap->my.bodyDir-tcmd.bodyDir));
	  return true;
	}
    }
  // we reached the target and are looking to ball, so we are !goingBackwardsTo()
  return false;
}

bool StandardSituationModule::doingGoalieFreeKickThem(AdvSkills *skills,
						      const Snapshot *worldSnap)

{
  DECISIONTRACE( "Balldist: " << worldSnap->my.distTo(worldSnap->ball) );

  // calc goalie position
  Vector2d intersect = calcPossibleGoalPoint(worldSnap);
  Vector2d pos = intersect - worldSnap->ball.pos;
  pos = (1. - 3./norm(pos))*pos + worldSnap->ball.pos;
  
  // if this pos is inside of FREE_KICK_FORBIDDEN_AREA
  if (sqrnorm(pos-worldSnap->ball.pos) <= sqr(FREE_KICK_FORBIDDEN_AREA))
    {
      // translate it PLAYER_SIZE meters in front of it
      pos =
	(pos-worldSnap->ball.pos).stretch((FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE)/
					  norm(pos-worldSnap->ball.pos))+
	worldSnap->ball.pos;
    }
  
  // if this is behind the goal
  if (pos.getX() < -FIELD_LENGTH/2)
    {
      ERRORTRACE( " This freekick is closer than "\
		  "FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE to the goal! :-( " );
      DECISIONTRACE( " Goalie turns to ball " );
      // goalie should do anything (walk to good position near the circle around 
      // the ball or so) not only look to ball
      Snapshot mySnap = *worldSnap, *ws = &mySnap;
      AdvSkills::TurnCommand tcmd;
      skills->exec(worldSnap, tcmd =
		   skills->turn(worldSnap, 				 
				worldSnap->my.bodyAngleTo(worldSnap->ball)+
				worldSnap->my.bodyDir));
      ws->my.bodyDir = tcmd.bodyDir;
      skills->exec(worldSnap, skills->turnHead(ws, ws->my.bodyAngleTo(ws->ball)+
					       ws->my.bodyDir));
      return true;
    }

  double stamina = STAMINA_INC_MAX;
  if (worldSnap->senseBody.stamina > 
      goalieStaminaFactor*2.*STAMINA_INC_MAX+RECOVER_DEC_THR*STAMINA_MAX)
    stamina = goalieStaminaFactor*2.*STAMINA_INC_MAX;
  
  if (!goingBackwardsTo(pos, stamina, skills, worldSnap))
    {
      lookAround(skills, worldSnap);
      skills->noAction();
      isSpecialSituation = true;
      DECISIONTRACE( "Using SPECIAL SITUATION channel" );
    }
  return true;
}

bool StandardSituationModule::doingFieldPlayerFreeKickThem(AdvSkills *skills,
							   const Snapshot *worldSnap,
							   int firstPlayer,
							   int secondPlayer,
							   bool needSecond,
							   Vector2d &targetPos1,
							   Vector2d &targetPos2)

{
  if (sqr(minWallDist) < sqrnorm(worldSnap->ourGoal.pos-worldSnap->ball.pos) &&
      sqrnorm(worldSnap->ourGoal.pos-worldSnap->ball.pos)< sqr(maxWallDist))
    DECISIONTRACE( " Ball is dangerous!, Balldist: " 
		   << worldSnap->my.distTo(worldSnap->ball) );
  
  Vector2d targetPos;

  // Am I the first player?
  if (worldSnap->my.number == firstPlayer && needSecond)
    {
      DECISIONTRACE( " I am the first of two players" );
      targetPos = targetPos1;
    }

  // Am I the only player?
  if (worldSnap->my.number == firstPlayer && !needSecond)
    {
      DECISIONTRACE( " I am the only player " );
      targetPos = targetPos1;
    }

  // if I am the second player
  if (worldSnap->my.number == secondPlayer && needSecond)
    {
      DECISIONTRACE( " I am the second player" );
      targetPos = targetPos2;
    }

  DECISIONTRACE( " My targetPos: " << targetPos );
  DRAWTRACE( POSITION(targetPos) );

  double stamina = STAMINA_INC_MAX;
  // if we have enough stamina run faster
  if (worldSnap->senseBody.stamina > 
      wallStaminaFactor*STAMINA_INC_MAX+RECOVER_DEC_THR*STAMINA_MAX)
    stamina = wallStaminaFactor*STAMINA_INC_MAX;
  
  if (!goingTo(targetPos, stamina, skills, worldSnap, &(worldSnap->ball.pos),
	       FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE))
    {
      DRAWTRACE( POSITION(targetPos) );
      lookAround(skills, worldSnap);
      skills->noAction();
    }
  return true;
  
}

bool StandardSituationModule::doingKickInCornerKickThem(AdvSkills *skills,
							const Snapshot *worldSnap,
							int firstPlayer,
							int secondPlayer,
							bool needSecond,
							Vector2d &targetPos1,
							Vector2d &targetPos2)
  
{
  Vector2d targetPos;

  // Am I the first player?
  if (worldSnap->my.number == firstPlayer && needSecond)
    {
      DECISIONTRACE( " I am the first of two players" );
      targetPos = targetPos1;
    }

  // Am I the only player?
  if (worldSnap->my.number == firstPlayer && !needSecond)
    {
      DECISIONTRACE( " I am the only player " );
      targetPos = targetPos1;
    }

  // if I am the second player
  if (worldSnap->my.number == secondPlayer && needSecond)
    {
      DECISIONTRACE( " I am the second player" );
      targetPos = targetPos2;
    }

  DECISIONTRACE( " My targetPos: " << targetPos );
  OTHERTRACE( " my pos: " << worldSnap->my.pos );
  DRAWTRACE( POSITION(targetPos) );
  
  double stamina = STAMINA_INC_MAX;
  if (worldSnap->senseBody.stamina >
      passiveStaminaFactor*STAMINA_INC_MAX+RECOVER_DEC_THR*STAMINA_MAX)
    stamina = passiveStaminaFactor*STAMINA_INC_MAX;
  
  if (!goingTo(targetPos, stamina, skills, worldSnap, &(worldSnap->ball.pos),
	       FREE_KICK_FORBIDDEN_AREA+PLAYER_SIZE))
    {
      DRAWTRACE( POSITION(targetPos) );
      lookAround(skills, worldSnap);
      skills->noAction();
    }
  return true;
  
}

// check if players of the other team are in a circle with radius r around pos
bool StandardSituationModule::posIsFree(const Vector2d &pos, const Snapshot *worldSnap, 
					const double r = 3.)
{
  double sqr_r = sqr(r);   // for axel
  for (int i = 1; i <= 11; i++)
    if (sqrnorm(worldSnap->otherTeam(i).pos-pos) < sqr_r)
      return false;
  return true;
}

void StandardSituationModule::calcWayTo(const Vector2d &from, Vector2d &to, 
					const double stamina,
					const Vector2d &m, const double &r, 
					const Snapshot *worldSnap)
{
  if (worldSnap);     // to avoid warnings

  OTHERTRACE( "Dist to target: " << norm(from-to) );
  // can we walk from "from" to "to" without entering the circle defined by m and r

  // if we are close to the target do nothing
  if (sqrnorm(to-from)<=sqr(PLAYER_SIZE+0.1)) 
    {
      to = from;
      return;
    }

  // if I am inside the circle
  double nrm = sqrnorm(from-m);
  if (nrm < sqr(r-0.2))
    {
      DECISIONTRACE( " walking out of the circle" );
      // if I am close to the middle of the circle
      if (nrm < sqr(1e-4))
	{
	  // go towards target
	  to -= from;
	  to *= (DASH_POWER_RATE*stamina+norm(worldSnap->my.v)+1)/norm(to);
	  to += worldSnap->my.pos;
	}
      else
	{ // take the shortest way out of the circle
	  to = from - m;
	  to *= (DASH_POWER_RATE*stamina+norm(worldSnap->my.v)+1)/norm(to);
	  to += from;
	}
      return;
    }

  bool weAreClose = sqrnorm((from-m).stretch(r/norm(from-m))+m-from) 
    <= sqr(PLAYER_SIZE+0.1);
  Vector2d c = (from-m).stretch(r/norm(from-m));
  // if the direct way leeds through the circle
  if ((fabs(asin(r/norm(m-from))) -  
      fabs(angleBetween(m-from, (1-PLAYER_SIZE/norm(to-m))*(to-m)+m-from)) >=
      -1e-4) && (sqrnorm(to-from) > sqrnorm(m-from)))
    {
      DECISIONTRACE( " Can not walk directly to " << to );
      if (weAreClose)
	{
	  Vector2d rightPos = c;
	  rightPos.perp();
	  Vector2d leftPos = rightPos;
	  leftPos.stretch(-1.0);
	  rightPos += from;
	  leftPos += from;
	  // go tangential to the circle closer to the target
	  if ( sqrnorm(leftPos - to) < sqrnorm(rightPos - to ) )
	    {
	      to = leftPos;
	      DECISIONTRACE( "close to the ball, going to the left side" );
	    }
	  else
	    {
	      to = rightPos;
	      DECISIONTRACE( "close to the ball, going to the right side" );
	    }
	}
      else
	{
	  // we are not close to the ball and have to decide if we go left or right
	  Vector2d rightPos = c;
	  rightPos.stretch(1.1);
	  rightPos.perp();
	  Vector2d leftPos = rightPos;
	  leftPos.stretch(-1.0);
	  leftPos += m;
	  rightPos += m;
	  OTHERTRACE( " leftPos: " << leftPos << " rightPos: " << rightPos);
	  OTHERTRACE( " c: " << c);

	  if ( sqrnorm(leftPos - to) < sqrnorm(rightPos - to) )
	    {
	      // go to the left pos
	      to = c;
	      to.turn( -fabs(acos(r / norm(m-from))) );
	      DECISIONTRACE( "not close to the ball, going to the left side" );
	    }
	  else
	    {
	      // go to the right pos
	      to = c;
	      to.turn( fabs(acos(r / norm(m-from))) ); 
	      DECISIONTRACE( " not close to the ball, going to the right side " );
	    }
	  to.stretch( 1.1 );
	  to += m;
	  // stretch "to" to the length we can go + a little more for security
	  // goingTo() should not turn to ball before reaching the real target
	  to -= from;
	  to *= (DASH_POWER_RATE*stamina+norm(worldSnap->my.v)+1)/norm(to);
	  to += from;
	  OTHERTRACE( " turnAngle: " << RAD_TO_DEG(fabs(acos(r / norm(m-from)))) 
		      << " to: " << to);
	}
    }
}

void StandardSituationModule::walkAroundPlayers(const Snapshot *worldSnap, Vector2d &target, 
						const double stamina = STAMINA_INC_MAX, const Vector2d *m = NULL, 
						const double r = 0)
{
  if (sqrnorm(target - worldSnap->my.pos) < sqr(PLAYER_SIZE+0.1))
    return;

  double sqr_r = sqr(r);
  // that is how far we can go in this cycle using stamina
  // + a litte more as we don't want goToPosition() returns 
  // FINISHED
  double stretch = stamina*DASH_POWER_RATE+norm(worldSnap->my.v)+1;
  // the angle we use to turn
  const double angle = DEG_TO_RAD(15.);
  // the length of the corridor
  const double length = 5;
  // when turning the corridor its length should be
  // increased by addLength to check if there is
  // a wall of players in front of me
  double addLength;
  // the direction we want to go to
  Vector2d dir = target-worldSnap->my.pos;
  dir *= 1./norm(dir);
  Vector2d mypos = worldSnap->my.pos,
    target1, target2;

  // how many players are in front of me
  // I am always in front of me!
  int ourCount = worldSnap->ourTeamInCorridor(mypos-0.1*dir, 
					      mypos+length*dir,
					      4*PLAYER_SIZE+0.2);
  int otherCount = worldSnap->otherTeamInCorridor(mypos-0.1*dir,
						  mypos+length*dir,
						  4*PLAYER_SIZE+0.2);
  DRAWTRACE( CORRIDOR(mypos-0.1*dir, mypos+length*dir, 4*PLAYER_SIZE+0.2, PI_2 ) );
  OTHERTRACE( " ourCount: " << ourCount << " otherCount: " << otherCount );
  if (ourCount + otherCount > 1)
    {
      // turning to the right till there is a way
      for (int i = 1; i < (2.*M_PI)/angle; i++)
	{
	  OTHERTRACE( ourCount+otherCount << " Player(s) in front of me " );
	  dir.turn(angle);
	  if (i < M_PI/angle)
	    addLength = length/cos(angle)-length;
	  else
	    addLength = 0;
	  if ( (ourCount = worldSnap->ourTeamInCorridor(mypos-0.1*dir, 
							mypos+(length+addLength)*dir, 
							4*PLAYER_SIZE+0.2))+
	       (otherCount = worldSnap->otherTeamInCorridor(mypos-0.1*dir, 
							    mypos+(length+addLength)*dir, 
							    4*PLAYER_SIZE+0.2)) <= 1)
	    {
	      DRAWTRACE( CORRIDOR(mypos-0.1*dir, mypos+(length+addLength)*dir, 
				  4*PLAYER_SIZE+0.2, PI_2 ) );
	      target1 = mypos + stretch*dir;
	      break;
	    }
	}

      dir = target-mypos;
      dir *= 1./norm(dir);
      ourCount = worldSnap->ourTeamInCorridor(mypos-0.1*dir, 
					      mypos+length*dir, 4*PLAYER_SIZE+0.2);
      otherCount = worldSnap->otherTeamInCorridor(mypos-0.1*dir, 
						  mypos+length*dir, 4*PLAYER_SIZE+0.2);
      // turning to the left till there is a way
      for (int i = 1; i < (2.*M_PI)/angle; i++)
	{
	  OTHERTRACE( ourCount+otherCount << " Player(s) in front of me " );
	  dir.turn(-angle);
	  if (i < M_PI/angle)
	    addLength = length/cos(angle)-length;
	  else
	    addLength = 0;
	  if ( (ourCount = worldSnap->ourTeamInCorridor(mypos-0.1*dir, 
							mypos+(length+addLength)*dir, 
							4*PLAYER_SIZE+0.2))+
	       (otherCount = worldSnap->otherTeamInCorridor(mypos-0.1*dir, 
							    mypos+(length+addLength)*dir, 
							    4*PLAYER_SIZE+0.2)) <= 1)
	    {
	      DRAWTRACE( CORRIDOR(mypos-0.1*dir, mypos+(length+addLength)*dir, 4*PLAYER_SIZE+0.2, PI_2 ) );
	      target2 = mypos + stretch*dir;
	      break;
	    }
	}

      // check if one of those targets is in the circle
      if ((m && r>0) && sqrnorm(target1-*m) < sqr_r)
	{
	  target = target2;
	  DECISIONTRACE( " going to target1 " << target1 << " as target2 in circle " );
	  DRAWTRACE( LLINE(mypos,mypos+10*(target-mypos)) );
	  return;
	}
      if ((m && r>0) && sqrnorm(target2-*m) < sqr_r)
	{
	  target = target1;
	  DECISIONTRACE( " going to target2 " << target2 << " as target1 in circle " );
	  DRAWTRACE( LLINE(mypos,mypos+10*(target-mypos)) );
	  return;
	}
	
      // which way is the shortest
      if (sqr(norm(target-target1)+2) < sqrnorm(target-target2))
	{
	  target = target1;
	  DECISIONTRACE( " going to target1 " << target1 );
	}
      else
	{
	  target = target2;
	  DECISIONTRACE( " going to target2 " << target1 );
	}
      DRAWTRACE( LLINE(mypos,mypos+10*(target-mypos)) );
    }
}

bool StandardSituationModule::doingFreeKickUs(AdvSkills *skills,
					      const Snapshot *worldSnap,
					      int firstPlayer, int secondPlayer,
					      bool needSecond,
					      Vector2d &targetPos1, 
					      Vector2d &targetPos2)
{
  Vector2d targetPos;
  
  if (worldSnap->my.number == firstPlayer)
    {
      if (needSecond)
	DECISIONTRACE( " I am the first player " );
      else
	DECISIONTRACE( " I am the only player " );
      targetPos = targetPos1;
    }

  if (worldSnap->my.number == secondPlayer)
    {
      DECISIONTRACE( " I am the second player " );
      targetPos = targetPos2;
    }

  DECISIONTRACE( " My targetPos: " << targetPos );
  DECISIONTRACE( " My pos: " << worldSnap->my.pos );

  // the players should walk around the ball
  double r = BALL_SIZE+PLAYER_SIZE;
  const Vector2d *m = &(worldSnap->ball.pos);

  double stamina = STAMINA_INC_MAX;
  if (worldSnap->senseBody.stamina >
      activeStaminaFactor*STAMINA_INC_MAX+RECOVER_DEC_THR*STAMINA_MAX)
    stamina = activeStaminaFactor*STAMINA_INC_MAX;
  

  if (!goingTo(targetPos, stamina, skills, worldSnap, m, r))
    { 
      // look what we can do
      if (worldSnap->playMode == PLAY_MODE_OFFSIDE_THEM)
	{
	  DECISIONTRACE( "we must not kick because of PLAY_MODE_OFFSIDE_THEM, doing nothing" );
	  lookAround(skills, worldSnap);
	  skills->noAction();
	  return false;
	}

      // first check if we can make a goal
      Grade g; g.set_AVERAGE();
      if (worldSnap->my.number == firstPlayer)
	{
	  // if a goalkick is better than AVERAGE
	  if (goalKickEval->ourEval(worldSnap, worldSnap->my.pos)
	      .is_better_than(g))
	    {
	      DECISIONTRACE( " Want to make a goal " );
	      // kick the ball and let the goalkickmodule handle it
	      skills->exec(worldSnap, skills->kick(worldSnap, Vector2d(0.1,0),
					    AdvSkills::KICKAWAY));
	      return true;
	    }

	  // calc if it is good to pass but
	  // wait for second player to arrive its position
	  // but only wait for waitingTime cycles
	  OTHERTRACE( " waitingTime: " << waitingTime );
	  if (needSecond && 
	      ((sqrnorm(targetPos2-worldSnap->ourTeam(secondPlayer).pos) < 
		sqr(1.7))) || waitingTime) 
	    {
	      // if there is a pass better than AVERAGE
	      for (int player = 1; player <= 11; player++)
		{
		  if (worldSnap->ourTeam(player).age < maxAge &&
		      firstPlayer != player)
		    {
		      // if the player is close to the outer line 
		      if (fabs(worldSnap->ourTeam(player).pos.getY())+2 > 
			  FIELD_WIDTH/2 || 
			  fabs(worldSnap->ourTeam(player).pos.getX())+2 >
			  FIELD_LENGTH/2)
			continue;  // check for the next one

		      if (passEval->ourEval(worldSnap, firstPlayer, player)
			  .is_better_than(g))
			{
			  DECISIONTRACE( " Want to pass " );
			  // just do it
			  skills->exec(worldSnap, skills->kick(worldSnap,
							Vector2d(0.1,0),
							AdvSkills::KICKAWAY));
			  return true;
			}
		    }
		} // end for

	      // second player has arrived but I don't want to pass or
	      // make a goal kick
	      if (!waitingTime)
		{
		  DECISIONTRACE( " Time exceeded and didn't want to pass " );
		  DECISIONTRACE( " Don't know what to do, so playing to second player " );
		  skills->exec(worldSnap, skills->kick(worldSnap, 2., targetPos2,
						AdvSkills::KICKAWAY));
		  return true;
		}
	      if (!needSecond || 
		  (sqrnorm(targetPos2-worldSnap->ourTeam(secondPlayer).pos) < sqr(1.7)))
		  DECISIONTRACE( " Don't know what to do so waiting " );
	      else
		  DECISIONTRACE( " Waiting for second player " );
	      lookAround(skills, worldSnap);
	      skills->noAction();
	      waitingTime--;
	      return false;
	    }
	  else
	    {	    
	      DECISIONTRACE( " Time exceeded, second player not on pos " );
	      DECISIONTRACE( " Don't know what to do, so just playing on" );
	      skills->exec(worldSnap, skills->kick(worldSnap, Vector2d(0.1,0),
					    AdvSkills::FIXED));
	      return true;
	    }
	}
      else
	{ // I am the second player
	  DECISIONTRACE( " Arrived my position and watching the ball " );
	  lookAround(skills, worldSnap);
	  skills->noAction();
	  // if I am behind the first player tell him what I see
	  if (worldSnap->my.pos.getX() < targetPos.getX())
	    isSpecialSituation = true;
	  return false;
	}
    }
  return true;
}


void StandardSituationModule::lookAround(AdvSkills *skills,
					 const Snapshot *worldSnap)
{
  // look to the left or the right of the ball, to get more see information 
  // if the ball is very near, look in front of me

  Angle turnHeadDir;

  if ( sqrnorm(worldSnap->ball.pos - worldSnap->my.pos) < 1.0 )  
    turnHeadDir = worldSnap->my.bodyAngleTo(Vector2d(0.0,0.0)) +
      worldSnap->my.bodyDir;
  else
    turnHeadDir = worldSnap->my.bodyAngleTo(worldSnap->ball) + 
      worldSnap->my.bodyDir;
  
  // decide where to look next
  if ( nextTurnAction == TURN_TO_LEFT_SIDE )
    turnHeadDir += DEG_TO_RAD(lookAroundAngle);
  else
    turnHeadDir += DEG_TO_RAD(-lookAroundAngle);
  AdvSkills::TurnHeadCommand thcmd = skills->turnHead(worldSnap, turnHeadDir);
  
  if ((thcmd.state == AdvSkills::THOK) || 
      (thcmd.state == AdvSkills::THANGLE))
    {
      skills->exec(worldSnap, thcmd);
      if ( nextTurnAction == TURN_TO_LEFT_SIDE )
	{
	  DECISIONTRACE( " turning head to the left " );
	  DECISIONTRACE( " turnHeadDir: " << RAD_TO_DEG(turnHeadDir) 
			 << " faceDir: " 
				 << RAD_TO_DEG(worldSnap->my.faceDir));
	  nextTurnAction = TURN_TO_RIGHT_SIDE;
	}
      else
	{
	  DECISIONTRACE(" turning head to the right ");
	  DECISIONTRACE( "turnHeadDir: "<< RAD_TO_DEG(turnHeadDir)
			 << " faceDir: " 
			 << RAD_TO_DEG(worldSnap->my.faceDir));
	  nextTurnAction = TURN_TO_LEFT_SIDE;
	}
    }
  else
    DECISIONTRACE( " player is not turning, because state = " 
		   << thcmd.state );
}

bool StandardSituationModule::doingKickInUs(AdvSkills *skills,
					    const Snapshot *worldSnap,
					    int firstPlayer, int secondPlayer,
					    bool needSecond,
					    Vector2d &targetPos1, 
					    Vector2d &targetPos2)
{
  Vector2d targetPos;
  
  if (worldSnap->my.number == firstPlayer)
    {
      if (needSecond)
	DECISIONTRACE( " I am the first player " );
      else
	DECISIONTRACE( " I am the only player " );
      targetPos = targetPos1;
    }

  if (worldSnap->my.number == secondPlayer)
    {
      DECISIONTRACE( " I am the second player " );
      targetPos = targetPos2;
    }

  DECISIONTRACE( " My targetPos: " << targetPos );
  DECISIONTRACE( " My pos: " << worldSnap->my.pos );

  // players should walk around the ball
  double r = BALL_SIZE+PLAYER_SIZE;
  const Vector2d *m = &(worldSnap->ball.pos);

  double stamina = STAMINA_INC_MAX;
  if (worldSnap->senseBody.stamina >
      activeStaminaFactor*STAMINA_INC_MAX+RECOVER_DEC_THR*STAMINA_MAX)
    stamina = activeStaminaFactor*STAMINA_INC_MAX;
  
  if (!goingTo(targetPos, stamina, skills, worldSnap, m, r))
    if (worldSnap->my.number == firstPlayer)
      {
	if (needSecond)
	  {
	    if (waitingTime)
	      {
		waitingTime--;
		// if second player has not arrived its position
		if (sqrnorm(worldSnap->ourTeam(secondPlayer).pos - 
			    targetPos2) >= sqr(1.7))
		  {
		    // wait for him
		    DECISIONTRACE( " Waiting for second Player to arrive its position" );
		    lookAround(skills, worldSnap);
		    skills->noAction();
		    return false;
		  }
		else
		  {
		    // second player arrived, so kick to him
		    AdvSkills::ExtKickCommand kcmd;
		    kcmd = skills->kick(worldSnap, 2., targetPos2,
					AdvSkills::KICKAWAY);
		    skills->exec(worldSnap, kcmd);
		    // when stdsit is over run into the field
		    double vz = worldSnap->my.pos.getY()/
		      fabs(worldSnap->my.pos.getY());
		    myPlan.setRunToPosition(worldSnap->my.pos-
					    Vector2d(0,vz*7),
					    STAMINA_INC_MAX,
					    7);
		    return true;
		  }
	      }
	  }
	// no second player needed or waitingTime exceeded
	if (!waitingTime)
	  DECISIONTRACE( " Time exceeded" );
	DECISIONTRACE( " Don't know what to do, so playing on" );
	// if we are in our half or close to the other goal
	if (worldSnap->ball.pos.getX() < 0 ||
	    worldSnap->ball.pos.getX() > 
	    (FIELD_LENGTH/2-PENALTY_AREA_LENGTH))
	  // kick to (0,0)
	  skills->exec(worldSnap,
		       skills->kick(worldSnap, (-worldSnap->ball.pos)
				    .stretch(0.5/norm(worldSnap->
						      ball.pos)),
				    AdvSkills::KICKAWAY));
	else  // kick to other goal
	  skills->exec(worldSnap,
		       skills->kick(worldSnap,(worldSnap->otherGoal.pos-worldSnap->ball.pos)
				    .stretch(0.5/norm(worldSnap->ball.pos-
						      worldSnap->otherGoal.pos)),
				    AdvSkills::KICKAWAY));
	
      }
    else  // I am the second player
      {
	DECISIONTRACE( " Arrived my position and watching ball" );
	lookAround(skills, worldSnap);
	skills->noAction();
	return false;
      }
  return true;
}

bool StandardSituationModule::doingCornerKickUs(AdvSkills *skills,
						const Snapshot *worldSnap,
						int firstPlayer, int secondPlayer,
						bool needSecond,
						Vector2d &targetPos1, 
						Vector2d &targetPos2)
{
  Vector2d targetPos;
  
  if (worldSnap->my.number == firstPlayer)
    {
      if (needSecond)
	DECISIONTRACE( " I am the first player " );
      else
	DECISIONTRACE( " I am the only player " );
      targetPos = targetPos1;
    }

  if (worldSnap->my.number == secondPlayer)
    {
      DECISIONTRACE( " I am the second player " );
      targetPos = targetPos2;
    }

  DECISIONTRACE( " My targetPos: " << targetPos );
  DECISIONTRACE( " My pos: " << worldSnap->my.pos );
  
  // players should walk around the ball
  double r = BALL_SIZE+PLAYER_SIZE;
  const Vector2d *m = &(worldSnap->ball.pos);
  
  double stamina = STAMINA_INC_MAX;
  if (worldSnap->senseBody.stamina >
      activeStaminaFactor*STAMINA_INC_MAX+RECOVER_DEC_THR*STAMINA_MAX)
    stamina = activeStaminaFactor*STAMINA_INC_MAX;
  
  if (!goingTo(targetPos, stamina, skills, worldSnap, m, r))
    if (worldSnap->my.number == firstPlayer)
      {
	if (needSecond)
	  {
	    if (waitingTime)
	      {
		waitingTime--;
		// if second player has not arrived its position
		if (sqrnorm(worldSnap->ourTeam(secondPlayer).pos - 
			    targetPos2) >= sqr(1.7))
		  {
		    // wait for him
		    DECISIONTRACE( " Waiting for second Player to arrive its position" );
		    lookAround(skills, worldSnap);
		    skills->noAction();
		    isSpecialSituation = true;
		    return false;
		  }
		else
		  {
		    // here targetPos2 is outside the FREE_KICK_FORBIDDEN_AREA circle
		    if (posIsFree(targetPos2, worldSnap))
		      {
			AdvSkills::ExtKickCommand kcmd;
			kcmd = skills->kick(worldSnap, 2., targetPos2,
					    AdvSkills::KICKAWAY);
			skills->exec(worldSnap, kcmd);
			// run into the field, when stdsit is over
			myPlan.setRunToPosition(Vector2d(0,0), STAMINA_INC_MAX, 7);
		      }
		    else
		      {
			// kick towards (0,0)
			myPlan.setNothing();
			skills->exec(worldSnap, 
				     skills->kick(worldSnap, (-worldSnap->ball.pos)
						  .stretch(0.5/norm(worldSnap->
								    ball.pos)),
						  AdvSkills::KICKAWAY));
		      }
		    return true;
		  }
	      }
	  }
	// no second player needed or waitingTime exceeded
	if (!waitingTime)
	  DECISIONTRACE( " Time exceeded" );
	DECISIONTRACE( " Don't know what to do, so playing on" );
	// kick towards (0,0)
	skills->exec(worldSnap,
		     skills->kick(worldSnap, (-worldSnap->ball.pos)
				  .stretch(0.5/norm(worldSnap->
						    ball.pos)),
				  AdvSkills::KICKAWAY));
	return true;
      }
    else
      {
	// I am the second player
	DECISIONTRACE( " Arrived my position and watching ball" );
	lookAround(skills, worldSnap);
	skills->noAction();
	return false;
      }
  
  // as we are goingTo
  return true;
}

bool StandardSituationModule::doingKickOff(AdvSkills *skills,
					   const Snapshot *worldSnap,
					   TacticalSnapshot *tacticalSnap)

{
  // moving to standard kickoff-positions and turning after that
  Vector2d moveTarget;
  moveTarget = calcKickOffTarget(worldSnap->my.number, worldSnap, tacticalSnap);
  if ((worldSnap->my.distTo(moveTarget) > 0.4) || 
      (worldSnap->my.pos.getX() > -0.1) )
    {
      if ((worldSnap->playMode == PLAY_MODE_BEFORE_KICK_OFF) ||
	  (worldSnap->playMode == PLAY_MODE_GOAL_US) ||
	  (worldSnap->playMode == PLAY_MODE_GOAL_THEM) 
	  )
	{
	  skills->basicMove(moveTarget.getX(),moveTarget.getY());
	  DECISIONTRACE( "doingKickOff: moves the player to the kickoff position at " 
			 << moveTarget );
	  return true;
	}      
      if (worldSnap->my.distTo(moveTarget) > 0.8)
	{
	  DECISIONTRACE( "doingKickOff: moves not possible, walking to" 
			 << moveTarget );
	  skills->exec(worldSnap, skills->goToPosition(worldSnap, moveTarget, 0.8));
	  return true;
	}
    }
  if (worldSnap->playMode == PLAY_MODE_BEFORE_KICK_OFF)
    {
      skills->basicTurn(M_PI/4.0);
      DECISIONTRACE( "doingKickOff: turns the player (cause we are the MRB)" );
      return true;
    }



  if (!weAreKickingOff)
    {
      DECISIONTRACE( "doingKickOff: otherTeam is kicking off, so turning to ball" );
      skills->basicTurn(worldSnap->my.bodyAngleTo(worldSnap->ball));
      return true;
    }  
   
  if (!skills->kickSafe(worldSnap))
    {
      DECISIONTRACE( "I am not responsible for the kick off, do nothing" );
      skills->noAction();
      return true;
    }
   

  // find out where to kick to
  int passReceiver;
  Vector2d kickTarget = calcKickOffKickTarget(worldSnap, tacticalSnap, passReceiver);
  DECISIONTRACE( "I am responsible for the kick off, kicking to " 
		 << kickTarget );     
  AdvSkills::ExtKickCommand kickCommand = skills->kick(worldSnap, 0.7, 
						       kickTarget, 1);
  
  skills->exec(worldSnap, kickCommand);

  // report tactical action 
  tacticalSnap->reportOurTacticalAction(TA_PASS, worldSnap->my.number, passReceiver);
  return true;
}



bool StandardSituationModule::doingGoalieFreeKickUs(AdvSkills *skills,
						    const Snapshot *worldSnap,
						    TacticalSnapshot *tacticalSnap)
{
  if (!iAmGoalie)
    { 
      if ( sqrnorm(worldSnap->my.pos - worldSnap->ball.pos) < sqr(ballAvoidDist) )
	{
	  avoidBall(skills, worldSnap, tacticalSnap);
	  return false;
	}
      DECISIONTRACE( "I am fieldplayer, doing nothing" ); 
      avoidBall(skills, worldSnap, tacticalSnap);
      return false;
    }

  // I am goalie
  if ( !skills->kickSafe(worldSnap) )
    {
      DECISIONTRACE( "Goalie wants to do the goalkick, but has to go to the ball first" );
      DECISIONTRACE( "dist2Ball=" << norm(worldSnap->my.pos - worldSnap->ball.pos) );
      Vector2d targetPos = worldSnap->ball.pos;
      targetPos.setX(targetPos.getX()-PLAYER_SIZE-BALL_SIZE);
      
      double r = BALL_SIZE+PLAYER_SIZE;
      Vector2d m = (r-1.5*BALL_SIZE)/norm(worldSnap->ball.pos-targetPos)*
	(worldSnap->ball.pos-targetPos) + 
	worldSnap->ball.pos;
      DRAWTRACE( CIRCLE(m, r) );
      calcWayTo(worldSnap->my.pos, targetPos, STAMINA_INC_MAX, m, r, worldSnap);
      walkAroundPlayers(worldSnap, targetPos, STAMINA_INC_MAX, &m, r);
      DECISIONTRACE( "Goalie going to " << targetPos );
      skills->exec(worldSnap, skills->goToPosition(worldSnap, targetPos, 0.5));
      waitingTime = 3; 
      return true;
    } // running to ball (if)
  
  // now getting some wm-info and let the positionmodule do it's job
  // (we want many many passpartners when doing the freekick)

  if (waitingTime > 8)
    {
      skills->basicTurn(M_PI/4.0);
      DECISIONTRACE( "Goalie wants to wait because waitingTime= " << waitingTime
		     << " turning" ); 
      waitingTime --;
      return true;
    }
  
  if (waitingTime > 6)
    {
      skills->basicTurn(-worldSnap->my.bodyDir);
      DECISIONTRACE( "Goalie turns to dir 0.0, waitingTime=" << waitingTime );
      waitingTime--;
      return true;
    }

  if (waitingTime > 0)
    {
      waitingTime --;
      DECISIONTRACE( "Goalie looking around, waitingTime=" << waitingTime );
      lookAround(skills, worldSnap);
      return true;
    }

  DECISIONTRACE( "I am goalie and want to do " << worldSnap->playMode );
  // if not ballkickable => walk towards it
  Vector2d moveTarget, kickTarget;
  if ( goalieMayMove &&
       (goalieMovesInCurrentSituation < GOALIE_MAX_MOVES) &&
       (goalieMovesInCurrentSituation < 1 || goalieShouldMoveAgain) )
    {
      calcGoalieMoveAndKick(worldSnap, tacticalSnap,
			    moveTarget, kickTarget, true);
      if (worldSnap->my.distTo(moveTarget) > 1.0 ) 
	{
	  goalieMovesInCurrentSituation++;
	  goalieShouldMoveAgain = false;
	  DECISIONTRACE( "goalie moving to " << moveTarget
			 << " goalieMovesInCurrSit: " << goalieMovesInCurrentSituation);
	  skills->basicMove(moveTarget.getX(), moveTarget.getY());
	  waitingTime = standardWaitingTime/2;
	  return true;
	}
    }
  else
    {
      // we have moved or may not, so do the kick

      calcGoalieMoveAndKick(worldSnap, tacticalSnap,
			    moveTarget, kickTarget, false);
      DECISIONTRACE( "goalie has already been moved or may not move, so waiting for"
		     << " the other modules" );
    }

  // next time another module should do some action so 
  waitingTime = waitingTimeAfterGoalieMove;
  nextCycleDoNothing = true;
  DECISIONTRACE( "doing nothing, plan to let other modules do their job" );
  skills->noAction();
  lookAround(skills, worldSnap);

  return true;
}



Vector2d StandardSituationModule::calcKickOffKickTarget(const Snapshot *worldSnap, 
							TacticalSnapshot *tacticalSnap,
							int &passReceiver)
{
  int ourSystem = getOurTacticalSystem(tacticalSnap);  
  int tacticalPassReceiver = 7; // the player I want to pass to (tactical number !!)

  passReceiver = tacticalPassReceiver;           // to avoid warnings

  // check what positions the other team has
  int upperY=0;
  int lowerY=0; 
  for (int i=1; i<=11; i++) 
    { 
      if (worldSnap->otherTeam(i).age > maxAge)
	continue;
      if (worldSnap->otherTeam(i).pos.getX() > 30.0)  
	continue; // players far away from the middle line don't matter
      if (worldSnap->otherTeam(i).pos.getY() < 3.0)
	lowerY++;
      if (worldSnap->otherTeam(i).pos.getY() > -3.0)
	upperY++;
    }

  switch(ourSystem)
    {
    case 442:
      if (upperY > lowerY)
	tacticalPassReceiver = 8;
      else if (upperY < lowerY)
	tacticalPassReceiver = 7;
      else
	tacticalPassReceiver = 10;
      break;
    case 343:
      tacticalPassReceiver = 7;
      break;
    case 352:
      tacticalPassReceiver = 7;
      break;
    default:
      passReceiver = 7;
      ERRORTRACE( "unknown tactical system, kicking into the crowd !!" );
      break;
    }
  DECISIONTRACE( "tactical system: " << ourSystem << " kick for player " 
		 << tacticalPassReceiver );


  // now get the player's position
  passReceiver = tacticalPassReceiver;
  for (int i=1; i<=11; i++)
    if (tacticalSnap->ourTeamTacticalNo(i) == tacticalPassReceiver)
      {
	passReceiver = i;
	break;
      }

  if (worldSnap->ourTeam(passReceiver).age < maxAge)
    return worldSnap->ourTeam(passReceiver).pos;
  else
    return Vector2d(-10.0, 0.0);
}


void StandardSituationModule::calcShortWays(const Snapshot *worldSnap, 
					    int &firstPlayer, int &secondPlayer,  
					    const Vector2d &targetPos1, 
					    const Vector2d &targetPos2)
{
  double d1 = norm(targetPos1-worldSnap->ourTeam(firstPlayer).pos)+
    norm(targetPos2-worldSnap->ourTeam(secondPlayer).pos);
  double d2 = norm(targetPos1-worldSnap->ourTeam(secondPlayer).pos)+
    norm(targetPos2-worldSnap->ourTeam(firstPlayer).pos);
  
  // which way is the shortest
  if (d1 - 2. > d2)
    {
      int tmp = firstPlayer;
      firstPlayer = secondPlayer;
      secondPlayer = tmp;
    }
  // if they are close
  if (fabs(d1-d2) < 2.)
    {
      // the player with the smaller number is the leftmost of these two
      if (secondPlayer > firstPlayer)
	{
	  int tmp = firstPlayer;
	  firstPlayer = secondPlayer;
	  secondPlayer = tmp;
	}
    }
}



int StandardSituationModule::getOurTacticalSystem(const TacticalSnapshot *tacticalSnap)
{
  // check what our tactical system is
  if ( (tacticalSnap->ourSystem.numDefenders == 4) &&
       (tacticalSnap->ourSystem.numMidfield == 4) &&
       (tacticalSnap->ourSystem.numOffenders == 2) )
    return 442;
  if ( (tacticalSnap->ourSystem.numDefenders ==3) &&
       (tacticalSnap->ourSystem.numMidfield == 4) &&
       (tacticalSnap->ourSystem.numOffenders == 3) )
    return 343;
  if ( (tacticalSnap->ourSystem.numDefenders ==3) &&
       (tacticalSnap->ourSystem.numMidfield == 5) &&
       (tacticalSnap->ourSystem.numOffenders == 2) )
    return 352;
  return 343;        //our default system
}



Vector2d StandardSituationModule::calcEmergencyMovePosition(const Snapshot *worldSnap)
{
  double valueLeft = 0.0, valueRight = 0.0;

  valueLeft -= 3.0 * 
    worldSnap->otherTeamInCorridor(Vector2d(-FIELD_LENGTH/2.0,-FIELD_WIDTH/3.0), 
				   Vector2d(-FIELD_WIDTH/6.0, -FIELD_WIDTH/3.0), 
				   FIELD_WIDTH/3.0);
  valueLeft += 1.0 * 
    worldSnap->ourTeamInCorridor(Vector2d(-FIELD_LENGTH/2.0,-FIELD_WIDTH/3.0), 
				 Vector2d(-FIELD_WIDTH/6.0, -FIELD_WIDTH/3.0), 
				 FIELD_WIDTH/3.0);

  valueRight -= 3.0 * 
    worldSnap->otherTeamInCorridor(Vector2d(-FIELD_LENGTH/2.0,FIELD_WIDTH/3.0), 
				   Vector2d(-FIELD_WIDTH/6.0, FIELD_WIDTH/3.0), 
				   FIELD_WIDTH/3.0);
  valueRight += 1.0 * 
    worldSnap->ourTeamInCorridor(Vector2d(-FIELD_LENGTH/2.0,FIELD_WIDTH/3.0), 
				 Vector2d(-FIELD_WIDTH/6.0, FIELD_WIDTH/3.0), 
				 FIELD_WIDTH/3.0);

  if (valueLeft > valueRight)  // left position is better
    return Vector2d(-FIELD_LENGTH / 2.0 + PENALTY_AREA_LENGTH - 1.5, 
		    -PENALTY_AREA_WIDTH / 2.0 + 1.5);
  else
    return Vector2d(-FIELD_LENGTH / 2.0 + PENALTY_AREA_LENGTH - 1.5, 
		    PENALTY_AREA_WIDTH / 2.0 - 1.5);
}

StandardSituationModuleInfo::StandardSituationModuleInfo() 
{
};

StandardSituationModuleInfo::~StandardSituationModuleInfo() 
{
};



Vector2d StandardSituationModule::calcIntersection(Vector2d A, Vector2d B, 
						   Vector2d C, Vector2d D)
  // intersect the lines AB and CD ?
  // Rem.: AB not parallel to CD
{  

  Vector2d E(A);
  E -= C;
  Vector2d F(A);
  F -= B;
  Vector2d G(D);
  G -= C;

  CMDTRACE( cout << "Intersection: A=" << A << ", B=" << B << ", C=" << C << ", D=" << D << endl );
  CMDTRACE( cout << "  E=" << E << ", F=" << F << ", G=" << G << endl );
  

  double detM = F.getX()*G.getY() - F.getY()*G.getX();

  if (fabs(detM) < eps)
    {
      CMDTRACE( cerr << "det(M)= " << detM << endl );
      return A;
    }

  double t = E.getX()*G.getY() - G.getX()*E.getY();
  double s = F.getX()*E.getY() - F.getY()*E.getX();

  t /= detM;
  s /= detM;

  Vector2d S1 = F;
  S1.stretch(1.0-t);
  S1 += B;

  Vector2d S2 = G;
  S2.stretch(s);
  S2 += C;

  if (sqrnorm(S1-S2) > sqr(0.2))
    CMDTRACE( cerr << "Error, norm = " << norm(S1-S2) << endl );
  
  CMDTRACE( cout << "  t=" << t << ", s=" << s << ", S1=" << S1 << ", S2=" << S2 << endl );

  return S1;
}




double StandardSituationModule::calcDistAroundCircle(Vector2d A, Vector2d B, Vector2d M, 
						     double r, const Snapshot *worldSnap)
{
  if (worldSnap);        // to avoid warnings

  Vector2d MMinusA = M;
  MMinusA -= A;
  Vector2d MMinusB = M;
  MMinusB -= B;

  double alpha = asin(r/norm(MMinusA));
  double beta  = asin(r/norm(MMinusB));

  Vector2d leftPos  = MMinusA;
  Vector2d rightPos = MMinusA;
  leftPos.turn(alpha);
  leftPos += A;
  rightPos.turn(-alpha);
  rightPos += A;
  
  Vector2d hatA, hatB; 
  if (sqrnorm(leftPos - B) < sqrnorm(rightPos - B))
    {
      hatA = leftPos;
      hatB = MMinusB;
      hatB.turn(-beta);
    }
  else
    {
      hatA = rightPos;
      hatB = MMinusB;
      hatB.turn(beta);
    }

  hatA += A;
  hatB += B;

  Vector2d S = calcIntersection(A, hatA, B, hatB);


  DRAWTRACE( LPOSITION(leftPos) );
  DRAWTRACE( LPOSITION(rightPos) );

  DRAWTRACE( CIRCLE(M, r) );
  DRAWTRACE( LINE(A, S) );
  DRAWTRACE( LINE(B, S) );
  DRAWTRACE( POSITION(hatA) );
  DRAWTRACE( POSITION(hatB) );
  DRAWTRACE( POSITION(M) );
  
  OTHERTRACE( "calcDistAroundCircle: A=" << A << ", B=" << B
	      << ", hatA=" << hatA << ", hatB=" << hatB );

  return (norm(S-A) + norm(S-B));
}

bool StandardSituationModule::mustAvoidBall(const Snapshot *worldSnap)
{
  return (sqrnorm(worldSnap->my.pos - worldSnap->ball.pos) < sqr(ballAvoidDist) );
}

void StandardSituationModule::avoidBall(AdvSkills *skills, const Snapshot *worldSnap, 
					TacticalSnapshot *tacticalSnap)
{
  if (!mustAvoidBall(worldSnap))
    {
      ERRORTRACE( "avoidBall: player is not closer than ballAvoidDist=" << ballAvoidDist 
		  << " so doing nothing" );
      skills->noAction();
    }

  Vector2d mypos(worldSnap->my.pos);
  Vector2d walkTarget(mypos);


  // check what my tactical task is
  int myTacticalNo = tacticalSnap->myTacticalNo;

  if (myTacticalNo == UNKNOWN)
    {
      DECISIONTRACE( "  tacticalNo unknown !!" );
      if (mypos.getY() < worldSnap->ball.pos.getY() )
	walkTarget.setXY(mypos.getX() + 2.*ballAvoidDist, mypos.getY() - ballAvoidDist);
      else
	walkTarget.setXY(mypos.getX() + 2.*ballAvoidDist, mypos.getY() + ballAvoidDist);
      
    }
  else if (tacticalSnap->ourTeamDefender(myTacticalNo))
    {
      walkTarget.setX(mypos.getX() - 1.5*ballAvoidDist);
    }
  else if (tacticalSnap->ourTeamMidfield(myTacticalNo))
    {
      if (mypos.getY() < worldSnap->ball.pos.getY() )
	walkTarget.setY(mypos.getY() - 1.5*ballAvoidDist);
      else
	walkTarget.setY(mypos.getY() + 1.5*ballAvoidDist);	
    }
  else if (tacticalSnap->ourTeamOffender(myTacticalNo))
    {
      walkTarget.setX(mypos.getX() + 1.5*ballAvoidDist);      
    }
  else // I am goalie
    {
      walkTarget.setX(mypos.getX() - 2.0*ballAvoidDist);
    }

  // check if walkTarget in the field 
  if (walkTarget.getY() > FIELD_WIDTH/2.0)
    walkTarget.setY(walkTarget.getY() - 2.*ballAvoidDist);
  if (walkTarget.getY() < -FIELD_WIDTH/2.0)
    walkTarget.setY(walkTarget.getY() + 2.*ballAvoidDist);
  
  if (walkTarget.getX() > FIELD_LENGTH/2.0)
    walkTarget.setX(walkTarget.getX() - 3.*ballAvoidDist);
  if (walkTarget.getX() < -FIELD_LENGTH/2.0)
    walkTarget.setX(walkTarget.getX() + 3.*ballAvoidDist);
  
  DECISIONTRACE( "avoidBall: I am near the ball, but do not want to" );
  DECISIONTRACE( " do the freekick, walking to " << walkTarget );

  calcWayTo(mypos, walkTarget, STAMINA_INC_MAX, worldSnap->ball.pos, 
	    BALL_SIZE+PLAYER_SIZE, worldSnap);
  DECISIONTRACE( " after calcWayto: walking to " << walkTarget );

  skills->exec(worldSnap, skills->goToPosition(worldSnap, walkTarget, PLAYER_SIZE));
  return;
}


Grade StandardSituationModule::evalFreeKickUs(AdvSkills* skills,
					      const Snapshot* worldSnap, 
					      TacticalSnapshot* tacticalSnap)
{
  if (skills);      // to avoid warnings

  Grade value;
    
  // if I am the goalie and want another module to act, do nothing and
  // return BAD
  if (waitingTime < 0)
    waitingTime = 3; 
  if (iAmGoalie && nextCycleDoNothing)
    {
      nextCycleDoNothing = false;
      value.set_BAD();
      DECISIONTRACE( "Goalie wants the other modules to kick away the ball" );
      return value;
    }
  
  // If I am the goalie and have catched the ball => move and kick
  if ( (lastTime_PLAY_ON <= worldSnap->lastCatchTime) &&
       ( worldSnap->ourGoalieNo != UNKNOWN) && 
       ( sqrnorm(worldSnap->ourTeam(worldSnap->ourGoalieNo).pos - worldSnap->ball.pos) < 1.0 ) )
    {
      DECISIONTRACE( "goalie has catched the ball in cycle" 
		     << worldSnap->lastCatchTime
		     << " lastTime_PLAY_ON= " << lastTime_PLAY_ON
		     << " so doing goalieMove&Kick" );
      goalieDoesFreeKick = true;
      goalieMayMove = true;
    }
  else
    {
      DECISIONTRACE( "goalie has not catched the ball (last catch in cycle " 
		     <<  worldSnap->lastCatchTime << ") so normal goalieMove&Kick" );
      if ( (worldSnap->ball.pos.getX() < -FIELD_LENGTH/2. + PENALTY_AREA_LENGTH -1.0 ) &&
	   (fabs(worldSnap->ball.pos.getY()) < PENALTY_AREA_WIDTH / 2.0 - 1.0) )
	{
	  DECISIONTRACE( "freekick in our penalty area, goalie should do it" );
	  goalieDoesFreeKick = true;
	  goalieMayMove = false;
	  goalieShouldMoveAgain = false;
	}
      else
	goalieDoesFreeKick = false;
    }
  
  if (goalieDoesFreeKick)
    {
      if (iAmGoalie)
	value.set_ABSOLUTE();
      else 
	{
	  // give players near the ball no chance to kick the ball
	  // better let them walk away
	  if (mustAvoidBall(worldSnap))
	    value.set_ABSOLUTE();
	  else
	    value.set_IMPOSSIBLE();
	}
      return value;
    }

   
  // goalie does not do the freekick
  firstPlayer = UNKNOWN;
  secondPlayer = UNKNOWN;
  // Do we need more than one player for the freekick
  needSecond = true;
  // if we are close to the edge of the field
  if (fabs(fabs(worldSnap->ball.pos.getY())-FIELD_WIDTH/2.0) < 2.0)
    //do a kick in
    calcPlayersAndPos4KickInUs(worldSnap, tacticalSnap, firstPlayer, secondPlayer, 
			       needSecond, targetPos1, targetPos2);
  else
    // do a free kick
    calcPlayersAndPos4FreeKickUs(worldSnap, tacticalSnap, firstPlayer, secondPlayer, 
				 needSecond, targetPos1, targetPos2);
  if (worldSnap->my.number == firstPlayer ||
      (worldSnap->my.number == secondPlayer && needSecond) )
    {
      value.set_ABSOLUTE();
      // max time to wait for second player to arrive its position
      if (waitingTime < 0)
	waitingTime = standardWaitingTime;
      return value;
    }

  DECISIONTRACE( "I am not responsible for ball" );
  if (mustAvoidBall(worldSnap))
    value.set_ABSOLUTE();
  else
    value.set_BAD();
  return value;
}
