//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// -*- c++ -*- ///////////////////////////////////////////////////
//
// $Id: compare_player.C,v 1.6 2000/10/19 15:27:56 goetz Exp $
// $Manager: holgi$
//
//  compare_player.C - replay comm.log and player.log files
//                     and compare with the soccer.log file
//
//  author: <holgi>
//
/////////////////////////////////////////////////////////////////////

#include "compare_player.h"

/////////////
//
// int CompPlayer::OpenFile(const char *logfilename)
//
//////////////

int CompPlayer::OpenFile(const char *logfilename)
{
  if (logfileopen == true) {
    cerr << "error - there's already a logfile open"
	 << endl;
    return 0;
  }
  
  logfile.open(logfilename, ios::in); // assign file logfilename with logfile
                                      // for reading

  if (!logfile) {                        // file doesn't exists
    cerr << "error while opening file "
         << logfilename
         << endl;
    return 0;
  }
  
  char testen[4];                          // read header of file
  int i = 0;
  
  while (i <= 2) {                         // error while reading header
    if (!(logfile >> testen[i])) {
      cerr << "error while reading header - can't read header at position "
	   << i+1
           << endl;
      return 0;
    }
    i++;
  }

  if ((testen[0] != 'U') &&              // wrong header
      (testen[1] != 'L') && 
      (testen[2] != 'G')) {   
    cerr << "error while reading header - wrong header"
	 << endl;
    return 0;
  }
  
  if (!(logfile >> testen[3])) {          // reading version-number
    cerr << "error while reading version"
         << endl; 
    return 0;
  }

  logfileopen = true; // set flag to true, because:
  return 1;           // logfile successfully read
}

/////////////
//
// CompPlayer::int nextSnapshot(char *ownteamname, Snapshot &snap)
//
////////////////

int CompPlayer::nextSnapshot(const char *ownteamname, 
			     int callingPlayer, Snapshot &snap)
{
  if (logfileopen == false) {
    cerr << "error while reading next snapshot - no logfile opened"
	 << endl;
    return 0;
  }
  
  showinfo_t show;

  // attention! for readnextShow(show):
  // showinfo_t with pmode between goal and kickoff should be not read
  // more accurate: don't read PM_AfterGoal_Left/Right
  //                go on with PM_KickOff_Left/Right


  if (readnextShow(show) == 1) {                                // sucessfully read showinfo_t
    if (ShowToSnap(ownteamname, callingPlayer, show, snap))     // sucessfully show -> snap
      return 1;
    else {                                                      // no show -> snap
      cerr << "error while converting showinfo_t to Snapshot"
	   << endl;
      return 0;
    }
  }
  else {                                                        // no showinfo_t read
    cerr << "error while reading next showinfo_t"
	 << endl;
    return 0;
  }
}

/////////////
//
// int CompPlayer::CloseFile()
//
//////////////

int CompPlayer::CloseFile()
{
  if (logfileopen == true) {
    logfile.close();
    return 1;
  }
  else {
    cerr << "error - there's no logfile to close"
	 << endl;
    return 0;
  } 
}

//////////
//
// int CompPlayer::readnextShow()
//
/////////////

int CompPlayer::readnextShow(showinfo_t &show)
{
  int mode = 0;

  do {
    mode = readMode();  // read next mode
    if (mode == -1)     // error occured while reading mode
      return 0;
    else {
      switch (mode) {   // mode should be  0, 1, 2 or 3
      case 0:                          
	break;                            // no info, do nothing
      case 1:
	show = readShow();                // read showinfo_t to show

     	// test, if show contains "bad" playmode = PM_AfterGoal_Left/Right
	// in such a "bad" case, set mode to -1 and continue with the next entry
        if ((show.pmode == PM_AfterGoal_Left) || (show.pmode == PM_AfterGoal_Right)) 
	  mode = -1;

	break;
      case 2:
	msginfo_t msginfo = readMsg();    // read msginfo_t for nothing
	break;
      case 3:
	drawinfo_t drawinfo = readDraw(); // read drawinfo_t for nothing
	break;
      default:
	cerr << "error in logfile - "
	     << mode
	     << " is no valid mode"
	     << endl;
	return 0;
	break;
      }
    }
  }
  while (mode != 1); // while no showinfo_t found in logfile

  return 1; // successfully read showinfo_t from logfile
}

/////////////
//
// int CompPlayer::ShowToSnap(char *ownteamname, showinfo_t show, snap &Snapshot)
//
//////////////

#warning compare_player::ShowToSnap does not work: wrong player numbers etc.
int CompPlayer::ShowToSnap(const char *ownteamname, 
			   int callingPlayer, showinfo_t show, Snapshot &snap)
{
  // set ball positions
  snap.ball.pos.setXY(double(show.pos[0].x)/SHOWINFO_SCALE, double(show.pos[0].y)/SHOWINFO_SCALE);
  
  // compute ball-velocity
  Vector2d BalVel;
  if (lastBallPos.getX() == -1000)
    BalVel.setRadAng(0,0);
  else {
    BalVel = snap.ball.pos - lastBallPos;
    BalVel.stretch(BALL_DECAY);
  }
  snap.ball.v  = BalVel;
  lastBallPos = snap.ball.pos; 
  
  // something does not work correctly here.
  snap.ball.v.setRadAng(0,0);

  // get side of calling player using ownteamname
  int side = -1;                                    // -1: error occured
  if (strcmp(ownteamname, show.team[0].name) == 0)  //  0: myTeam = players  1-11
    side = 0;                                       //  1: myTeam = players 12-22
  if (strcmp(ownteamname, show.team[1].name) == 0)
    side = 1;
  if (side == -1) {
    cerr << "error - teamname of calling player isn't matching with teamnames in the logfile"
	 << endl;
    return 0;
  }
  
  // set time
  snap.time = int(show.time);

  // set side 
  int helpside = 0;
  if (side == 0)
    helpside = int(show.pos[1].side);
  else
    helpside = int(show.pos[12].side);
  if (helpside == 1)
    snap.ourSide = SIDE_LEFT;
  if (helpside == -1) 
    snap.ourSide = SIDE_RIGHT;
  if (helpside == 0) {
    cerr << "error - teams side couldn't be read "
	 << endl;
    return 0;
  }
  
  // set score
  if (side == 0) {
    snap.ourScore = int(show.team[0].score);
    snap.otherScore = int(show.team[1].score);
  }
  else {
    snap.ourScore = int(show.team[1].score);
    snap.otherScore = int(show.team[0].score);
  }

  // set players' positions and numbers and age
  // the first 11 players ...
  for (int i = 1; i <= 11; i++) {
    double posx = double(show.pos[i].x)/SHOWINFO_SCALE; // get the x-pos ...
    double posy = double(show.pos[i].y)/SHOWINFO_SCALE; // ... and the y-pos ...
    int number = int(show.pos[i].unum);                 // ... of player with number ... 
    if (side == 0) {
      WorldPlayer player;
      player.pos.setXY(posx, posy);                     // ... and save them to myTeam ...
      player.number = number;
      player.age = 0;
      snap.setOurTeam(number, player);
    }
    else {
      WorldPlayer player;
      player.pos.setXY(posx, posy);                     // ... or otherTeam (depending on side)
      player.number = number;
      player.age = 0;
      snap.setOtherTeam(number, player);
    }
  }
  // ... and for the last 11 player the same procedure
  for (int i = 12; i <= 22; i++) {
    double posx = double(show.pos[i].x)/SHOWINFO_SCALE;
    double posy = double(show.pos[i].y)/SHOWINFO_SCALE;
    int number = int(show.pos[i].unum);
    if (side == 0) {
      WorldPlayer player;
      player.pos.setXY(posx, posy);                     // ... or otherTeam (depending on side)
      player.number = number;
      player.age = 0;
      snap.setOtherTeam(number, player);
    }
    else {
      WorldPlayer player;
      player.pos.setXY(posx, posy);                     // ... and save them to myTeam ...
      player.number = number;
      player.age = 0;
      snap.setOurTeam(number, player);
    }
  }

  // set snap.my to the right values
  snap.my = snap.ourTeam(callingPlayer);
  
  return 1;
}

///////////////
//
// int CompPlayer::readMode()
//
///////////////

int CompPlayer::readMode()
{
  char dummy;
  short mode = 0;
  
  if (!(logfile >> dummy))  // try do read character
    return -1;
  else {
    mode = short(dummy);
    return int(mode);       // and return the int-value
  }
}

/////////////
//
// showinfo_t CompPlayer::readShow()
//
////////////

showinfo_t CompPlayer::readShow()
{
  showinfo_t wanted;  
  char *buffer = (char*)&wanted;
  for (unsigned int i = 0; i < sizeof(wanted); ++i)
    logfile.get(buffer[i]);
  return wanted;
}

//////////
//
// msginfo_t CompPlayer::readMsg()
//
/////////////

msginfo_t CompPlayer::readMsg()
{
  msginfo_t wanted;
  char *buffer = (char*)&wanted;
  for (unsigned int i = 0; i < sizeof(wanted); ++i)
    logfile.get(buffer[i]);
  return wanted;
}

//////////
//
// drawinfo_t CompPlayer::readDraw()
//
/////////////

drawinfo_t CompPlayer::readDraw()
{
  drawinfo_t wanted;
  char *buffer = (char*)&wanted;
  for (unsigned int i = 0; i < sizeof(wanted); ++i)
    logfile.get(buffer[i]);
  return wanted;
}
