//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// drawables.h - Graphical objects for displaying in the replayer
//
// $Id: drawables.C,v 1.3 2000/10/19 15:27:56 goetz Exp $
// $master: goetz$
//
////////////////////////////////////////////////////////////////////////

#include "coordinates.h"
#include "worldmap.h"
#include "drawables.h"

DPosition::DPosition(const double x, const double y, const bool bright):
  Drawables(x, y, bright)
{}

void DPosition::draw(WorldMap *wm)
{
  wm->drawCircle(x, y, 1, true, bright);
}


DCircle::DCircle (const double x, const double y, const double rad, 
		  const bool grid, const bool bright) :
  Drawables(x, y, bright), rad(rad), grid(grid)
{}

void DCircle::draw (WorldMap *wm)
{
  wm->drawCircle(x, y, rad, grid, bright);
}

DLine::DLine (const double x1, const double y1, 
	      const double x2, const double y2,
	      const bool bright) :
  Drawables(x1, y1, bright), end_x(x2), end_y(y2)
{}

void DLine::draw (WorldMap *wm)
{
  wm-> drawLine(x, y, end_x, end_y, bright);
}

DText::DText (const double x, const double y, 
	      const char *text, 
	      const bool bright) :
  Drawables(x, y, bright)
{
  this->text= new char[strlen(text)+1];
  strcpy(this->text, text);
}

DText::~DText()
{
  delete text;
}

void DText::draw (WorldMap *wm)
{
  wm->drawText(x, y, text, bright);
}

DrawablesList::DrawablesList() : 
  dlist(0) 
{}

DrawablesList::~DrawablesList()
{
  clear();
}

DrawablesList& DrawablesList::add(Drawables *d)
{
  d->next= dlist;
  dlist= d;
  return *this;
} 

DrawablesList& DrawablesList::clear()
{
  while (remove()) ;
  return *this;
}

bool DrawablesList::remove()
{
  if (dlist) {
    Drawables *d= dlist;
    dlist= dlist->next;
    delete d;
    return true;
  } 
  return false;
}

DrawablesList& DrawablesList::draw(WorldMap *wm)
{
  Drawables *d= dlist;
  while (d) {
    d->draw(wm);
    d= d->next;
  }
  return *this;
}

