//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// drawables.h - Graphical objects for displaying in the replayer
//
// $Id: drawables.h,v 1.3 2000/10/19 15:27:57 goetz Exp $
// $master: goetz$
//
////////////////////////////////////////////////////////////////////////

#ifndef _DRAWABLES_H_
#define _DRAWABLES_H_

#include "coordinates.h"
#include "worldmap.h"

/** all Drawables objects
 */
class Drawables 
{
  friend class DrawablesList; 
public:
  Drawables(const double x, const double y, const bool bright) :
    x(x), y(y), bright(bright) {};
  virtual void draw(WorldMap *wm) = 0;
private:
  Drawables *next; // for list structure
protected:
  double x;
  double y;
  bool bright;
};

/** position object
 */
class DPosition : public Drawables 
{
public:
  DPosition (const double x, const double y, const bool bright= false);
  virtual void draw (WorldMap *wm);
};

/** circle object
 */
class DCircle : public Drawables
{
public:
  DCircle (const double x, const double y, const double rad, 
	   const bool grid= false, const bool bright= false);
  virtual void draw (WorldMap *wm);
protected:
  double rad;
  bool grid;
};

/** line object
 */
class DLine : public Drawables
{
public:
  DLine (const double x1, const double y1, 
	 const double x2, const double y2,
	 const bool bright= false);
  virtual void draw (WorldMap *wm);
protected:
  double end_x;
  double end_y;
};

/** text string object
 */
class DText : public Drawables
{
public:
  DText (const double x, const double y, 
	 const char *text, 
	 const bool bright= false);
  virtual ~DText ();
  virtual void draw (WorldMap *wm);
protected:
  char *text;
};

/** storage class for Drawables's
 */
class DrawablesList
{
public:
  /** with list initialization! */
  DrawablesList();
  /** with garbage colleciton: cleaning up the list */
  ~DrawablesList();
  /** add a Drawables to the end of the list */
  DrawablesList& add(Drawables *d); 
  /** remove all list entries */
  DrawablesList& clear();
  /** remove last element added to list
   *  true if done; false if empty list
   */
  bool remove();
  /** draw all contained Drawabless to the given WorldMap */
  DrawablesList& draw(WorldMap *wm);
private:
  Drawables* dlist;
};

#endif
