//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//// -*- c++ -*- ///////////////////////////////////////////////////
//
// $Id: replayer.C,v 1.16 2000/10/19 15:27:57 goetz Exp $
// $Manager: eri$
//
//  replayer.C - replay MRB99 comm.log and player.log files
//
//  author: <eri>
//
/////////////////////////////////////////////////////////////////////


#include "replayer.h"


// main
/////////////////////////////////////////////////////////////////////



int main (int argc, const char** argv) {

  if (argc == 3) {

    RePlayer* replayer = new RePlayer (argv[1], atoi(argv[2]));
    
    // depending on what you want to do now, 
    // call either replay() for interactive replaying,
    // or evaluate() for numerical evaluation of snapshots (ie comparison
    // with objective snapshots etc).

    replayer->replay();
    // replayer->evaluate();
    
  } else cerr << "Syntax: compplayer <teamName> <playerNo>" << endl
	      << "  if a coach snapshot logfile is present," << endl
	      << "  it will be read and displayed as well."
	      << endl;

  cout << endl;
  
}


// Constructor / Destructor
/////////////////////////////////////////////////////////////////////



RePlayer::RePlayer (const char* _team, int _number) {

  number = _number;
  team = new char[strlen(_team)];
  strcpy (team, _team);

  char displayTitle [1024];
  sprintf (displayTitle, "%s No. %d", team, number);
    


  // 1. read player output logfile
  char playerlog [1024];
  sprintf (playerlog, "%s%s%d%s", PLAYER_LOG, team, number, EXTENSION);
  cout << "reading " << playerlog << " ...";
  playerlogReader = new PlayerlogReader(playerlog);


  // 2. read subjective snapshots
  char snaplog [1024];
  sprintf (snaplog, "%s%s%d%s", SNAP_LOG, team, number, EXTENSION);
  
  cout << "reading " << snaplog << " ...";
  snaplogReader = new SnaplogReader(snaplog);
  maxTime = snaplogReader->getMaxTime();
  
  // 3. try to read objective coach snapshots
  sprintf (snaplog, "%s%s%d%s", SNAP_LOG, team, 0, EXTENSION);
  
  cout << "reading " << snaplog << " ...";
  coachlogReader = new SnaplogReader(snaplog);
  if ( coachlogReader->getMaxTime() == -1 ) {
    twoSnap = false;
    mapwindow_two = NULL;
    mapwindow = new WorldMap_r (displayTitle, MAPWINDOW_SIZE);
  } else {

    if ( coachlogReader->getMaxTime() < maxTime )
      maxTime = coachlogReader->getMaxTime();

    twoSnap = true;
    mapwindow = NULL;
    mapwindow_two = new WorldMap_r_two_snap(displayTitle, MAPWINDOW_SIZE);
  }
  
  drawables= new DrawablesList();
}
 


RePlayer::~RePlayer() {
  
  if (twoSnap) mapwindow_two->Close();
  else mapwindow->Close();

  delete(mapwindow);
  delete(mapwindow_two);
  delete(drawables);
  delete (playerlogReader);
  delete (snaplogReader);
  delete (coachlogReader);
  delete (team);
  
}


// public methods
/////////////////////////////////////////////////////////////////////

void RePlayer::evaluate() {
  // exemplary way of calculating difference of subjective and 
  // objective ball pos

  // implement any other method here, if you want to use the compplayer
  // for different, non-interactive purposes
  
  // test for error (coachlogs missing)
  if (!twoSnap) {
    cerr << "need snaplog and coachlog !" << endl;
    return;
  }

  double sum = 0;
  int count = 0;
  
  // loop over all snapshots read from both files
  for (int i=1; i<maxTime ; i++) {

    // get player snapshot
    Snapshot *playersnap;
    playersnap = snaplogReader->getSnapshot(i);
    
    // get coach snapshot
    Snapshot *coachsnap;
    coachsnap = coachlogReader->getSnapshot(i);
    if (coachsnap) { // coachsnap exists
      // set "my" explicitly, as otherwise snapshot->my is coach !
      coachsnap->my = coachsnap->ourTeam(number);
    }

    if (playersnap && coachsnap) { // both snaps exist

      // do evaluation here:
      Vector2d absposdiff = coachsnap->ball.pos - playersnap->ball.pos;
      sum += absposdiff.getLength();
      count += 1;

    } else {
      cerr << "*** did not get snapshots for time " << i << endl;
    }
    
  }
  
  cout << "mean ball position error over timesteps 1.." << maxTime 
       << " = " << sum/count << endl;
  
}


void RePlayer::replay() {
  // interactive control method: 
  // read input, parse input, playTo, and printOutput.
  
  char input [1024];

  bool showhelp= false;
  bool autodel= true;
  bool showplayergraph= true;
  
  int  time=1;
  if (time > maxTime) time = maxTime;
  
  int  stopTime=time;
  bool playOn = false;
  
  // open displays
  if (twoSnap) mapwindow_two->Open();
  else mapwindow->Open();

  while (time > -1) {
    // map snapshots
    Snapshot *playersnap;
    Snapshot *coachsnap;
    playersnap = snaplogReader->getSnapshot(time);

    // draw graphics from player log
    if (showplayergraph) 
      playerlogReader->drawOutput(time, drawables);   
      
    if (twoSnap) {
      coachsnap = coachlogReader->getSnapshot(time);
      if (coachsnap)
	coachsnap->my = coachsnap->ourTeam(number); // set "my" explicitly !
      if (playersnap && coachsnap) {
	mapwindow_two->Map(playersnap, coachsnap, 20);
      } else {
	cout << "Replayer: playersnap or coachsnap missing for time=" 
	     << time << endl;
      }
      // draw lines, circles and stuff
      drawables->draw(mapwindow_two);
    } else {
      if (playersnap) {
	mapwindow->Map(playersnap, 20);
      } else {
	cout << "Replayer: playersnap missing for time=" << time << endl;
      }
      mapwindow->Map(playersnap, 20);
      // draw lines, circles and stuff
      drawables->draw(mapwindow);
    }
    
    // test for end of play loop
    if (stopTime == time) playOn = false;
    
    // do intercative stuff only when not in play loop
    if (!playOn) {
  
      playerlogReader->printCommands(time);  // print player commands
      
      playerlogReader->printOutput(time);    // print additional output
      
      // do interactive input
      cout << endl
	   << "--------------------------------------------------------------" 
	   << endl 
	   << "---(+, -, <time>, f<time>, b<time>, -1 to exit)--(Time "
	   << time << "/" << maxTime << ")------" 
	   << endl
	   << "Draw: c<x>,<y>,<rad>[g][l], p<x>,<y>[l], "
	   << endl
	   << "l<x1>,<y1>,<x2>,<y2>[l], u, n, a, g, ? gives help"
	   << endl;

      if (!autodel)
	cout << "auto-delete feature switched off." << endl;
      if (!showplayergraph)
	cout << "player log graphic show feature switched off." << endl;

      cout << endl << endl;
      
      if (showhelp) {
	showhelp= false;
	cout << "Help: (options of drawing commands)"
	     << endl
	     << "All drawing commands: l=light color"
	     << endl
	     << "c - circle: g= with grid"
	     << endl
	     << "p - position"
	     << endl
	     << "l - line"
	     << endl
	     << "n - new (clear all objects)"
	     << endl
	     << "u - undo (remove last object)"
	     << endl
	     << "a - toggle auto delete after each cycle"
	     << endl
	     << "g - toggle show player graphics feature"
	     << endl
	     << "DON'T enter any whitespaces!"
	     << endl << endl;
      }

      cout << "Command: ";
      
      cin >> input;
      cout << endl;
      
      // scan input
      int t;
      double x, y, x2, y2, rad;
      char c1(0), c2(0);

      if ( input[0] == '?' || input[0] == 'h' ) {
	showhelp= true;
      } else if (input[0] == 'n') {
	drawables->clear();
      } else if (input[0] == 'u') {
	drawables->remove();
      } else if (input[0] == 'a') {
	autodel= !autodel;
      } else if (input[0] == 'g') {
	showplayergraph= !showplayergraph;
      } else if (sscanf (input, "c%lf,%lf,%lf%c%c", 
			 &x, &y, &rad, &c1, &c2) >= 3) {
	drawables->add(new DCircle(x, y, rad, 
				  c1 == 'g' || c2 == 'g',
 				  c1 == 'l' || c2 == 'l' ));
      } else if (sscanf (input, "l%lf,%lf,%lf,%lf%c", 
			 &x, &y, &x2, &y2, &c1) >= 4) {
	drawables->add(new DLine(x, y, x2, y2, c1 == 'l'));
      } else if (sscanf (input, "p%lf,%lf%c", 
			 &x, &y, &c1) >= 2) {
	drawables->add(new DPosition(x, y, c1 == 'l'));
      } else if (sscanf (input, "%d" , &t) == 1) {
	time = t;
	// time can be set to "-1" here to quit
	if (autodel) drawables->clear();
      } else if (sscanf (input, "f%d", &t ) == 1) {
	stopTime = time+t;
	if (stopTime > maxTime) stopTime = maxTime;
	playOn = true;
      } else if (sscanf (input, "b%d", &t ) == 1) {
	stopTime = time-t;
	if (stopTime < 0) stopTime = 0;
	playOn = true;
      } else if (input[0]=='+') {
	time++;
	if (time > maxTime) time = maxTime;
	if (autodel) drawables->clear();
      } else if (input[0]=='-') {
	time--;
	if (time < 0) time = 0;
	if (autodel) drawables->clear();
      } 
    } else { // in play loop
      
      cout << "(Time " << time << "/" << maxTime << ")" << endl << endl;
      if (stopTime > time) time++;
      else if (stopTime < time) time--;

      if (autodel) drawables->clear();

      usleep (100*1000);
      
    }
    
  } // main while loop
  
}
