//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
//
// $Id: worldmap_r_two_snap.h,v 1.5 2000/10/19 15:27:58 goetz Exp $	
// $Manager: goetz$
// 
// 
//  worldmap_r_two_snap.h - World model display with additional relative map
//                          And display of two snapshots at the same time
//                          in same window
// 
//  ( Christian Bauer <cebix> )
//   Goetz Schwandtner
//
//
//////////////////////////////////////////////////////////////////////

#ifndef _WcORLDMAP_R_TWO_SNAP_H_
#define _WORLDMAP_R_TWO_SNAP_H_

#include "worldmodel.h"
#include "worldmap_r.h"
#include "coordinates.h"

#include <X11/Xlib.h>
#include <X11/keysym.h>

// Display of actual worldmodel data
class WorldMap_r_two_snap : public WorldMap_r {
public:
  // create new worldmaps
  // absolute worldmap has magnification factor "mag"
  // relative worldmap has magnification factor "relmag"
  // If "relmag" is set to negative values, "relmag" gets same value as
  // "mag". If "relmag" is set to 0.0, it means that there will be no
  // relative map at all
  WorldMap_r_two_snap(const char *windowtitle = NULL, 
		      double mag = 5.0,
		      double relmag = -1.0);


  // NOTE: Destructor & close remain the same as in superclass

  // Open map window. True if oK
  bool Open();

  // Display snapshot with onbjects with age <= max_age only
  void Map(Snapshot *snap_primary, Snapshot *snap_secondary, int max_age);
protected:
  // private drawing functions
  void drawownplayer(WorldPlayer ownplayer);
  void drawball(WorldBall ball);
  void drawplayer(WorldPlayer player);
  void drawmyteam(Snapshot *snap);
  void drawotherteam(Snapshot *snap);
  void setownpos(WorldPlayer ownplayer);

  // colors in dark, hence beginning with "d"
  XColor dred_color;
  XColor dgreen_color;
  XColor dblue_color;
  XColor dyellow_color;
  unsigned long dred;
  unsigned long dgreen;
  unsigned long dblue;
  unsigned long dyellow;

  // Position of own player for relative map window in secondary snapshot
  double sec_ownx;
  double sec_owny;
};

#endif
